SystemOrganization addCategory: #'System-Support'!

!Behavior methodsFor: '*System-Support' stamp: ''!
allCallsOn
	"Answer a SortedCollection of all the methods that refer to me by name or as part of an association in a global dict."


	^ (self  systemNavigation allReferencesTo:  (self environment associationAt: self theNonMetaClass name)), (self  systemNavigation allCallsOn:  self theNonMetaClass name)	! !

!Behavior methodsFor: '*System-Support' stamp: ''!
allCallsOn: aSymbol
	"Answer a SortedCollection of all the methods that call on aSymbol."


	^ self  systemNavigation allCallsOn: aSymbol from: self .
	! !

!Behavior methodsFor: '*System-Support' stamp: ''!
allUnsentMessages
	"Answer an array of all the messages defined by the receiver that are not sent anywhere in the system."

	^ SystemNavigation default allUnsentMessagesIn: self selectors! !

IdentityDictionary subclass: #SystemDictionary
	instanceVariableNames: 'cachedClassNames cachedNonClassNames'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'System-Support'!

!SystemDictionary commentStamp: 'nice 3/6/2010 21:56' prior: 0!
I represent a special dictionary used as global namespace for class names :

	Smalltalk globals classNames.

and for traits too:

	Smalltalk globals traitNames.
	
and a few other globals:

	(Smalltalk globals keys reject: [:k | (Smalltalk globals at: k) isBehavior])
			collect: [:k | k -> (Smalltalk globals at: k) class].

As the above example let you guess, the global namespace of Smalltalk system is accessed through:

	Smalltalk globals.!

!SystemDictionary class methodsFor: 'instance creation' stamp: ''!
withOrganizer: anOrganizer
	| dictionary |
	dictionary := self new.
	dictionary organization: anOrganizer. "update too the back pointer"
	^dictionary.
	! !

!SystemDictionary methodsFor: 'class and trait names' stamp: ''!
allClasses  
	"Return all the class defines in the Smalltalk SystemDictionary"
	"Smalltalk allClasses"

	^ self classNames collect: [:name | self at: name]! !

!SystemDictionary methodsFor: 'class and trait names' stamp: ''!
allClassesAndTraits
	"Return all the classes and traits defined in the Smalltalk SystemDictionary"

	^ self classNames , self traitNames collect: [:each | self at: each]! !

!SystemDictionary methodsFor: 'class and trait names' stamp: ''!
allClassesAndTraitsDo: aBlock
	^self allClassesAndTraits do: aBlock! !

!SystemDictionary methodsFor: 'class and trait names' stamp: ''!
allClassesDo: aBlock
	"Evaluate the argument, aBlock, for each class in the system."

	(self classNames collect: [:name | self at: name]) do: aBlock! !

!SystemDictionary methodsFor: 'class and trait names' stamp: ''!
allTraits
	"Return all traits defined in the Smalltalk SystemDictionary"

	^ self traitNames collect: [:each | self at: each]! !

!SystemDictionary methodsFor: 'class and trait names' stamp: ''!
allTraitsDo: aBlock
	"Evaluate the argument, aBlock, for each trait in the system."

	(self traitNames collect: [:name | self at: name]) do: aBlock! !

!SystemDictionary methodsFor: 'dictionary access' stamp: ''!
associationOrUndeclaredAt: key 
	"return an association or install in undeclared.  Used for mating up ImageSegments."

	^ self associationAt: key ifAbsent: [
		Undeclared at: key put: nil.
		Undeclared associationAt: key]! !

!SystemDictionary methodsFor: 'dictionary access' stamp: ''!
at: aKey put: anObject 
	"Override from Dictionary to check Undeclared and fix up
	references to undeclared variables."
	| index element |
	(self includesKey: aKey) ifFalse: 
		[self declare: aKey from: Undeclared.
		self flushClassNameCache].
	super at: aKey put: anObject.
	^ anObject! !

!SystemDictionary methodsFor: 'accessing' stamp: ''!
bindingOf: varName
	"SystemDictionaries includes symbols only"
	
	^ super bindingOf: varName asSymbol! !

!SystemDictionary methodsFor: 'class and trait names' stamp: ''!
classNamed: className 
	^self classOrTraitNamed: className.! !

!SystemDictionary methodsFor: 'classes and traits' stamp: 'ul 10/8/2010 21:22'!
classNames
	"Answer a sorted collection of all class names. Use the return value of #fillCaches to avoid concurrency issues."
	
	^cachedClassNames ifNil: [ self fillCaches at: 1 ]! !

!SystemDictionary methodsFor: 'class and trait names' stamp: ''!
classOrTraitNamed: aString
	"aString is either a class or trait name or a class or trait name followed by ' class' or 'classTrait' respectively.
	Answer the class or metaclass it names."

	| meta baseName baseClass |
	(aString endsWith: ' class')
		ifTrue: [ 
			meta := true.
			baseName := aString copyFrom: 1 to: aString size - 6 ]
		ifFalse: [ 
			(aString endsWith: ' classTrait')
				ifTrue: [ 
					meta := true.
					baseName := aString copyFrom: 1 to: aString size - 11 ]
				ifFalse: [ 
					meta := false.
					baseName := aString ] ].
	baseClass := Smalltalk globals at: baseName asSymbol ifAbsent: [ ^ nil ].
	meta
		ifTrue: [ ^ baseClass classSide ]
		ifFalse: [ ^ baseClass ]! !

!SystemDictionary methodsFor: 'accessing' stamp: ''!
environment
	"for conversion from SmalltalkImage to SystemDictionary"
	
	^ self ! !

!SystemDictionary methodsFor: 'classes and traits' stamp: 'ul 10/8/2010 22:44'!
fillCaches
	"Fill cachedClassNames and cachedNonClassNames. Return an array with the calculated values."

	| classNames nonClassNames |
	classNames := OrderedCollection new: self size.
	nonClassNames := OrderedCollection new.
	self keysAndValuesDo: [ :key :value |
		"The key == value name test below addresses two separate issues:
			1) Obsolete classes, where key = #Foo and value name = 'AnObsoleteFoo'
			2) Aliases, i.e., Smalltalk at: #OtherName put: aClass"
		((value isKindOf: Class) and: [ key == value name ])
			ifTrue: [ classNames add: key ]
			ifFalse: [ nonClassNames add: key ] ].
	classNames sort.
	cachedNonClassNames := nonClassNames sort.
	cachedClassNames := classNames.
	^{ classNames. nonClassNames }! !

!SystemDictionary methodsFor: 'classes and traits' stamp: 'ul 10/8/2010 21:23'!
flushClassNameCache
	"Smalltalk flushClassNameCache"
	"Force recomputation of the cached list of class names and non-class names."

	cachedClassNames := cachedNonClassNames := nil! !

!SystemDictionary methodsFor: 'class and trait names' stamp: ''!
forgetClass: aClass logged: aBool 
	"Delete the class, aClass, from the system.
	Note that this doesn't do everything required to dispose of a class - to do that use Class>>removeFromSystem."

	aBool ifTrue: [SystemChangeNotifier uniqueInstance classRemoved: aClass fromCategory: aClass category].		
	self organization removeElement: aClass name.
	Smalltalk removeFromStartUpList: aClass.
	Smalltalk removeFromShutDownList: aClass.
	self removeKey: aClass name ifAbsent: [].
	self flushClassNameCache! !

!SystemDictionary methodsFor: 'testing' stamp: 'MarcusDenker 11/30/2010 11:10'!
hasBindingThatBeginsWith: aString
	"Use the cached class and non-class names for better performance."

	| name searchBlock |
	searchBlock := [ :element |
		(element beginsWith: aString)
			ifTrue: [ 0 ]
			ifFalse: [
				aString < element
					ifTrue: [ -1 ]
					ifFalse: [ 1 ] ] ].
	name := self classNames 
		findBinary: searchBlock
		ifNone: [nil].
	name ifNotNil: [ ^true ].
	name := self nonClassNames 
		findBinary: searchBlock
		ifNone: [nil].
	^name notNil! !

!SystemDictionary methodsFor: 'class and trait names' stamp: ''!
hasClassNamed: aString
	"Answer whether there is a class of the given name, but don't intern aString if it's not alrady interned."

	Symbol hasInterned: aString ifTrue: 
		[:aSymbol | ^ (self at: aSymbol ifAbsent: [nil]) isKindOf: Class].
	^ false! !

!SystemDictionary methodsFor: 'deprecated' stamp: ''!
hasSpecialSelector: aLiteral ifTrueSetByte: aBlock

	self deprecated: 'Use Smalltalk instead' on: '20 March 2010' in: 'Pharo1.1'.
	^ Smalltalk hasSpecialSelector: aLiteral ifTrueSetByte: aBlock! !

!SystemDictionary methodsFor: 'classes and traits' stamp: 'ul 10/8/2010 21:22'!
nonClassNames
	"Answer a sorted collection of all non-class names. Use the return value of #fillCaches to avoid concurrency issues."
	
	^cachedNonClassNames ifNil: [ self fillCaches at: 2 ]! !

!SystemDictionary methodsFor: 'accessing' stamp: ''!
organization
	"Return the organizer for the receiver"
	
	^ self at: #SystemOrganization! !

!SystemDictionary methodsFor: 'accessing' stamp: ''!
organization: anOrganization
	"Return the organizer for the receiver"
	
	anOrganization environment: self.
	^self at: #SystemOrganization put: anOrganization! !

!SystemDictionary methodsFor: 'accessing' stamp: ''!
poolUsers
	"Answer a dictionary of pool name -> classes that refer to it.
	Also includes any globally know dictionaries (such as
	Smalltalk, Undeclared etc) which although not strictly
	accurate is potentially useful information"
	"Smalltalk poolUsers"
	| poolUsers |
	poolUsers := Dictionary new.
	self keys
		do: [:k | | pool refs | "yes, using isKindOf: is tacky but for reflective code like
			this it is very useful. If you really object you can:-
			a) go boil your head.
			b) provide a better answer.
			your choice."
			(((pool := self at: k) isKindOf: Dictionary)
					or: [pool isKindOf: SharedPool class])
				ifTrue: [refs := self systemNavigation allClasses
								select: [:c | c sharedPools identityIncludes: pool]
								thenCollect: [:c | c name].
					refs
						add: (self systemNavigation
								allReferencesTo: (self associationAt: k)).
					poolUsers at: k put: refs]].
	^ poolUsers! !

!SystemDictionary methodsFor: 'printing' stamp: ''!
printElementsOn: aStream
	aStream nextPutAll:'(lots of globals)'! !

!SystemDictionary methodsFor: 'class and trait names' stamp: ''!
removeClassNamed: aName
	"Invoked from fileouts:  if there is currently a class in the system named aName, then remove it.  If anything untoward happens, report it in the Transcript.  "

	| oldClass |
	(oldClass := self at: aName asSymbol ifAbsent: [nil]) == nil
		ifTrue:
			[Transcript cr; show: 'Removal of class named ', aName, ' ignored because ', aName, ' does not exist.'.
			^ self].

	oldClass removeFromSystem! !

!SystemDictionary methodsFor: 'dictionary access' stamp: ''!
removeKey: key ifAbsent: aBlock
	"Remove key (and its associated value) from the receiver. If key is not in
	the receiver, answer the result of evaluating aBlock. Otherwise, answer
	the value externally named by key."
	
	self flushClassNameCache.
	^super removeKey: key ifAbsent: aBlock! !

!SystemDictionary methodsFor: 'renaming' stamp: ''!
renameClass: aClass as: newName 
	"Rename the class, aClass, to have the title newName."

	^self renameClass: aClass from: aClass name to: newName! !

!SystemDictionary methodsFor: 'renaming' stamp: ''!
renameClass: aClass from: oldName 
	"Rename the class, aClass, to have the title newName."

	^self renameClass: aClass from: oldName to: aClass name! !

!SystemDictionary methodsFor: 'renaming' stamp: ''!
renameClass: aClass from: oldName to: newName
	"Rename the class, aClass, to have the title newName."

	| oldref category |
	category := self organization categoryOfElement: oldName.
	self organization classify: newName under: category.
	self organization removeElement: oldName.
	oldref := self associationAt: oldName.
	self removeKey: oldName.
	oldref key: newName.
	self add: oldref.  "Old association preserves old refs"
	Smalltalk renamedClass: aClass from: oldName to: newName.
	self flushClassNameCache.
	SystemChangeNotifier uniqueInstance classRenamed: aClass from: oldName to: newName inCategory: category! !

!SystemDictionary methodsFor: 'renaming' stamp: ''!
renameClassNamed: oldName as: newName
	"Invoked from fileouts:  if there is currently a class in the system named oldName, then rename it to newName.  If anything untoward happens, report it in the Transcript.  "

	| oldClass |
	(oldClass := self at: oldName asSymbol ifAbsent: [nil]) == nil
		ifTrue:
			[Transcript cr; show: 'Class-rename for ', oldName, ' ignored because ', oldName, ' does not exist.'.
			^ self].

	oldClass rename: newName! !

!SystemDictionary methodsFor: 'class and trait names' stamp: ''!
traitNames
	"Answer a SortedCollection of all traits (not including class-traits) names."
	| names |
	names := OrderedCollection new.
	self do: 
		[:cl | ((cl isKindOf: Trait)
			and: [(cl name beginsWith: 'AnObsolete') not])
				ifTrue: [names add: cl name]].
	^ names! !

!SystemDictionary methodsFor: 'copying' stamp: ''!
veryDeepCopyWith: deepCopier
	"Return self.  I can't be copied.  Do not record me."! !

Categorizer subclass: #SystemOrganizer
	instanceVariableNames: 'environment'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'System-Support'!

!SystemOrganizer commentStamp: '<historical>' prior: 0!
My instances provide an organization for the classes in the system, just as a ClassOrganizer organizes the messages within a class. The only difference is the methods for fileIn/Out.!

!SystemOrganizer class methodsFor: 'cleanup' stamp: ''!
cleanUp: agressive
	"Remove empty categories when cleaning aggressively"

	agressive ifTrue: [SystemOrganization removeEmptyCategories].
! !

!SystemOrganizer methodsFor: 'accessing' stamp: ''!
addCategory: newCategory
	| r |
	r := super addCategory: newCategory.
	SystemChangeNotifier uniqueInstance classCategoryAdded: newCategory.
	^ r! !

!SystemOrganizer methodsFor: 'remove' stamp: 'LucFabresse 10/24/2010 17:42'!
categoriesMatching: matchString
	"Return all matching categories"
	
	self categories ifNil: [ ^#() ].
	^ self categories select: [:c | matchString match: c]! !

!SystemOrganizer methodsFor: 'utils' stamp: ''!
classesInCategory: category
	^ (self listAtCategoryNamed: category) collect: [ :className | self environment at: className ] ! !

!SystemOrganizer methodsFor: 'utils' stamp: ''!
commentInventory: categoryName
	"SystemOrganization commentInventory: 'Morphic*'"

	| classes commentedClasses |
	classes := OrderedCollection new.
	self categories
		withIndexCollect: [ :cat :idx | 
			(categoryName match: cat)
				ifTrue: [ classes addAll: (self listAtCategoryNumber: idx) ]
				ifFalse: [ nil ] ].
	commentedClasses := classes select: [ :catCls | (self environment at: catCls) hasComment ].
	^ 'There are ' , classes size asString , ' classes in ' , categoryName , ' of which ' , commentedClasses size asString
		, ' have comments and ' , (classes size - commentedClasses size) asString , ' do not yet have comments.'! !

!SystemOrganizer methodsFor: 'accessing' stamp: ''!
environment
	 ^ environment ifNil: [ environment := Smalltalk globals]! !

!SystemOrganizer methodsFor: 'accessing' stamp: ''!
environment: aSystemDictionary
	 environment := aSystemDictionary.
! !

!SystemOrganizer methodsFor: 'filein/out' stamp: ''!
fileOut  
	"SystemOrganization fileOut"

	| internalStream |
	internalStream := (String new: 30000) writeStream.
	internalStream nextPutAll: 'SystemOrganization changeFromCategorySpecs: #('; cr;
		print: SystemOrganization;  "ends with a cr"
		nextPutAll: ')!!'; cr.

	FileStream writeSourceCodeFrom: internalStream baseName: (FileDirectory default nextNameFor: 'SystemOrganization' extension: 'st') isSt: true
! !

!SystemOrganizer methodsFor: 'filein/out' stamp: ''!
fileOutCategory: category 
	"Store on the file named category (a string) concatenated with '.st' all the 
	classes associated with the category."
	
	| internalStream |
	internalStream := (String new: 1000) writeStream.
	self fileOutCategory: category on: internalStream initializing: true.
	^ FileStream writeSourceCodeFrom: internalStream baseName: category isSt: true ! !

!SystemOrganizer methodsFor: 'filein/out' stamp: ''!
fileOutCategory: category on: aFileStream 
	"Store on the file associated with aFileStream, all the classes associated 
	with the category and any requested shared pools."
	^self fileOutCategory: category on: aFileStream initializing: true! !

!SystemOrganizer methodsFor: 'filein/out' stamp: ''!
fileOutCategory: category on: aFileStream initializing: aBool
	"Store on the file associated with aFileStream, all the traits and classes associated 
	with the category and any requested shared pools in the right order."

	| first poolSet tempClass classes traits |
	traits := self orderedTraitsIn: category.
	classes := self superclassOrder: category.
	poolSet := Set new.
	classes do:  [:class | class sharedPools do: [:eachPool | poolSet add: eachPool]].
	poolSet size > 0 ifTrue: [
		tempClass := Class new.
		tempClass shouldFileOutPools ifTrue: [
			poolSet := poolSet select: [:aPool |
				tempClass shouldFileOutPool: (Smalltalk globals keyAtIdentityValue: aPool)].
			poolSet do: [:aPool | tempClass fileOutPool: aPool onFileStream: aFileStream]]].
	first := true.
	traits, classes do: [:each | 
		first
			ifTrue: [first := false]
			ifFalse: [aFileStream cr; nextPut: Character newPage; cr].
		each
			fileOutOn: aFileStream
			moveSource: false
			toFile: 0
			initializing: false].
	aBool ifTrue: [classes do: [:cls | cls fileOutInitializerOn: aFileStream]].! !

!SystemOrganizer methodsFor: 'private' stamp: ''!
ifClassOrganizerDo: aBlock
	"Do nothing, since this is not a class organizer"! !

!SystemOrganizer methodsFor: 'initialize-release' stamp: 'LucFabresse 10/24/2010 19:57'!
initialize
	super initialize.
	categoryArray := Array new.
	categoryStops := Array new.
	elementArray := Array new! !

!SystemOrganizer methodsFor: 'utils' stamp: ''!
orderedTraitsIn: category
	"Answer an OrderedCollection containing references to the traits in the 
	category whose name is the argument, category (a string). The traits 
	are ordered so they can be filed in." 

	| behaviors traits |
	behaviors := (self listAtCategoryNamed: category asSymbol) collect: [ :title | self environment at: title ].
	traits := behaviors reject: [ :each | each isBehavior ].
	traits := traits
		asSortedCollection: [ :t1 :t2 | (t2 traitComposition allTraits includes: t1) or: [ (t1 traitComposition allTraits includes: t2) not ] ].
	^ traits asArray! !

!SystemOrganizer methodsFor: 'remove' stamp: ''!
removeCategoriesMatching: matchString
	"Remove all matching categories with their classes"
	(self categoriesMatching: matchString) do:
		[:c | self removeSystemCategory: c]! !

!SystemOrganizer methodsFor: 'accessing' stamp: ''!
removeCategory: cat 
	| r |
	r := super removeCategory: cat.
	SystemChangeNotifier uniqueInstance classCategoryRemoved: cat.
	^ r! !

!SystemOrganizer methodsFor: 'utils' stamp: ''!
removeMissingClasses
	"Remove any class names that are no longer in the Smalltalk dictionary. Used for cleaning up after garbage collecting user-generated classes."

	"SystemOrganization removeMissingClasses"
 
	elementArray copy
		do: [ :el | 
			(self environment includesKey: el)
				ifFalse: [ self removeElement: el ] ]! !

!SystemOrganizer methodsFor: 'remove' stamp: ''!
removeSystemCategory: category
	"remove all the classes and traits associated with the category"

	 (self orderedTraitsIn: category) , (self superclassOrder: category)
		reverseDo: [:each | each removeFromSystem].

	self removeCategory: category.
! !

!SystemOrganizer methodsFor: 'accessing' stamp: ''!
renameCategory: oldCatString toBe: newCatString
	| r |
	r := super renameCategory: oldCatString toBe: newCatString.
	SystemChangeNotifier uniqueInstance 
		classCategoryRenamedFrom: oldCatString to: newCatString.
	^ r! !

!SystemOrganizer methodsFor: 'utils' stamp: ''!
superclassOrder: category
	"Answer an OrderedCollection containing references to the classes in the 
	category whose name is the argument, category (a string). The classes 
	are ordered with superclasses first so they can be filed in."

	| behaviors classes |
	behaviors := (self listAtCategoryNamed: category asSymbol) collect: [ :title | self environment at: title ].
	classes := behaviors select: [ :each | each isBehavior ].
	^ ChangeSet superclassOrder: classes ! !

!SystemOrganizer methodsFor: 'query' stamp: ''!
uncommentedClassesIn: categoryName
	"SystemOrganization uncommentedClassesIn: 'Morphic*'"

	| classes |
	classes := OrderedCollection new.
	self categories
		withIndexCollect: [ :cat :idx | 
			(categoryName match: cat)
				ifTrue: [ classes addAll: (self listAtCategoryNumber: idx) ]
				ifFalse: [ nil ] ].
	^ (classes collect: [ :clsName | self environment at: clsName ] thenSelect: [ :cls | cls hasComment not ]) asArray! !

Model subclass: #Project
	instanceVariableNames: 'world changeSet transcript parentProject displayDepth projectParameters'
	classVariableNames: 'AllProjects CurrentProject UIProcess'
	poolDictionaries: ''
	category: 'System-Support'!

!Project commentStamp: 'tk 12/2/2004 12:38' prior: 0!
A Project stores the state of a complete Squeak desktop, including
the windows, and the currently active changeSet.  A project knows who
its parent project is.  When you change projects, whether by entering
or exiting, the screen state of the project being exited is saved in
that project.

A project is retained by its view in the parent world.  It is
effectively named by the name of its changeSet, which can be changed
either by renaming in a changeSorter, or by editing the label of its
view from the parent project.

As the site of major context switch, Projects are the locus of
swapping between the old MVC and the new Morphic worlds.  The
distinction is based on whether the variable 'world' contains a
WorldMorph or a ControlManager.

Saving and Loading
Projects may be stored on the disk in external format.  (Project
named: 'xxx') exportSegment, or choose 'store project on file...'.
Projects may be loaded from a server and stored back.  Storing on a
server never overwrites;  it always makes a new version.  A project
remembers the url of where it lives in urlList.  The list is length
one, for now.  The url may point to a local disk instead of a server.
All projects that the user looks at, are cached in the Squeaklet
folder.  Sorted by server.  The cache holds the most recent version
only.

When a project is loaded into Squeak, its objects are converted to
the current version.  There are three levels of conversion.  First,
each object is converted from raw bits to an object in its old
format.  Then it is sent some or all of these messages:
	comeFullyUpOnReload: smartRefStream  		Used to
re-discover an object that already exists in this image, such as a
resource, global variable, Character, or Symbol.  (sent to objects in
outPointers)
	convertToCurrentVersion: varDict refStream: smartRefStrm
		fill in fields that have been added to a class since
the object was stored.  Used to set the extra inst var to a default
value.  Or, return a new object of a different class.  (sent to
objects that changed instance variables)
	fixUponLoad: aProject refStream: smartRefStrm
	change the object due to conventions that have changed on the
project level.  (sent to all objects in the incoming project)

Here is the calling sequence for storing out a Project:
Project saveAs
Project storeOnServer
Project storeOnServerWithProgressInfo
Project storeOnServerInnards
Project exportSegmentFileName:directory:
Project exportSegmentWithChangeSet:fileName:directory:
ImageSegment writeForExportWithSources:inDirectory:changeSet:
---------
Isolation (not used any more)
When you accept a method, the entire system feels the change, except
projects that are "isolated".  In an isolated project, all new global
variables (including new classes) arestored in the project-local
environment, and all changes to preexisting classes are revoked when
you leave the project.  When you enter another project, that
project's changes are invoked.  Invocation and revocation are handled
efficiently by swapping pointers.  To make a project be isolated,
choose 'isolate changes of this project' from the 'changes...'
section of the screen menu.  You can use an isolated project for
making dangerous change to a system, and you can get out if it
crashes.  A foreign application can have the separate environment it
wants.  Also, you can freeze part of the system for a demo that you
don't want to disturb.  An isolated project shares methods with all
subprojects inside it, unless they are isolated themselves.   Each
isolated project is the head of a tree of projects with which it
shares all methods.

You may 'assert' all changes ever made in the current project to take
effect above this project.  This amounts to exporting all the globals
in the current environment, and zapping the revocation lists to that
the current state of the world will remain in force upon exit from
this project.

[Later: A project may be 'frozen'.  Asserts do not apply to it after
that.  (Great for demos.)  You should be informed when an assert was
blocked in a frozen project.]

Class definitions are layered by the isolation mechanism.  You are
only allowed to change the shape of a class in projects that lie
within its isolation scope.  All versions of the methods are
recompiled, in all projects.  If you remove an inst var that is in
use in an isolated project, it will become an Undeclared global.  It
is best not to remove an inst var when it is being used in another
isolated project. [If we recompile them all, why can't we diagnose
the problem before allowing the change??]

Senders and Implementors do not see versions of a method in isolated
projects.  [again, we might want to make this possible at a cost].
When you ask for versions of a method, you will not get the history
in other isolated projects.

Moving methods and classes between changeSets, and merging changeSets
has no effect on which methods are in force.  But, when you look at a
changeSet from a different isolated project, the methods will contain
code that is not in force.  A changeSet is just a list of method
names, and does not keep separate copies of any code.

When finer grained assertion is needed, use the method (aProject
assertClass: aClass from: thisProject warn: warnConflicts).

How isolated changes work: The first time a class changes, store its
MethodDictionary object.  Keep parallel arrays of associations to
Classes and MethodDictionaries.  Traverse these and install them when
you enter an "ioslated project".  When you leave, store this
project's own MethodDictionaries there.
	To do an assert, we must discover which methods changed here,
and which changed only in the project we are asserting into.  There
is one copy of the 'virgin' method dictionaries in the system.  It is
always being temporarily stored by the currently inForce isolated
project.

isolatedHead - true for the top project, and for each isolated
project.  false or nil for any subproject that shares all methods
with its parent project.

inForce -  true if my methods are installed now.  false if I am
dormant. [is this equivalent to self == Project Current?]

classArray - list of associations to classes

methodDictArray - the method dictionaries of those classes before we
started changing methods.  They hang onto the original
compiledMethods.  (If this project is dormant, it contains the method
dictionaries of those classes as they will be here, in this project).

orgArray - the class organizations of the classes in classArray.

UsingIsolation (class variable) - No longer used.

When you want to save a project in export format from within that
very project, it gets tricky.  We set two flags in parentProject,
exit to it, and let parentProject write the project.
ProjectViewMorph in parentProject checks in its step method, does the
store, clears the flags, and reenters the subProject.

!

!Project class methodsFor: 'constants' stamp: ''!
current
	"Answer the project that is currently being used."

	^CurrentProject! !

!Project class methodsFor: 'utilities' stamp: ''!
interruptName: labelString preemptedProcess: theInterruptedProcess
	"Create a Notifier on the active scheduling process with the given label."
	| preemptedProcess projectProcess |
	ActiveHand ifNotNil:[ActiveHand interrupted].
	ActiveWorld := World. "reinstall active globals"
	ActiveHand := World primaryHand.
	ActiveHand interrupted. "make sure this one's interrupted too"
	ActiveEvent := nil.

	projectProcess := self uiProcess.	"we still need the accessor for a while"
	preemptedProcess := theInterruptedProcess ifNil: [Processor preemptedProcess].
	"Only debug preempted process if its priority is >= projectProcess' priority"
	preemptedProcess priority < projectProcess priority 
		ifTrue:[preemptedProcess := projectProcess].
	preemptedProcess suspend.
	ToolSet interrupt: preemptedProcess label: labelString! !

!Project class methodsFor: 'initialization' stamp: ''!
localeChanged
	self current localeChanged! !

!Project class methodsFor: 'utilities' stamp: ''!
resumeProcess: aProcess
	"Adopt aProcess as the project process -- probably because of proceeding from a debugger"

	UIProcess := aProcess.
	UIProcess resume! !

!Project class methodsFor: 'utilities' stamp: ''!
spawnNewProcess

	UIProcess := [
		[World doOneCycle.  Processor yield.  false] whileFalse: [].
	] newProcess priority: Processor userSchedulingPriority.
	UIProcess resume! !

!Project class methodsFor: 'utilities' stamp: ''!
spawnNewProcessAndTerminateOld: terminate

	self spawnNewProcess.
	terminate
		ifTrue: [Processor terminateActive]
		ifFalse: [Processor activeProcess suspend]! !

!Project class methodsFor: 'instance creation' stamp: ''!
uiProcess

	^ UIProcess! !

!Project methodsFor: 'accessing' stamp: ''!
changeSet
	^ changeSet! !

!Project methodsFor: 'initialization' stamp: 'MarcusDenker 10/28/2010 13:17'!
initialize
	super initialize.
	changeSet := ChangeSet new.
	transcript := TranscriptStream new.
	displayDepth := Display depth.
	parentProject := CurrentProject.
! !

!Project methodsFor: 'project parameters' stamp: ''!
initializeProjectParameters
	"Initialize the project parameters."

	projectParameters := IdentityDictionary new.
	^ projectParameters! !

!Project methodsFor: 'accessing' stamp: ''!
isCurrentProject

	^self == CurrentProject! !

!Project methodsFor: 'accessing' stamp: ''!
labelString
	^ changeSet name! !

!Project methodsFor: 'language' stamp: ''!
localeChanged
	"Set the project's natural language as indicated"
	self projectParameterAt: #localeID put: LocaleID current.
	self updateLocaleDependents! !

!Project methodsFor: 'language' stamp: ''!
localeID
	"Answer the natural language for the project"

	^ LocaleID current! !

!Project methodsFor: 'accessing' stamp: ''!
name
	changeSet == nil ifTrue: [^ 'no name'].
	^ changeSet name! !

!Project methodsFor: 'project parameters' stamp: ''!
noteThatParameter: prefSymbol justChangedTo: aBoolean
	"Provides a hook so that a user's toggling of a project parameter might precipitate some immediate action"

! !

!Project methodsFor: 'project parameters' stamp: ''!
parameterAt: aSymbol
	^self parameterAt: aSymbol ifAbsent:[nil]! !

!Project methodsFor: 'project parameters' stamp: ''!
parameterAt: aSymbol ifAbsent: aBlock
	projectParameters ifNil:[^aBlock value].
	^projectParameters at: aSymbol ifAbsent: aBlock! !

!Project methodsFor: 'accessing' stamp: ''!
parent
	^ parentProject! !

!Project methodsFor: 'accessing' stamp: ''!
projectChangeSet
	^ changeSet! !

!Project methodsFor: 'project parameters' stamp: ''!
projectParameterAt: aSymbol put: aValue
	"Set the given project parameter to the given value"

	self projectParameters at: aSymbol put: aValue.
	self noteThatParameter: aSymbol justChangedTo: aValue.
	^ aValue! !

!Project methodsFor: 'project parameters' stamp: ''!
projectParameters
	^ projectParameters ifNil: [self initializeProjectParameters]! !

!Project methodsFor: 'initialization' stamp: ''!
setChangeSet: aChangeSet

	changeSet := aChangeSet
! !

!Project methodsFor: 'language' stamp: ''!
updateLocaleDependents
	"Set the project's natural language as indicated"
	LanguageEnvironment localeChanged.
! !

!Project methodsFor: 'accessing' stamp: ''!
world
	^ world! !

Object subclass: #Author
	instanceVariableNames: 'fullName'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'System-Support'!

Author class
	instanceVariableNames: 'uniqueInstance'!

!Author commentStamp: 'MiguelCoba 7/25/2009 01:09' prior: 0!
I am responsible for the full name used to identify the current code author.!

Author class
	instanceVariableNames: 'uniqueInstance'!

!Author class methodsFor: 'utilities' stamp: ''!
changeStamp 
	"Answer a string to be pasted into source code to mark who changed it and when."
	^ Author fullName , ' ' , Date today mmddyyyy, ' ',
		((String streamContents: [:s | Time now print24: true on: s]) copyFrom: 1 to: 5)! !

!Author class methodsFor: 'utilities' stamp: ''!
changeStampPerSe
	"Answer a string to be pasted into source code to mark who changed it and when."

	^ (Author fullNamePerSe ifNil: ['.']) , ' ' , Date today mmddyyyy, ' ',
		((String streamContents: [:s | Time now print24: true on: s]) 
			copyFrom: 1 to: 5)! !

!Author class methodsFor: 'utilities' stamp: ''!
fixStamp: changeStamp 
	| parts |
	parts := changeStamp findTokens: ' '.
	(parts size > 0 and: [parts last first isLetter]) ifTrue:
		["Put initials first in all time stamps..."
		^ String streamContents:
				[:s | s nextPutAll: parts last.
				parts allButLast do: [:p | s space; nextPutAll: p]]].
	^ changeStamp! !

!Author class methodsFor: 'convenience' stamp: ''!
fullName
	^ Author uniqueInstance fullName! !

!Author class methodsFor: 'convenience' stamp: ''!
fullName: aString 
	^ Author uniqueInstance checkAndSetFullNameWith: aString! !

!Author class methodsFor: 'convenience' stamp: ''!
fullNamePerSe
	^ Author uniqueInstance fullNamePerSe! !

!Author class methodsFor: 'instance creation' stamp: ''!
new
	self error: 'Author is a singleton -- send uniqueInstance instead'! !

!Author class methodsFor: 'convenience' stamp: ''!
requestFullName
	^ Author uniqueInstance requestFullName! !

!Author class methodsFor: 'instance creation' stamp: ''!
reset
	^ uniqueInstance := nil! !

!Author class methodsFor: 'instance creation' stamp: ''!
uniqueInstance
	^ uniqueInstance ifNil: [ uniqueInstance := super new ]! !

!Author class methodsFor: 'instance creation' stamp: ''!
uniqueInstance: anInstance
	"Needed by AuthorTest to restore saved instance"
	^ uniqueInstance := anInstance! !

!Author class methodsFor: 'testing-support' stamp: ''!
useAuthor: aString during: aBlock
	^ self uniqueInstance useAuthor: aString during: aBlock! !

!Author methodsFor: 'accessing' stamp: 'Lr 12/9/2010 17:34'!
checkAndSetFullNameWith: aString 
	| s allowed |
	aString ifNil: [^self].
	s := ''.
	allowed := ($A to: $Z)
				, ($a to: $z)
				, ($0 to: $9).
	aString
		do: [:c | (allowed includes: c)
				ifTrue: [s := s , c asString]].
	self fullName: s! !

!Author methodsFor: 'accessing' stamp: ''!
fullName
	"Answer the full name to be used to identify the current code author."

	[fullName isEmptyOrNil] whileTrue: [
		self requestFullName.
		(fullName isNil or:[fullName = 'FirstnameLastname' translated]) ifTrue: [fullName := nil. ^'Anonymous']].
	^ fullName! !

!Author methodsFor: 'accessing' stamp: ''!
fullName: aString
	fullName := aString.! !

!Author methodsFor: 'accessing' stamp: ''!
fullNamePerSe

	"Answer the currently-prevailing author full name, such as it is, empty or not"

	^ fullName! !

!Author methodsFor: 'compatibility' stamp: ''!
ifUnknownAuthorUse: aString during: aBlock
	"If no author is set use the name aString while executing aBlock."
	"for compatibility with 1.0"
	
	| previous |
	fullName isEmptyOrNil ifFalse: [ ^ aBlock value ].
	fullName := aString.
	^ aBlock ensure: [ self reset ]! !

!Author methodsFor: 'initialization' stamp: ''!
initialize
	super initialize.
	fullName := ''.! !

!Author methodsFor: 'ui-requests' stamp: ''!
messagePrompt
	^
'Please type your full name. 
It will be used to sign the changes you make to the image.
Spaces, accents, dashes, underscore and similar characters are not allowed
'! !

!Author methodsFor: 'ui-requests' stamp: ''!
requestFullName
	| initialAnswer |
	initialAnswer := fullName isEmptyOrNil
				ifTrue: ['FirstnameLastname' translated]
				ifFalse: [fullName].
	self checkAndSetFullNameWith: ( UIManager default
				request: self messagePrompt
				initialAnswer: initialAnswer
				title: 'Author identification' translated)! !

!Author methodsFor: 'compatibility' stamp: ''!
reset
	fullName := ''! !

!Author methodsFor: 'testing-support' stamp: ''!
useAuthor: aString during: aBlock
	| previous |
	previous := fullName.
	fullName := aString.
	^ aBlock ensure: [ fullName := previous ]! !

Object subclass: #AutoStart
	instanceVariableNames: 'parameters'
	classVariableNames: 'Active InstalledLaunchers UpdateFromServerAtStartup'
	poolDictionaries: ''
	category: 'System-Support'!

!AutoStart class methodsFor: 'private' stamp: ''!
active
	^ Active == true! !

!AutoStart class methodsFor: 'private' stamp: ''!
active: aBoolean
	Active := aBoolean! !

!AutoStart class methodsFor: 'accessing' stamp: ''!
addLauncher: launcher
	self installedLaunchers add: launcher! !

!AutoStart class methodsFor: 'accessing' stamp: ''!
addLauncherFirst: launcher
	self installedLaunchers addFirst: launcher! !

!AutoStart class methodsFor: 'updating' stamp: ''!
checkForUpdates
	| availableUpdate updateServer |
	World 
		ifNotNil: [
			World install.
			ActiveHand position: 100@100].
	^self processUpdates.
	! !

!AutoStart class methodsFor: 'initialization' stamp: ''!
deinstall
	"AutoStart deinstall"

	Smalltalk removeFromStartUpList: AutoStart.
	InstalledLaunchers := nil! !

!AutoStart class methodsFor: 'initialization' stamp: ''!
initialize
	"AutoStart initialize"
	"Order: ExternalSettings, SecurityManager, AutoStart"
	Smalltalk addToStartUpList: AutoStart after: SecurityManager.
	Smalltalk addToShutDownList: AutoStart after: SecurityManager.! !

!AutoStart class methodsFor: 'private' stamp: ''!
installedLaunchers
	InstalledLaunchers ifNil: [
		InstalledLaunchers := OrderedCollection new].
	^InstalledLaunchers! !

!AutoStart class methodsFor: 'updating' stamp: 'Jb 10/23/2010 12:14'!
processUpdates
	"Process update files from a well-known update server.  This method is called at system startup time,   Only if the UpdateFromServerAtStartup class variable is true is the actual update processing undertaken automatically"
	| choice |
	(self updateFromServerAtStartup) ifTrue:
		[choice := UIManager default chooseFrom: #('Yes, Update' 'No, Not now' 'Don''t ask again')
			title: 'Shall I look for new code\updates on the server?' withCRs.
		choice = 1 ifTrue: [
			UpdateStreamer new updateFromServer].
		choice = 3 ifTrue: [
			self updateFromServerAtStartup: false.
			self inform: 'Remember to save you image to make this setting permant.']].
	^false! !

!AutoStart class methodsFor: 'accessing' stamp: ''!
removeLauncher: launcher
	self installedLaunchers remove: launcher ifAbsent: []! !

!AutoStart class methodsFor: 'initialization' stamp: ''!
shutDown: quitting
	self active: false! !

!AutoStart class methodsFor: 'initialization' stamp: ''!
startUp: resuming
	"The image is either being newly started (resuming is true), or it's just been snapshotted.
	If this has just been a snapshot, skip all the startup stuff."

	| startupParameters launchers ws |
	self active
		ifTrue: [ ^ self ].
	self active: true.
	resuming
		ifFalse: [ ^ self ].
	self checkForUpdates
		ifTrue: [ ^ self ].
	startupParameters := SmalltalkImage current extractParameters.
	launchers := self installedLaunchers collect: [ :launcher | launcher new ].
	launchers do: [ :launcher | launcher parameters: startupParameters ].
	ws := Smalltalk globals at: #WorldState ifAbsent: [ nil ].
	(ws notNil and: [Smalltalk globals includesKey: #World])
		ifTrue: [
			launchers do: [:launcher |
				ws addDeferredUIMessage: [launcher startUp]] ]
		ifFalse: [
			launchers do: [:launcher |
				launcher startUp]].
! !

!AutoStart class methodsFor: 'settings' stamp: ''!
updateFromServerAtStartup
	^ UpdateFromServerAtStartup ifNil: [UpdateFromServerAtStartup := false]! !

!AutoStart class methodsFor: 'settings' stamp: ''!
updateFromServerAtStartup: aBoolean
	UpdateFromServerAtStartup := aBoolean! !

Object subclass: #Beeper
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'System-Support'!

Beeper class
	instanceVariableNames: 'default'!

!Beeper commentStamp: 'gk 2/26/2004 22:44' prior: 0!
Beeper provides simple audio (or in some other way) feedback to the user.

The recommended use is "Beeper beep" to give the user the equivalence of a beep. If you want to force the beep to use the primitive in the VM for beeping, then use "Beeper beepPrimitive". In either case, if sounds are disabled there will be no beep.

The actual beeping, when you use "Beeper beep", is done by sending a #play message to a registered playable object. You can register your own playable object by invoking the class side method #setDefault: passing in an object that responds to the #play message.

The default playable object is an instance of Beeper itself which implements #play on the instance side. That implementation delegates the playing of the beep to the default SoundService.

Note that #play is introduced as a common interface between AbstractSound and Beeper.
This way we can register instances of AbstractSound as playable entities, for example:

	Beeper setDefault: (SampledSound new
						setSamples: self coffeeCupClink
						samplingRate: 12000).

Then "Beeper beep" will play the coffeeCup sound.!

Beeper class
	instanceVariableNames: 'default'!

!Beeper class methodsFor: 'beeping' stamp: ''!
beep
	"The preferred way of producing an audible feedback.
	The default playable entity (an instance of Beeper)
	also uses the pluggable SoundService
	mechanism, so it will use the primitive beep only
	if there is no other sound mechanism available."
	
	self default 
		ifNil: [self beepPrimitive]
		ifNotNil: [ self default play].
! !

!Beeper class methodsFor: 'beeping' stamp: ''!
beepPrimitive
	"Make a primitive beep. Only use this if
	you want to force this to be a primitive beep.
	Otherwise use Beeper class>>beep
	since this method bypasses the current
	registered playable entity."

	SoundSettings soundEnabled 
		ifTrue: [self primitiveBeep]! !

!Beeper class methodsFor: 'customize' stamp: ''!
clearDefault
	"Clear the default playable.
	Will be lazily initialized in Beeper class >>default."

	default := nil! !

!Beeper class methodsFor: 'customize' stamp: ''!
default
	"When the default is not defined it is
	initialized using #newDefault."

	default isNil 
		ifTrue: [default := self newDefault ].
	^ default! !

!Beeper class methodsFor: 'customize' stamp: ''!
newDefault
	"Subclasses may override me to provide a default beep.
	This base implementation returns an instance of Beeper
	which uses the pluggable sound service."

	^ self new! !

!Beeper class methodsFor: 'private' stamp: ''!
primitiveBeep
	"Make a primitive beep. Not to be called directly.
	It is much better to use Beeper class>>beep
	or Beeper class>>beepPrimitive
	since this method bypasses the current
	registered playable entity and does not
	check SoundSettings class>>soundEnabled."

	<primitive: 140>
	self primitiveFailed! !

!Beeper class methodsFor: 'customize' stamp: ''!
setDefault: aPlayableEntity
	"Set the playable entity used when making a beep.
	The playable entity should implement the message #play."

	default := aPlayableEntity! !

!Beeper methodsFor: 'play interface' stamp: ''!
play
	"This is how the default Beeper makes a beep,
	by sending beep to the default sound service.
	The sound system will check if sounds are enabled."

	SoundService default beep! !

Object subclass: #DefaultExternalDropHandler
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'System-Support'!

!DefaultExternalDropHandler commentStamp: 'dgd 4/5/2004 19:07' prior: 0!
An alternative default handler that uses the file-list services to process files.
!

!DefaultExternalDropHandler class methodsFor: 'initialization' stamp: ''!
initialize
	"initialize the receiver"
	ExternalDropHandler defaultHandler: self new! !

!DefaultExternalDropHandler class methodsFor: 'initialization' stamp: ''!
unload
	"initialize the receiver"
	ExternalDropHandler defaultHandler: nil! !

!DefaultExternalDropHandler methodsFor: 'private' stamp: ''!
chooseServiceFrom: aCollection
	"private - choose a service from aCollection asking the user if  
	needed"
	aCollection size = 1
		ifTrue: [^ aCollection anyOne].
	^ UIManager default chooseFrom: (aCollection collect: [:each | each label]) values: aCollection.
! !

!DefaultExternalDropHandler methodsFor: 'private' stamp: ''!
servicesForFileNamed: aString 
	"private - answer a collection of file-services for the file named  
	aString"
	| allServices |
	allServices := FileServices itemsForFile: aString.
	^ allServices
		reject: [:svc | self unwantedSelectors includes: svc selector]! !

!DefaultExternalDropHandler methodsFor: 'private' stamp: ''!
unwantedSelectors
	"private - answer a collection well known unwanted selectors "
	^ #(#removeLineFeeds: #addFileToNewZip: #compressFile: #putUpdate: )! !

Object subclass: #ExternalDropHandler
	instanceVariableNames: 'action type extension'
	classVariableNames: 'DefaultHandler RegisteredHandlers'
	poolDictionaries: ''
	category: 'System-Support'!

!ExternalDropHandler class methodsFor: 'private' stamp: ''!
chooseServiceFrom: aCollection
	"private - choose a service from aCollection asking the user if  
	needed"
	aCollection size = 1 ifTrue: [^ aCollection anyOne].
	^ UIManager default chooseFrom: (aCollection collect: [:each | 
			each label]) values: aCollection.
! !

!ExternalDropHandler class methodsFor: 'private' stamp: ''!
defaultGZipHandler
	^ExternalDropHandler
		type: nil
		extension: 'gz'
		action: [:stream :pasteUp :event |
			stream viewGZipContents]! !

!ExternalDropHandler class methodsFor: 'accessing' stamp: ''!
defaultHandler
	DefaultHandler ifNil: [DefaultHandler := ExternalDropHandler type: nil extension: nil action: [:dropStream | dropStream edit]].
	^DefaultHandler! !

!ExternalDropHandler class methodsFor: 'accessing' stamp: ''!
defaultHandler: externalDropHandler
	DefaultHandler := externalDropHandler! !

!ExternalDropHandler class methodsFor: 'private' stamp: ''!
defaultImageHandler
	
	^ExternalDropHandler
		type: 'image/'
		extension: nil
		action: [:stream :pasteUp :event | | image sketch |
			stream binary.
			image := Form fromBinaryStream: ((RWBinaryOrTextStream with: stream contents) reset).
			sketch := World drawingClass withForm: image.
			pasteUp addMorph: sketch centeredNear: event position.
			image := sketch := nil]! !

!ExternalDropHandler class methodsFor: 'initialization' stamp: ''!
initialize
	"ExternalDropHandler initialize"

	self resetRegisteredHandlers.
	self
		registerHandler: self defaultImageHandler;
		registerHandler: self defaultGZipHandler.
	! !

!ExternalDropHandler class methodsFor: 'accessing' stamp: ''!
lookupExternalDropHandler: stream

	| types extension serviceHandler |
	types := stream mimeTypes.

	types ifNotNil: [
		self registeredHandlers do: [:handler | 
			(handler matchesTypes: types)
				ifTrue: [^handler]]].

	extension := FileDirectory extensionFor: stream name.
	self registeredHandlers do: [:handler | 
		(handler matchesExtension: extension)
				ifTrue: [^handler]].
	serviceHandler := self lookupServiceBasedHandler: stream.
	^serviceHandler
		ifNil: [self defaultHandler]! !

!ExternalDropHandler class methodsFor: 'accessing' stamp: ''!
lookupServiceBasedHandler: dropStream
	"the file was just droped, let's do our job"
	| fileName services theOne |
	fileName := dropStream name.

	services := (FileList itemsForFile: fileName)
		reject: [:svc | self unwantedSelectors includes: svc selector].

	"no service, default behavior"
	services isEmpty
		ifTrue: [^nil].

	theOne := self chooseServiceFrom: services.
	^theOne
		ifNotNil: [ExternalDropHandler type: nil extension: nil action: [:stream | theOne performServiceFor: stream]]! !

!ExternalDropHandler class methodsFor: 'accessing' stamp: ''!
registerHandler: aHandler
	self registeredHandlers add: aHandler! !

!ExternalDropHandler class methodsFor: 'initialization' stamp: ''!
registerStandardExternalDropHandlers
	"ExternalDropHandler registerStandardExternalDropHandlers"

	self registeredHandlers add: (
		ExternalDropHandler
			type: 'image/'
			extension: nil
			action: [:stream :pasteUp :event |
				pasteUp addMorph: (World drawingClass withForm: (Form fromBinaryStream: stream binary)) centeredNear: event position])! !

!ExternalDropHandler class methodsFor: 'private' stamp: ''!
registeredHandlers
	RegisteredHandlers ifNil: [RegisteredHandlers := OrderedCollection new].
	^RegisteredHandlers! !

!ExternalDropHandler class methodsFor: 'private' stamp: ''!
resetRegisteredHandlers
	RegisteredHandlers := nil! !

!ExternalDropHandler class methodsFor: 'instance creation' stamp: ''!
type: aType extension: anExtension action: anAction 
	^self new type: aType extension: anExtension action: anAction ! !

!ExternalDropHandler class methodsFor: 'private' stamp: ''!
unwantedSelectors
	"private - answer a collection well known unwanted selectors "
	^ #(#removeLineFeeds: #addFileToNewZip: #compressFile: #putUpdate: )! !

!ExternalDropHandler methodsFor: 'accessing' stamp: ''!
extension
	^extension! !

!ExternalDropHandler methodsFor: 'accessing' stamp: ''!
handle: dropStream in: pasteUp dropEvent: anEvent
	
	^action cull: dropStream 
			cull: pasteUp 
			cull: anEvent
	! !

!ExternalDropHandler methodsFor: 'testing' stamp: ''!
matchesExtension: aExtension
	(self extension isNil or: [aExtension isNil])
		ifTrue: [^false].
	FileDirectory activeDirectoryClass isCaseSensitive
		ifTrue: [^extension = aExtension]
		ifFalse: [^extension sameAs: aExtension]! !

!ExternalDropHandler methodsFor: 'testing' stamp: ''!
matchesTypes: types
	(self type isNil or: [types isNil])
		ifTrue: [^false].
	^types anySatisfy: [:mimeType | mimeType beginsWith: self type]! !

!ExternalDropHandler methodsFor: 'accessing' stamp: ''!
type
	^type! !

!ExternalDropHandler methodsFor: 'initialize' stamp: ''!
type: aType extension: anExtension action: anAction 
	action := anAction.
	type := aType.
	extension := anExtension! !

Object subclass: #ExternalSemaphoreTable
	instanceVariableNames: ''
	classVariableNames: 'ProtectTable'
	poolDictionaries: ''
	category: 'System-Support'!

!ExternalSemaphoreTable commentStamp: '<historical>' prior: 0!
By John M McIntosh johnmci@smalltalkconsulting.com
This class was written to mange the external semaphore table. When I was writing a Socket test server I discovered various race conditions on the access to the externalSemaphore table. This new class uses class side methods to restrict access using a mutex semaphore. It seemed cleaner to deligate the reponsibility here versus adding more code and another class variable to SystemDictionary 

Note that in Smalltalk recreateSpecialObjectsArray we still directly play with the table.!

!ExternalSemaphoreTable class methodsFor: 'accessing' stamp: ''!
clearExternalObjects
	"Clear the array of objects that have been registered for use in non-Smalltalk code."

	ProtectTable critical: [Smalltalk specialObjectsArray at: 39 put: Array new].
! !

!ExternalSemaphoreTable class methodsFor: 'accessing' stamp: ''!
externalObjects
	^ProtectTable critical: [Smalltalk specialObjectsArray at: 39].! !

!ExternalSemaphoreTable class methodsFor: 'initialize' stamp: ''!
initialize
	ProtectTable := Semaphore forMutualExclusion! !

!ExternalSemaphoreTable class methodsFor: 'accessing' stamp: ''!
registerExternalObject: anObject
	^ ProtectTable critical: [self safelyRegisterExternalObject: anObject]
! !

!ExternalSemaphoreTable class methodsFor: 'accessing' stamp: ''!
safelyRegisterExternalObject: anObject
	"Register the given object in the external objects array and return its index. If it is already there, just return its index."

	| objects firstEmptyIndex obj sz newObjects |
	objects := Smalltalk specialObjectsArray at: 39.

	"find the first empty slot"
	firstEmptyIndex := 0.
	1 to: objects size do: [:i |
		obj := objects at: i.
		obj == anObject ifTrue: [^ i].  "object already there, just return its index"
		(obj == nil and: [firstEmptyIndex = 0]) ifTrue: [firstEmptyIndex := i]].

	"if no empty slots, expand the array"
	firstEmptyIndex = 0 ifTrue: [
		sz := objects size.
		newObjects := objects species new: sz + 20.  "grow linearly"
		newObjects replaceFrom: 1 to: sz with: objects startingAt: 1.
		firstEmptyIndex := sz + 1.
		Smalltalk specialObjectsArray at: 39 put: newObjects.
		objects := newObjects].

	objects at: firstEmptyIndex put: anObject.
	^ firstEmptyIndex
! !

!ExternalSemaphoreTable class methodsFor: 'accessing' stamp: ''!
safelyUnregisterExternalObject: anObject
	"Unregister the given object in the external objects array. Do nothing if it isn't registered.
	JMM change to return if we clear the element, since it should only appear once in the array"

	| objects |
	anObject ifNil: [^ self].
	objects := Smalltalk specialObjectsArray at: 39.
	1 to: objects size do: [:i |
		(objects at: i) == anObject ifTrue: 
		[objects at: i put: nil.
		^self]].
! !

!ExternalSemaphoreTable class methodsFor: 'accessing' stamp: ''!
unregisterExternalObject: anObject
	ProtectTable critical: [self safelyUnregisterExternalObject: anObject]
! !

Object subclass: #ExternalSettings
	instanceVariableNames: ''
	classVariableNames: 'RegisteredClients'
	poolDictionaries: ''
	category: 'System-Support'!

!ExternalSettings commentStamp: '<historical>' prior: 0!
ExternalSettings manages settings kept externally, e.g. files.
Objects can register themselves as clients to be notified at startup time to read their settings.

Eventually all the preferences should be managed through this mechanism.
!

!ExternalSettings class methodsFor: 'accessing' stamp: ''!
assuredPreferenceDirectory
	"Answer the preference directory, creating it if necessary"

	|  prefDir |
	prefDir := self preferenceDirectory.
	prefDir
		ifNil:
			[prefDir := FileDirectory default directoryNamed: self preferenceDirectoryName.
			prefDir assureExistence].
	^ prefDir! !

!ExternalSettings class methodsFor: 'initialization' stamp: ''!
initialize
	"ExternalSettings initialize"
	"Order: ExternalSettings, SecurityManager, AutoStart"
	Smalltalk addToStartUpList: self.
	Smalltalk addToShutDownList: self! !

!ExternalSettings class methodsFor: 'accessing' stamp: ''!
parseServerEntryArgsFrom: stream 
	"Args are in the form <argName>: <argValueString> delimited by end of line.
	It's not a very robust format and should be replaced by something like XML later.
	But it avoids evaluating the entries for security reasons."
	| entries lineStream entryName entryValue |
	entries := Dictionary new.
	stream skipSeparators.
	[ stream atEnd ] whileFalse: 
		[ lineStream := stream nextLine readStream.
		entryName := lineStream upTo: $:.
		lineStream skipSeparators.
		entryValue := lineStream upToEnd.
		(entryName isEmptyOrNil or: [ entryValue isEmptyOrNil ]) ifFalse: 
			[ entries 
				at: entryName
				put: entryValue withoutTrailingBlanks ].
		stream skipSeparators ].
	^ entries! !

!ExternalSettings class methodsFor: 'accessing' stamp: ''!
preferenceDirectory
	| prefDirName path |
	prefDirName := self preferenceDirectoryName.
	path := Smalltalk vmPath.
	^(FileDirectory default directoryExists: prefDirName)
		ifTrue: [FileDirectory default directoryNamed: prefDirName]
		ifFalse: [
			((FileDirectory on: path) directoryExists: prefDirName)
				ifTrue: [(FileDirectory on: path) directoryNamed: prefDirName]
				ifFalse: [nil]]
! !

!ExternalSettings class methodsFor: 'accessing' stamp: ''!
preferenceDirectoryName
	^'prefs'! !

!ExternalSettings class methodsFor: 'accessing' stamp: ''!
registerClient: anObject
	"Register anObject as a settings client to be notified on startup."

	self registeredClients add: anObject! !

!ExternalSettings class methodsFor: 'private' stamp: ''!
registeredClients
	RegisteredClients ifNil: [RegisteredClients := Set new].
	^RegisteredClients! !

!ExternalSettings class methodsFor: 'initialization' stamp: ''!
shutDown
	"Look for external defs and load them."
	"ExternalSettings shutDown"

	self registeredClients do: [:client | 
		client releaseExternalSettings]! !

!ExternalSettings class methodsFor: 'initialization' stamp: ''!
startUp
	"Look for external defs and load them."
	"ExternalSettings startUp"

	| prefDir |
	prefDir := self preferenceDirectory.
	prefDir
		ifNil: [^self].
	self registeredClients do: [:client | 
		client fetchExternalSettingsIn: prefDir]! !

Object subclass: #LRUCache
	instanceVariableNames: 'size factory calls hits values'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'System-Support'!

!LRUCache commentStamp: '<historical>' prior: 0!
I'm a cache of values, given a key I return a Value from the cache or from the factory!

!LRUCache class methodsFor: 'instance creation' stamp: ''!
size: aNumber factory: aBlock 
	"answer an instance of the receiver"
	^ self new initializeSize: aNumber factory: aBlock! !

!LRUCache class methodsFor: 'testing' stamp: ''!
test
	" 
	LRUCache test 
	"
	| c |
	c := LRUCache
				size: 5
				factory: [:key | key * 2].
	c at: 1.
	c at: 2.
	c at: 3.
	c at: 4.
	c at: 1.
	c at: 5.
	c at: 6.
	c at: 7.
	c at: 8.
	c at: 1.
	^ c! !

!LRUCache class methodsFor: 'testing' stamp: ''!
test2
	" 
	LRUCache test2.  
	Time millisecondsToRun:[LRUCache test2]. 
	MessageTally spyOn:[LRUCache test2].  
	"
	| c |
	c := LRUCache
				size: 600
				factory: [:key | key * 2].
	1
		to: 6000
		do: [:each | c at: each].
	^ c! !

!LRUCache methodsFor: 'accessing' stamp: ''!
at: aKey 
	"answer the object for aKey, if not present in the cache creates it.
	Clone the factory block before calling in case of multiple processes!!"
	
	| element keyHash |
	calls := calls + 1.
	keyHash := aKey hash.
	1
		to: size
		do: [:index | 
			element := values at: index.
			(keyHash
						= (element at: 2)
					and: [aKey
							= (element at: 1)])
				ifTrue: ["Found!!"
					hits := hits + 1.
					values
						replaceFrom: 2
						to: index
						with: (values first: index - 1).
					values at: 1 put: element.
					^ element at: 3]].
	"Not found!!"
	element := {aKey. keyHash. factory shallowCopy value: aKey}.
	values
		replaceFrom: 2
		to: size
		with: values allButLast.
	values at: 1 put: element.
	^ element at: 3! !

!LRUCache methodsFor: 'initialization' stamp: ''!
initializeSize: aNumber factory: aBlock 
	"initialize the receiver's size and factory"
	size := aNumber.
	values := Array new: aNumber withAll: {nil. nil. nil}.
	factory := aBlock.
	calls := 0.
	hits := 0! !

!LRUCache methodsFor: 'printing' stamp: ''!
printOn: aStream 
	"Append to the argument, aStream, a sequence of characters 
	that identifies the receiver."
	aStream nextPutAll: self class name;
		 nextPutAll: ' size:';
		 nextPutAll: size asString;
		 nextPutAll: ', calls:';
		 nextPutAll: calls asString;
		 nextPutAll: ', hits:';
		 nextPutAll: hits asString;
		 nextPutAll: ', ratio:'.
	calls = 0
		ifTrue: [aStream nextPutAll: 0 asFloat asString]
		ifFalse: [aStream nextPutAll: (hits / calls) asFloat asString]! !

!Object methodsFor: '*System-Support' stamp: ''!
systemNavigation

	^ SystemNavigation default! !

Object subclass: #SecurityManager
	instanceVariableNames: 'privateKeyPair trustedKeys keysFileName securityChecksEnabled showSecurityStatus startInUntrustedDirectory warnAboutInsecureContent automaticKeyGeneration'
	classVariableNames: 'Default'
	poolDictionaries: ''
	category: 'System-Support'!

!SecurityManager class methodsFor: 'accessing' stamp: ''!
default
	^Default ifNil: [Default := self new]! !

!SecurityManager class methodsFor: 'initialization' stamp: ''!
initialize
	"SecurityManager initialize"

	"Order: ExternalSettings, SecurityManager, AutoStart"

	Default := self new.
	Smalltalk addToStartUpList: self after: ExternalSettings.
	Smalltalk addToShutDownList: self! !

!SecurityManager class methodsFor: 'initialization' stamp: ''!
shutDown
	self default shutDown.! !

!SecurityManager class methodsFor: 'initialization' stamp: ''!
startUp
	self default startUp.! !

!SecurityManager methodsFor: 'accessing' stamp: ''!
addTrustedKey: aPublicKey
	"Add a public key to the list of trusted keys"
	trustedKeys := (trustedKeys copyWithout: aPublicKey) copyWith: aPublicKey.! !

!SecurityManager methodsFor: 'settings' stamp: ''!
automaticKeyGeneration
	^ automaticKeyGeneration ifNil: [automaticKeyGeneration := false]! !

!SecurityManager methodsFor: 'testing' stamp: ''!
canWriteImage
	"SecurityManager default canWriteImage"
	"Primitive. Return true if the right to write an image hasn't been revoked."
	<primitive: 'primitiveCanWriteImage' module: 'SecurityPlugin'>
	^true "assume so unless otherwise proven"! !

!SecurityManager methodsFor: 'security operations' stamp: ''!
disableFileAccess
	"SecurityManager default disableFileAccess"
	"Primitive. Disable unlimited access to files.
	Cannot be revoked from the image."
	<primitive: 'primitiveDisableFileAccess' module: 'FilePlugin'>
	^self primitiveFailed! !

!SecurityManager methodsFor: 'security operations' stamp: ''!
disableImageWrite
	"SecurityManager default disableImageWrite"
	"Primitive. Disable writing to an image file.
	Cannot be revoked from the image."
	<primitive: 'primitiveDisableImageWrite' module: 'SecurityPlugin'>
	^self primitiveFailed! !

!SecurityManager methodsFor: 'security operations' stamp: ''!
disableSocketAccess
	"SecurityManage default disableSocketAccess"
	"Primitive. Disable access to sockets.
	Cannot be revoked from the image."
	<primitive: 'primitiveDisableSocketAccess' module: 'SocketPlugin'>
	^self primitiveFailed! !

!SecurityManager methodsFor: 'security operations' stamp: ''!
enterRestrictedMode
	"Some insecure contents was encountered. Close all doors and proceed."
	self isInRestrictedMode ifTrue:[^true].
	self securityChecksEnabled ifFalse:[^true]. "it's been your choice..."
	self warnAboutInsecureContent ifTrue:[
		( UIManager default chooseFrom: #('Load it anyways' 'Do not load it')
			title: 
'You are about to load some insecure content.
If you continue, access to files as well as
some other capabilities will be limited.')
			 = 1 ifFalse:[
				"user doesn't really want it"
				^false.
			].
	].
	"here goes the actual restriction"
	self flushSecurityKeys.
	self disableFileAccess.
	self disableImageWrite.
	"self disableSocketAccess."
	FileDirectory setDefaultDirectory: self untrustedUserDirectory.
	^true
! !

!SecurityManager methodsFor: 'security operations' stamp: ''!
fileInObjectAndCode: aStream
	| trusted |
	trusted := self positionToSecureContentsOf: aStream.
	trusted ifFalse:[self enterRestrictedMode ifFalse:[
		aStream close.
		^nil]].
	^aStream fileInObjectAndCode! !

!SecurityManager methodsFor: 'initialization' stamp: ''!
flushSecurityKey: aKey
	"Flush a security key"
	| n |
	n := aKey first.
	1 to: n basicSize do:[:i| n basicAt: i put: 0].
	n := aKey second.
	1 to: n basicSize do:[:i| n basicAt: i put: 0].
! !

!SecurityManager methodsFor: 'initialization' stamp: ''!
flushSecurityKeys
	"Flush all keys"
	privateKeyPair ifNotNil:[
		self flushSecurityKey: privateKeyPair first.
		self flushSecurityKey: privateKeyPair last.
	].
	privateKeyPair := nil.
	trustedKeys do:[:key| self flushSecurityKey: key].
	trustedKeys := #().! !

!SecurityManager methodsFor: 'private' stamp: ''!
generateKeyPairInBackground
	"SecurityManager default generateKeyPairInBackground"
	"Silently generate a key set on the local machine while running in the background."
	| guesstimate startTime |
	guesstimate := [ 10 benchmark ] timeToRun * 150.
	startTime := Time millisecondClockValue.
	privateKeyPair := nil.
	[ self generateLocalKeyPair ] fork.
	UIManager default informUserDuring: 
		[ :bar | 
		[ privateKeyPair == nil ] whileTrue: 
			[ bar value: 'Initializing security system (' , ((Time millisecondClockValue - startTime) * 100 // guesstimate) printString , '%)'.
			(Delay forSeconds: 1) wait ] ]! !

!SecurityManager methodsFor: 'private' stamp: ''!
generateLocalKeyPair
	"SecurityManager default generateLocalKeyPair"
	"Generate a key set on the local machine."
	| dsa |
	dsa := DigitalSignatureAlgorithm new.
	dsa initRandomFromString: 
		Time millisecondClockValue printString, 
		Date today printString, 
		OSPlatform platformName printString.
	privateKeyPair := dsa generateKeySet.
	self storeSecurityKeys.! !

!SecurityManager methodsFor: 'testing' stamp: ''!
hasFileAccess
	"SecurityManager default hasFileAccess"
	"Return true if the right to access arbitrary files hasn't been revoked"
	<primitive: 'primitiveHasFileAccess' module: 'FilePlugin'>
	^true "assume so unless otherwise proven"! !

!SecurityManager methodsFor: 'testing' stamp: ''!
hasSocketAccess
	"SecurityManager default hasSocketAccess"
	"Return true if the right to access sockets hasn't been revoked"
	<primitive: 'primitiveHasSocketAccess' module: 'SocketPlugin'>
	^true "assume so unless otherwise proven"! !

!SecurityManager methodsFor: 'initialization' stamp: ''!
initialize
	super initialize.
	privateKeyPair := nil.
	trustedKeys := #().
	keysFileName := 'Pharo.keys'.! !

!SecurityManager methodsFor: 'testing' stamp: ''!
isInRestrictedMode
	"Return true if we're in restricted mode"
	^(self canWriteImage 
		or:[self hasFileAccess 
		"or:[self hasSocketAccess]"]) not! !

!SecurityManager methodsFor: 'accessing' stamp: ''!
keysFileName
	^keysFileName! !

!SecurityManager methodsFor: 'accessing' stamp: ''!
keysFileName: aFileName
	keysFileName := aFileName! !

!SecurityManager methodsFor: 'filein/out' stamp: ''!
loadSecurityKeys
	"SecurityManager default loadSecurityKeys"
	"Load the keys file for the current user"
	| fd loc file keys |
	self isInRestrictedMode ifTrue: [ ^ self ]. "no point in even trying"
	loc := self secureUserDirectory. "where to get it from"
	loc last = FileDirectory pathNameDelimiter ifFalse:[
		loc := loc copyWith: FileDirectory pathNameDelimiter.
	].

	fd := FileDirectory on: loc.
	file := [fd readOnlyFileNamed: keysFileName] 
				on: FileDoesNotExistException 
				do:[:ex| ^ self "no keys file"].
				
	keys := Object readFrom: file.
	privateKeyPair := keys first.
	trustedKeys := keys last.
	file close.! !

!SecurityManager methodsFor: 'security operations' stamp: ''!
positionToSecureContentsOf: aStream
	| bytes trusted part1 part2 sig hash dsa okay pos |
	aStream binary.
	pos := aStream position.
	bytes := aStream next: 4.
	bytes = 'SPRJ' asByteArray ifFalse:[
		"was not signed"
		aStream position: pos.
		^false].
	part1 := (aStream nextInto: (LargePositiveInteger basicNew: 20)) normalize.
	part2 := (aStream nextInto: (LargePositiveInteger basicNew: 20)) normalize.
	sig := Array with: part1 with: part2.
	hash := SecureHashAlgorithm new hashStream: aStream.
	dsa := DigitalSignatureAlgorithm new.
	trusted := self trustedKeys.
	okay := (trusted detect:[:key| dsa verifySignature: sig ofMessageHash: hash publicKey: key]
			ifNone:[nil]) notNil.
	aStream position: pos+44.
	^okay! !

!SecurityManager methodsFor: 'accessing' stamp: ''!
primSecureUserDirectory
	<primitive: 'primitiveGetSecureUserDirectory' module: 'SecurityPlugin'>
	^ nil! !

!SecurityManager methodsFor: 'private' stamp: ''!
primUntrustedUserDirectory
	"Primitive. Return the untrusted user directory that is the root directory for files that are visible even in restricted mode."
	<primitive: 'primitiveGetUntrustedUserDirectory' module: 'SecurityPlugin'>
	^ nil! !

!SecurityManager methodsFor: 'private' stamp: ''!
printStateOn: stream
 	"Print the current state of myself onto stream.
 	Used to gather information in the debug log."
 
 	stream
 		nextPutAll: 'SecurityManager state:'; cr;
 		nextPutAll: 'Restricted: '; nextPutAll: self isInRestrictedMode asString; cr;
 		nextPutAll: 'FileAccess: '; nextPutAll: self hasFileAccess asString; cr;
 		nextPutAll: 'SocketAccess: '; nextPutAll: self hasSocketAccess asString; cr;
 		nextPutAll: 'Working Dir '; nextPutAll: FileDirectory default pathName asString; cr;
 		nextPutAll: 'Trusted Dir '; nextPutAll: self secureUserDirectory asString; cr;
 		nextPutAll: 'Untrusted Dir '; nextPutAll: self untrustedUserDirectory asString; cr;
 		cr! !

!SecurityManager methodsFor: 'accessing' stamp: ''!
secureUserDirectory
	"SecurityManager default secureUserDirectory"
	| dir |
	dir := self primSecureUserDirectory.
	^ dir
		ifNil: [FileDirectory default pathName]
		ifNotNil: [(FilePath pathName: dir isEncoded: true) asSqueakPathName]! !

!SecurityManager methodsFor: 'settings' stamp: ''!
securityChecksEnabled
	^ securityChecksEnabled ifNil: [securityChecksEnabled := false]! !

!SecurityManager methodsFor: 'settings' stamp: ''!
securityChecksEnabled: aBoolean
	securityChecksEnabled := aBoolean! !

!SecurityManager methodsFor: 'settings' stamp: ''!
showSecurityStatus
	^ showSecurityStatus ifNil: [showSecurityStatus := false]! !

!SecurityManager methodsFor: 'settings' stamp: ''!
showSecurityStatus: aBoolean
	showSecurityStatus := aBoolean! !

!SecurityManager methodsFor: 'initialization' stamp: ''!
shutDown
	"Flush existing keys"
	self flushSecurityKeys.! !

!SecurityManager methodsFor: 'accessing' stamp: ''!
signingKey
	"Return the key used for signing projects"
	^privateKeyPair ifNotNil:[privateKeyPair first]! !

!SecurityManager methodsFor: 'settings' stamp: ''!
startInUntrustedDirectory
	^ startInUntrustedDirectory ifNil: [startInUntrustedDirectory := false]! !

!SecurityManager methodsFor: 'settings' stamp: ''!
startInUntrustedDirectory: aBoolean
	startInUntrustedDirectory := aBoolean! !

!SecurityManager methodsFor: 'initialization' stamp: ''!
startUp
	"Attempt to load existing keys"
	self loadSecurityKeys.
	(privateKeyPair == nil 
		and:[self isInRestrictedMode not
		and:[self automaticKeyGeneration]]) ifTrue:[
			self generateKeyPairInBackground.
	].! !

!SecurityManager methodsFor: 'filein/out' stamp: ''!
storeSecurityKeys
	"Store the keys file for the current user"
	"SecurityManager default storeSecurityKeys"

	| fd loc file |
	self isInRestrictedMode ifTrue:[^self]. "no point in even trying"
	loc := self secureUserDirectory. "where to put it"
	loc last = FileDirectory pathNameDelimiter ifFalse:
		[loc := loc copyWith: FileDirectory pathNameDelimiter].
	fd := FileDirectory on: loc.
	fd assureExistence.
	fd deleteFileNamed: self keysFileName ifAbsent:[].
	file := fd newFileNamed: self keysFileName.
	{privateKeyPair. trustedKeys} storeOn: file.
	file close! !

!SecurityManager methodsFor: 'accessing' stamp: ''!
trustedKeys
	"Return an array of trusted public keys for verifying some project"
	privateKeyPair ifNil:[^trustedKeys].
	^{privateKeyPair second}, trustedKeys! !

!SecurityManager methodsFor: 'accessing' stamp: ''!
untrustedUserDirectory
	"SecurityManager default untrustedUserDirectory"
	| dir |
	dir := self primUntrustedUserDirectory.
	^ dir
		ifNil: [FileDirectory default pathName]
		ifNotNil: [(FilePath pathName: dir isEncoded: true) asSqueakPathName]! !

!SecurityManager methodsFor: 'settings' stamp: ''!
warnAboutInsecureContent
	^ warnAboutInsecureContent ifNil: [warnAboutInsecureContent := false]! !

!SecurityManager methodsFor: 'settings' stamp: ''!
warnAboutInsecureContent: aBoolean
	warnAboutInsecureContent := aBoolean! !

Object subclass: #SmalltalkImage
	instanceVariableNames: 'globals'
	classVariableNames: 'CheckChangesFileAvailability CheckSourcesFileAvailability EndianCache LastImageName LastQuitLogPosition LastStats LowSpaceProcess LowSpaceSemaphore MemoryHogs ShutDownList SourceFileVersionString SpecialSelectors StartUpList StartupStamp WordSize'
	poolDictionaries: ''
	category: 'System-Support'!

!SmalltalkImage commentStamp: 'Jb 10/23/2010 11:23' prior: 0!
not a good name but i work on it.!

!SmalltalkImage class methodsFor: 'settings' stamp: ''!
checkChangesFileAvailability
	^ CheckChangesFileAvailability ifNil: [CheckChangesFileAvailability := true]! !

!SmalltalkImage class methodsFor: 'settings' stamp: ''!
checkChangesFileAvailability: aBoolean
	CheckChangesFileAvailability := aBoolean! !

!SmalltalkImage class methodsFor: 'settings' stamp: ''!
checkSourcesFileAvailability
	^ CheckSourcesFileAvailability ifNil: [CheckSourcesFileAvailability := true]! !

!SmalltalkImage class methodsFor: 'settings' stamp: ''!
checkSourcesFileAvailability: aBoolean
	CheckSourcesFileAvailability := aBoolean! !

!SmalltalkImage class methodsFor: 'cleanup' stamp: 'AlainPlantec 12/1/2010 09:53'!
cleanUp
	"Flush caches"
	
	"should probably distribute cleaning to systemDictionary"
	
	Smalltalk globals flushClassNameCache.
	Undeclared removeUnreferencedKeys! !

!SmalltalkImage class methodsFor: 'instance creation' stamp: ''!
current
	^Smalltalk! !

!SmalltalkImage class methodsFor: 'initialization' stamp: ''!
howToGetClassesNotRegisteredInStartups

	| classes startClassNames stopClassNames |
	classes := ((SystemNavigation default allCallsOn:
#'addToStartUpList:') collect: [ :e | e methodClass ]),
((SystemNavigation default allCallsOn: #'addToStartUpList:after:')
collect: [ :e | e methodClass ]).
startClassNames := classes collect: [ :n | n instanceSide name ].
(Smalltalk class classVarNamed: 'StartUpList') do: [ :s |
	startClassNames remove: s ifAbsent: [] ].


classes := ((SystemNavigation default allCallsOn:
#'addToShutDownList:') collect: [ :e | e methodClass ]),
((SystemNavigation default allCallsOn: #'addToShutDownList:after:')
collect: [ :e | e methodClass ]).
stopClassNames := classes collect: [ :n | n instanceSide name ].
(Smalltalk class classVarNamed: 'ShutDownList') do: [ :s |
	stopClassNames remove: s ifAbsent: [] ].

((startClassNames, stopClassNames) asSet asSortedCollection remove: #InputEventSensor ; yourself)
	"do: [:each | (Smalltalk at: each) initialize] does not work because of dependencies"! !

!SmalltalkImage class methodsFor: 'initialization' stamp: ''!
initialize
	"SmalltalkImage initialize"

	self initializeStartUpList.
	self initializeShutDownList.
	Smalltalk addToStartUpList: SmalltalkImage! !

!SmalltalkImage class methodsFor: 'initialization' stamp: ''!
initializeShutDownList
	"SmalltalkImage initializeShutDownList"

	| oldList |
	oldList := ShutDownList.
	ShutDownList := OrderedCollection new.	"These get processed from the bottom up..."
	#(#Delay #DisplayScreen #InputEventFetcher #Form #PasteUpMorph #StrikeFont #Color #FileDirectory #SoundPlayer #HttpUrl #Password)
		do: [ :clsName | Smalltalk globals at: clsName ifPresent: [ :cls | Smalltalk addToShutDownList: cls ] ].
	oldList
		ifNotNil: [ 
			oldList
				reverseDo: [ :className | Smalltalk globals at: className ifPresent: [ :theClass | Smalltalk addToShutDownList: theClass ] ] ]! !

!SmalltalkImage class methodsFor: 'initialization' stamp: ''!
initializeStartUpList
	"SmalltalkImage initializeStartUpList"

	| oldList |
	oldList := StartUpList.
	StartUpList := OrderedCollection new.	"These get processed from the top down..."
	#(#Delay #DisplayScreen #Cursor #InputEventFetcher #ProcessorScheduler #LanguageEnvironment #FileDirectory #NaturalLanguageTranslator #ShortIntegerArray #ShortRunArray #CrLfFileStream)
		do: [ :clsName | Smalltalk globals at: clsName ifPresent: [ :cls | Smalltalk addToStartUpList: cls ] ].
	oldList
		ifNotNil: [ 
			oldList
				do: [ :className | Smalltalk globals at: className ifPresent: [ :theClass | Smalltalk addToStartUpList: theClass ] ] ].
	#(#PasteUpMorph) do: [ :clsName | Smalltalk globals at: clsName ifPresent: [ :cls | Smalltalk addToStartUpList: cls ] ]! !

!SmalltalkImage class methodsFor: 'instance creation' stamp: ''!
new

	self error: 'Use current'.! !

!SmalltalkImage class methodsFor: 'initialization' stamp: ''!
startUp
	SystemChangeNotifier uniqueInstance
		notify: Smalltalk
		ofAllSystemChangesUsing: #event:! !

!SmalltalkImage class methodsFor: 'initialization' stamp: ''!
temporaryFixStartUpList
	"SmalltalkImage temporaryFixStartUpList"

	#(#CPUWatcher #Clipboard #CommandHistory #ExternalSettings #FT2Handle #FreeTypeCache #SecurityManager #FreeTypeFontProvider #FreeTypeSettings #SmalltalkImage #HostSystemMenus #HostWindowProxy #InternetConfiguration #Locale #LogicalFont #MCMethodDefinition #MultiByteFileStream #OSPlatform #ProcessBrowser #AutoStart #Symbol #SystemDictionary #TTCFont #UITheme #UUIDGenerator #WeakArray)
		do: [ :each | (Smalltalk globals at: each) initialize ]! !

!SmalltalkImage class methodsFor: 'initialize-release' stamp: ''!
wordSize
	^ WordSize ifNil: [WordSize := [Smalltalk vmParameterAt: 40] on: Error do: [4]]! !

!SmalltalkImage methodsFor: 'shrinking' stamp: ''!
abandonSources
	"Smalltalk abandonSources"
	"Replaces every method by a copy with the 4-byte source pointer 
	 replaced by a string of all arg and temp names, followed by its
	 length. These names can then be used to inform the decompiler."
	"wod 11/3/1998: zap the organization before rather than after
	 condensing changes."
	"eem 7/1/2009 13:59 update for the closure schematic temp names regime"
	| oldMethods newMethods bTotal bCount |
	(self confirm: 'This method will preserve most temp names
(up to about 15k characters of temporaries)
while allowing the sources file to be discarded.
-- CAUTION --
If you have backed up your system and
are prepared to face the consequences of
abandoning source code files, choose Yes.
If you have any doubts, you may choose No
to back out with no harm done.')
			== true
		ifFalse: [^ self inform: 'Okay - no harm done'].
	oldMethods := OrderedCollection new: CompiledMethod instanceCount.
	newMethods := OrderedCollection new: CompiledMethod instanceCount.
	bTotal := 0.
	bCount := 0.
	self systemNavigation allBehaviorsDo: [:b | bTotal := bTotal + 1].
	'Saving temp names for better decompilation...'
		displayProgressAt: Sensor cursorPoint
		from: 0
		to: bTotal
		during:
			[:bar |
			self systemNavigation allBehaviorsDo:
				[:cl |  "for test: (Array with: Arc with: Arc class) do:"
				bar value: (bCount := bCount + 1).
				cl selectorsAndMethodsDo:
					[:selector :m |
					| oldCodeString methodNode |
					m fileIndex > 0 ifTrue:
						[oldCodeString := cl sourceCodeAt: selector.
						methodNode := cl compilerClass new
											parse: oldCodeString
											in: cl
											notifying: nil.
						oldMethods addLast: m.
						newMethods addLast: (m copyWithTempsFromMethodNode: methodNode)]]]].
	oldMethods asArray elementsExchangeIdentityWith: newMethods asArray.
	self systemNavigation allBehaviorsDo: [:b | b zapOrganization].
	self condenseChanges.
	SmalltalkImage checkSourcesFileAvailability: false! !

!SmalltalkImage methodsFor: 'shrinking' stamp: ''!
abandonTempNames
	"Replaces every method by a copy with no source pointer or
	encoded temp names."
	"Smalltalk abandonTempNames"
	| continue oldMethods newMethods n |
	continue := self confirm: '-- CAUTION --
If you have backed up your system and
are prepared to face the consequences of
abandoning all source code, hit Yes.
If you have any doubts, hit No,
to back out with no harm done.'.
	continue
		ifFalse: [^ self inform: 'Okay - no harm done'].
	self garbageCollect.
	oldMethods := OrderedCollection new.
	newMethods := OrderedCollection new.
	n := 0.
	'Removing temp names to save space...'
		displayProgressAt: Sensor cursorPoint
		from: 0
		to: CompiledMethod instanceCount
		during: [:bar | self systemNavigation
				allBehaviorsDo: [:cl | cl selectorsAndMethodsDo: [:sel :m | 
							bar value: (n := n + 1).
							oldMethods addLast: m.
							newMethods
								addLast: (m copyWithTrailerBytes: #(0 ))]]].
	oldMethods asArray elementsExchangeIdentityWith: newMethods asArray.
	Smalltalk closeSourceFiles.
	self flag: #shouldUseAEnsureBlockToBeSureThatTheFileIsClosed.
	"sd: 17 April 2003"
	SmalltalkImage checkChangesFileAvailability: false.
	SmalltalkImage checkSourcesFileAvailability: false! !

!SmalltalkImage methodsFor: 'sources, change log' stamp: 'AlainPlantec 10/18/2010 19:26'!
aboutThisSystem
	"Identify software version"
	| text dialog width |
	text := Smalltalk systemInformationString withCRs.
	width := 0.
	text
		linesDo: [:l | width := width 
						max: (UITheme current textFont widthOfStringOrText: l)].
	dialog := LongMessageDialogWindow new entryText: text.
	dialog iconMorph image: ThemeIcons pharoIcon.
	dialog title: 'About Pharo'.
	dialog open.
	dialog textMorph minWidth: 20;
		 minHeight: 20.
	dialog
		width: (width + 120 min: Display width - 50).
	dialog position: 25 @ 25! !

!SmalltalkImage methodsFor: 'image, changes name' stamp: ''!
add: aClass toList: startUpOrShutDownList after: predecessor
	"Add the name of aClass to the startUp or shutDown list.
	Add it after the name of predecessor, or at the end if predecessor is nil."

	| name earlierName |
	name := aClass name.
	(self at: name ifAbsent: [nil]) == aClass ifFalse:
		[self error: name , ' cannot be found in Smalltalk dictionary.'].
	predecessor == nil
		ifTrue: ["No-op if alredy in the list."
				(startUpOrShutDownList includes: name) ifFalse:
					[startUpOrShutDownList == StartUpList
						ifTrue: ["Add to end of startUp list"
								startUpOrShutDownList addLast: name]
						ifFalse: ["Add to front of shutDown list"
								startUpOrShutDownList addFirst: name]]]
		ifFalse: ["Add after predecessor, moving it if already there."
				earlierName := predecessor name.
				(self at: earlierName) == predecessor ifFalse:
					[self error: earlierName , ' cannot be found in Smalltalk dictionary.'].
				(startUpOrShutDownList includes: earlierName) ifFalse:
					[self error: earlierName , ' cannot be found in the list.'].
				startUpOrShutDownList remove: name ifAbsent:[].
				startUpOrShutDownList add: name after: earlierName]! !

!SmalltalkImage methodsFor: 'snapshot and quit' stamp: ''!
addToShutDownList: aClass
	"This will add a ref to this class at the BEGINNING of the shutDown list."

	self addToShutDownList: aClass after: nil! !

!SmalltalkImage methodsFor: 'snapshot and quit' stamp: ''!
addToShutDownList: aClass after: predecessor

	self add: aClass toList: ShutDownList after: predecessor! !

!SmalltalkImage methodsFor: 'snapshot and quit' stamp: ''!
addToStartUpList: aClass
	"This will add a ref to this class at the END of the startUp list."

	self addToStartUpList: aClass after: nil! !

!SmalltalkImage methodsFor: 'snapshot and quit' stamp: ''!
addToStartUpList: aClass after: predecessor

	self add: aClass toList: StartUpList after: predecessor! !

!SmalltalkImage methodsFor: 'to clean later' stamp: ''!
allClasses  
	"Return all the class defines in the Smalltalk SmalltalkImage"
	"Smalltalk allClasses"
	"May be we could discard this one in the future"
	^ self classNames collect: [:name | self at: name]! !

!SmalltalkImage methodsFor: 'class and trait names' stamp: ''!
allClassesAndTraits
	"Return all the classes and traits defined in the Smalltalk SmalltalkImage"

	^ self classNames , self traitNames collect: [:each | self at: each]! !

!SmalltalkImage methodsFor: 'class and trait names' stamp: ''!
allClassesAndTraitsDo: aBlock
	^self allClassesAndTraits do: aBlock! !

!SmalltalkImage methodsFor: 'to clean later' stamp: ''!
allClassesDo: aBlock
	"Evaluate the argument, aBlock, for each class in the system."

	(self classNames collect: [:name | self at: name]) do: aBlock! !

!SmalltalkImage methodsFor: 'to clean later' stamp: ''!
allTraits
	"Return all traits defined in the Smalltalk SmalltalkImage"

	^ self traitNames collect: [:each | self at: each]! !

!SmalltalkImage methodsFor: 'system attribute' stamp: ''!
argumentAt: i
	"Answer the i-th argument of the command line, or nil if not so many argument."
	
	^self getSystemAttribute: 2 + i! !

!SmalltalkImage methodsFor: 'system attribute' stamp: ''!
arguments
	"Answer an array with all the command line arguments.
	This does not include imagePath, documentPath nor any option."
	
	"Smalltalk commandLine arguments"
	
	^Array streamContents: [:str |
		| arg i |
		i := 1.
		[i > 998 or: [(arg := self argumentAt: i) == nil]]
			whileFalse:
				[str nextPut: arg.
				i := i + 1]].! !

!SmalltalkImage methodsFor: 'to clean later' stamp: ''!
associationAt: aKey
	"DO NOT DEPRECATE - used by ImageSegments"

	^globals associationAt: aKey! !

!SmalltalkImage methodsFor: 'to clean later' stamp: ''!
associationAt: key ifAbsent: aBlock
	"Obsoleted."

	self flag: #tobeprecated.
	"self deprecated: 'Use Smalltalk globals'."
	^globals associationAt: key ifAbsent: aBlock! !

!SmalltalkImage methodsFor: 'to clean later' stamp: ''!
associationDeclareAt: aKey

	^ globals associationDeclareAt: aKey! !

!SmalltalkImage methodsFor: 'to clean later' stamp: ''!
associationOrUndeclaredAt: key 
	"return an association or install in undeclared.  Used for mating up ImageSegments."

	^ globals associationOrUndeclaredAt: key! !

!SmalltalkImage methodsFor: 'sources, change log' stamp: ''!
assureStartupStampLogged
	"If there is a startup stamp not yet actually logged to disk, do it now."
	| changesFile |
	StartupStamp ifNil: [^ self].
	(SourceFiles isNil or: [(changesFile := SourceFiles at: 2) == nil]) ifTrue: [^ self].
	changesFile isReadOnly ifTrue:[^self].
	changesFile setToEnd; cr; cr.
	changesFile nextChunkPut: StartupStamp asString; cr.
	StartupStamp := nil.
	self forceChangesToDisk.! !

!SmalltalkImage methodsFor: 'to clean later' stamp: ''!
at: aKey
	"Answer the global associated with the key."

	^globals at: aKey! !

!SmalltalkImage methodsFor: 'to clean later' stamp: ''!
at: key ifAbsent: aBlock
	"Answer the global associated with the key or, if key isn't found,
	answer the result of evaluating aBlock."

	^globals at: key ifAbsent: aBlock! !

!SmalltalkImage methodsFor: 'to clean later' stamp: ''!
at: aKey ifAbsentPut: aBlock 
	"Return the global at the given key. 
	If key is not included in the receiver store the result 
	of evaluating aBlock as new value."

	^globals at: aKey ifAbsentPut: aBlock! !

!SmalltalkImage methodsFor: 'to clean later' stamp: ''!
at: key ifPresent: aBlock
	"Lookup the given key in the globals. If it is present, answer the value of evaluating the given block with the value associated with the key. Otherwise, answer nil."

	^globals at: key ifPresent: aBlock! !

!SmalltalkImage methodsFor: 'to clean later' stamp: ''!
at: aKey put: anObject 
	"Set the global at key to be anObject.  If key is not found, create a
	new entry for key and set is value to anObject. Answer anObject."

	^globals at: aKey put: anObject! !

!SmalltalkImage methodsFor: 'to clean later' stamp: ''!
bindingOf: varName
	"Answer the binding of some variable resolved in the scope of the receiver"

	^globals bindingOf: varName! !

!SmalltalkImage methodsFor: 'vm' stamp: ''!
buildDate			
	"Return a String reflecting the build date of the VM"
	"Smalltalk buildDate"

	^self getSystemAttribute: 1006! !

!SmalltalkImage methodsFor: 'memory space' stamp: ''!
bytesLeft
	"Answer the number of bytes of space available. Does a full garbage collection."

	^ self garbageCollect
! !

!SmalltalkImage methodsFor: 'memory space' stamp: ''!
bytesLeft: aBool
	"Return the amount of available space. If aBool is true, include possibly available swap space. If aBool is false, include possibly available physical memory. For a report on the largest free block currently availabe within Pharo memory but not counting extra memory use #primBytesLeft."
	<primitive: 112>
	^self primBytesLeft! !

!SmalltalkImage methodsFor: 'memory space' stamp: ''!
bytesLeftString
	"Return a string describing the amount of memory available"
	| availInternal availPhysical availTotal |
	self garbageCollect.
	availInternal := self primBytesLeft.
	availPhysical := self bytesLeft: false.
	availTotal := self bytesLeft: true.
	(availTotal > (availInternal + 10000)) "compensate for mini allocations inbetween"
		ifFalse:[^availInternal asStringWithCommas, ' bytes available'].
	^String streamContents:[:s|
		s nextPutAll: availInternal asStringWithCommas, 	' bytes (internal) '; cr.
		s nextPutAll: availPhysical asStringWithCommas,	' bytes (physical) '; cr.
		s nextPutAll: availTotal asStringWithCommas, 	' bytes (total)     '].! !

!SmalltalkImage methodsFor: 'endian' stamp: ''!
calcEndianness
	| bytes word blt |
	"What endian-ness is the current hardware?  The String '1234' will be stored into a machine word.  On BigEndian machines (the Mac), $1 will be the high byte if the word.  On LittleEndian machines (the PC), $4 will be the high byte."
	"SmalltalkImage current endianness"

	bytes := #[0 0 0 0].  "(1 2 3 4) or (4 3 2 1)"
	word := WordArray with: 16r01020304.
	blt := (BitBlt toForm: (Form new hackBits: bytes)) 
				sourceForm: (Form new hackBits: word).
	blt combinationRule: Form over.  "store"
	blt sourceY: 0; destY: 0; height: 1; width: 4.
	blt sourceX: 0; destX: 0.
	blt copyBits.  "paste the word into the bytes"
	bytes first = 1 ifTrue: [^ #big].
	bytes first = 4 ifTrue: [^ #little].
	self error: 'Ted is confused'.! !

!SmalltalkImage methodsFor: 'image, changes names' stamp: ''!
changeImageNameTo: aString
	
	self imageName: aString asSqueakPathName.
	LastImageName := self imageName! !

!SmalltalkImage methodsFor: 'image, changes names' stamp: ''!
changesName
	"Answer the name for the changes file corresponding to the image file name."
	"Smalltalk changesName"

	| imName |
	imName := FileDirectory baseNameFor: self imageName.
	^ imName, FileDirectory dot, 'changes'! !

!SmalltalkImage methodsFor: 'sources, change log' stamp: ''!
checkAndOpenSourcesAndChanges
	"Open the changes and sources files and install them in SourceFiles. 
	Inform the user of problems regarding write permissions or CR/CRLF mixups."
	"Note: SourcesName and imageName are full paths; 
	changesName is a  local name."
	| sources changes msg wmsg |
	msg := 'Pharo cannot locate &fileRef.

Please check that the file is named properly and is in the
same directory as this image.'.
	wmsg := 'Pharo cannot write to &fileRef.

Please check that you have write permission for this file.

You won''t be able to save this image correctly until you fix this.'.

	sources := FileDirectory openSources: self sourcesName forImage: LastImageName.
	changes := FileDirectory openChanges: self changesName forImage: LastImageName.

	((sources == nil or: [sources atEnd])
			and: [self checkSourcesFileAvailability])
		ifTrue: [OSPlatform isMacOS
				ifTrue: [msg := msg , '
Make sure the sources file is not an Alias.'].
self inform: (msg copyReplaceAll: '&fileRef' with: 'the sources file named ' , self sourcesName)].

	(changes == nil
			and: [self checkChangesFileAvailability])
		ifTrue: [self inform: (msg copyReplaceAll: '&fileRef' with: 'the changes file named ' , self sourcesName)].

	(self checkChangesFileAvailability and: [changes notNil])
		ifTrue: [changes isReadOnly
				ifTrue: [self inform: (wmsg copyReplaceAll: '&fileRef' with: 'the changes file named ' , self changesName)].

			((changes next: 200)
					includesSubString: String crlf)
				ifTrue: [self inform: 'The changes file named ' , self changesName , '
has been injured by an unpacking utility.  Crs were changed to CrLfs.
Please set the preferences in your decompressing program to 
"do not convert text files" and unpack the system again.']].

	SourceFiles := Array with: sources with: changes! !

!SmalltalkImage methodsFor: 'settings' stamp: ''!
checkChangesFileAvailability
	^ self class checkChangesFileAvailability! !

!SmalltalkImage methodsFor: 'settings' stamp: ''!
checkSourcesFileAvailability
	^ self class checkSourcesFileAvailability! !

!SmalltalkImage methodsFor: 'class and trait names' stamp: ''!
classNamed: className 
	^self globals classOrTraitNamed: className.! !

!SmalltalkImage methodsFor: 'class and trait names' stamp: ''!
classNames
	^globals classNames! !

!SmalltalkImage methodsFor: 'class and trait names' stamp: ''!
classOrTraitNamed: aString
	"aString is either a class or trait name or a class or trait name followed by ' class' or 'classTrait' respectively.
	Answer the class or metaclass it names."

	| meta baseName baseClass |
	(aString endsWith: ' class')
		ifTrue: [ 
			meta := true.
			baseName := aString copyFrom: 1 to: aString size - 6 ]
		ifFalse: [ 
			(aString endsWith: ' classTrait')
				ifTrue: [ 
					meta := true.
					baseName := aString copyFrom: 1 to: aString size - 11 ]
				ifFalse: [ 
					meta := false.
					baseName := aString ] ].
	baseClass := Smalltalk globals at: baseName asSymbol ifAbsent: [ ^ nil ].
	meta
		ifTrue: [ ^ baseClass classSide ]
		ifFalse: [ ^ baseClass ]! !

!SmalltalkImage methodsFor: 'housekeeping' stamp: ''!
cleanOutUndeclared 
	Undeclared removeUnreferencedKeys! !

!SmalltalkImage methodsFor: 'cleanup' stamp: ''!
cleanUp	
	"Smalltalk cleanUp"
	"Gently clean up"
	
	^self cleanUp: false! !

!SmalltalkImage methodsFor: 'cleanup' stamp: ''!
cleanUp: aggressive
	"Clean up. When aggressive is true, this will destroy change sets, etc."
	"Smalltalk cleanUp: false"
	"Smalltalk cleanUp: true"

	^self cleanUp: aggressive except: #()! !

!SmalltalkImage methodsFor: 'cleanup' stamp: ''!
cleanUp: aggressive except: exclusions
	"Clean up. When aggressive is true, this will destroy resources, change sets, etc.
	Leave out any classes specifically listed in exclusions."

	"Smalltalk cleanUp: true except: {Point . ChangeSet}"

	^self cleanUp: aggressive except: exclusions confirming: aggressive! !

!SmalltalkImage methodsFor: 'cleanup' stamp: ''!
cleanUp: aggressive except: exclusions confirming: aBool
	"Clean up. When aggressive is true, this will destroy resources, change sets, etc.
	Leave out any classes specifically listed in exclusions."

	"Smalltalk cleanUp: true except: {Point . ChangeSet} 
		- will ask for confirmation"
	"SmalltalkImage current cleanUp: true except: {} confirming: false 
		- will not ask for confirmation"

	| classes |
	aBool ifTrue:[
		"Give the user a chance to bail"
		(self confirm: 'Cleanup will destroy projects, change sets and more.
Are you sure you want to proceed?') ifFalse: [^self]].

	"Find all classes implementing #cleanUp or cleanUp:"
	classes := Smalltalk allClasses select: [:aClass| 
		(aClass class includesSelector: #cleanUp) 
			or: [aClass class includesSelector: #cleanUp:]].

	"Leave out the classes in the exclusion set"
	classes := classes reject: [:aClass| exclusions includes: aClass].

	"Arrange classes in superclass order, superclasses before subclasses.
	This will ensure that specific cleanup (like MethodDictionary compaction)
	will run after generic superclass cleanup (HashedCollection rehashing).
	Otherwise generic superclass cleanup might undo specific one (in this
	case rehashing will undo a good bit of MD compaction)."
	classes := Class superclassOrder: classes.

	"Run the cleanup code"
	classes 
		do:[:aClass| aClass cleanUp: aggressive]
		displayingProgress: [:aClass| 'Cleaning up in ', aClass name].! !

!SmalltalkImage methodsFor: 'special objects' stamp: ''!
clearExternalObjects
	"Clear the array of objects that have been registered for use in non-Smalltalk code."
	"Smalltalk clearExternalObjects"

	ExternalSemaphoreTable clearExternalObjects
! !

!SmalltalkImage methodsFor: 'sources, change log' stamp: ''!
closeSourceFiles
	"Shut down the source files if appropriate.  1/29/96 sw: changed so that the closing and nilification only take place if the entry was a FileStream, thus allowing stringified sources to remain in the saved image file"

	1 to: 2 do: [:i |
		((SourceFiles at: i) isKindOf: FileStream)
			ifTrue:
				[(SourceFiles at: i) close.
				SourceFiles at: i put: nil]]! !

!SmalltalkImage methodsFor: 'system attribute' stamp: ''!
commandLine
	"Answer the object to query about command line."
	
	^self! !

!SmalltalkImage methodsFor: 'special objects' stamp: ''!
compactClassesArray
	"Smalltalk compactClassesArray"
	"Return the array of 31 classes whose instances may be
	represented compactly"
	^ self specialObjectsArray at: 29! !

!SmalltalkImage methodsFor: 'housekeeping' stamp: ''!
condenseChanges
	"Move all the changes onto a compacted sources file."
	"Smalltalk condenseChanges"
	| f oldChanges |
	f := FileStream fileNamed: 'ST80.temp'.
	f header; timeStamp.
	'Condensing Changes File...'
		displayProgressAt: Sensor cursorPoint
		from: 0
		to: self classNames size + self traitNames size
		during: [:bar | | count | 
			count := 0.
			self
				allClassesAndTraitsDo: [:classOrTrait | 
					bar value: (count := count + 1).
					classOrTrait moveChangesTo: f.
					classOrTrait putClassCommentToCondensedChangesFile: f.
					classOrTrait classSide moveChangesTo: f]].
	self lastQuitLogPosition: f position.
	f trailer; close.
	CompiledMethod cleanUp.
	
	oldChanges := SourceFiles at: 2.
	oldChanges close.
	FileDirectory default deleteFileNamed: oldChanges name , '.old';
		 rename: oldChanges name toBe: oldChanges name , '.old';
		 rename: f name toBe: oldChanges name.
	self setMacFileInfoOn: oldChanges name.
	SourceFiles
		at: 2
		put: (FileStream oldFileNamed: oldChanges name)! !

!SmalltalkImage methodsFor: 'housekeeping' stamp: ''!
condenseSources	
	"Move all the changes onto a compacted sources file."
	"Smalltalk condenseSources"

	| f dir newVersionString |
	self fixUpProblemsWithAllCategory.
	"The above removes any concrete, spurious '-- all --' categories, which mess up the process."
	dir := FileDirectory default.
	newVersionString := UIManager default request: 'Please designate the version
for the new source code file...' initialAnswer: Smalltalk  sourceFileVersionString.
	newVersionString ifNil: [^ self].
	newVersionString = self sourceFileVersionString ifTrue:
		[^ self error: 'The new source file must not be the same as the old.'].
	Smalltalk sourceFileVersionString: newVersionString.

	"Write all sources with fileIndex 1"
	f := FileStream newFileNamed: Smalltalk sourcesName.
	f header; timeStamp.
'Condensing Sources File...'
	displayProgressAt: Sensor cursorPoint
	from: 0 to: self classNames size + self traitNames size
	during:
		[:bar | | count | count := 0.
		Smalltalk allClassesAndTraitsDo:
			[:classOrTrait | bar value: (count := count + 1).
			classOrTrait fileOutOn: f moveSource: true toFile: 1]].
	f trailer; close.
	
	CompiledMethod cleanUp.
	
	"Make a new empty changes file"
	self closeSourceFiles.
	dir rename: Smalltalk changesName
		toBe: Smalltalk changesName , '.old'.
	(FileStream newFileNamed: Smalltalk changesName)
		header; timeStamp; close.
	self lastQuitLogPosition: 0.

	self setMacFileInfoOn: SmalltalkImage current changesName.
	self setMacFileInfoOn: SmalltalkImage current sourcesName.
	self openSourceFiles.
	self inform: 'Source files have been rewritten!!
Check that all is well,
and then save/quit.'! !

!SmalltalkImage methodsFor: 'memory space' stamp: ''!
createStackOverflow
	"For testing the low space handler..."
	"Smalltalk installLowSpaceWatcher; createStackOverflow"

	self createStackOverflow.  "infinite recursion"! !

!SmalltalkImage methodsFor: 'sources, change log' stamp: ''!
datedVersion
	"Answer the version of this release."

	^SystemVersion current datedVersion! !

!SmalltalkImage methodsFor: 'to clean later' stamp: ''!
do: aBlock
	"Obsoleted."
	self flag: #tobeprecated.
	"self deprecated: 'Use Smalltalk globals'."
	^globals do: aBlock
	
! !

!SmalltalkImage methodsFor: 'system attribute' stamp: ''!
documentPath
	"Answer the absolute path of the document passed to the vm or nil if none."
	"Smalltalk commandLine documentPath"
	
	^self getSystemAttribute: 2! !

!SmalltalkImage methodsFor: 'endian' stamp: ''!
endianness

	EndianCache ifNil: [EndianCache := self calcEndianness].
	^ EndianCache.
! !

!SmalltalkImage methodsFor: 'accessing' stamp: ''!
environment
	"For conversion from Smalltalk to SystemDictionary"

	^globals! !

!SmalltalkImage methodsFor: 'sources, change log' stamp: ''!
event: anEvent
	"Hook for SystemChangeNotifier"

	(anEvent isRemoved and: [anEvent itemKind = SystemChangeNotifier classKind]) ifTrue: [
		anEvent item acceptsLoggingOfCompilation 
			ifTrue: [self logChange: 'Smalltalk globals removeClassNamed: #' , anEvent item name].
	].
	anEvent isDoIt 
		ifTrue: [self logChange: anEvent item].
	(anEvent isRemoved and: [anEvent itemKind = SystemChangeNotifier methodKind]) ifTrue: [
		anEvent itemClass acceptsLoggingOfCompilation 
			ifTrue: [self logChange: anEvent itemClass name , ' removeSelector: #' , anEvent itemSelector]].! !

!SmalltalkImage methodsFor: 'miscellaneous' stamp: ''!
exitToDebugger
	"Primitive. Enter the machine language debugger, if one exists. Essential.
	See Object documentation whatIsAPrimitive."

	<primitive: 114>
	self primitiveFailed! !

!SmalltalkImage methodsFor: 'special objects' stamp: ''!
externalObjects
	"Return an array of objects that have been registered for use in non-Smalltalk code. Smalltalk objects should be referrenced by external code only via indirection through this array, thus allowing the objects to move during compaction. This array can be cleared when the VM re-starts, since variables in external code do not survive snapshots. Note that external code should not attempt to access a Smalltalk object, even via this mechanism, while garbage collection is in progress."
	"Smalltalk externalObjects"

	^ ExternalSemaphoreTable externalObjects
! !

!SmalltalkImage methodsFor: 'sources, change log' stamp: ''!
externalizeSources   
	"Write the sources and changes streams onto external files."
 	"Smalltalk externalizeSources"
	"the logic of this method is complex because it uses changesName and self changesName
	may be this is normal - sd"
	
	| sourcesName changesName aFile |
	sourcesName := self sourcesName.
	(FileDirectory default fileExists: sourcesName)
		ifTrue: [^ self inform:
'Sorry, you must first move or remove the
file named ', sourcesName].
	changesName := self changesName.
	(FileDirectory default fileExists: changesName)
		ifTrue: [^ self inform:
'Sorry, you must first move or remove the
file named ', changesName].

	aFile :=  FileStream newFileNamed: sourcesName.
	aFile nextPutAll: SourceFiles first originalContents.
	aFile close.
	self setMacFileInfoOn: sourcesName.
	SourceFiles at: 1 put: (FileStream readOnlyFileNamed: sourcesName).

	aFile := FileStream newFileNamed: self changesName.
	aFile nextPutAll: SourceFiles last contents.
	aFile close.
	"On Mac, set the file type and creator (noop on other platforms)"
	FileDirectory default
		setMacFileNamed: Smalltalk changesName
		type: 'STch'
		creator: 'FAST'.
	SourceFiles at: 2 put: (FileStream oldFileNamed: changesName).

	self inform: 'Sources successfully externalized'.
! !

!SmalltalkImage methodsFor: 'vm parameters' stamp: ''!
extraVMMemory
	"Answer the current setting of the 'extraVMMemory' VM parameter. See the comment in extraVMMemory: for details."

	^ self vmParameterAt: 23
! !

!SmalltalkImage methodsFor: 'vm parameters' stamp: ''!
extraVMMemory: extraBytesToReserve
	"Request that the given amount of extra memory be reserved for use by the virtual machine to leave extra C heap space available for things like plugins, network and file buffers, and so on. This request is stored when the image is saved and honored when the image is next started up. Answer the previous value of this parameter."

	extraBytesToReserve < 0
		ifTrue: [self error: 'VM memory reservation must be non-negative'].
	^ self vmParameterAt: 23 put: extraBytesToReserve
! !

!SmalltalkImage methodsFor: 'system attribute' stamp: ''!
extractMinusParameters
	"Returns a dictionary matching parameters beginning by a minus character and the list of the next parameters.
	Here is an example: 
		I suppose I run pharo with: `pharo myImage.image -minus1 'Hello' -minus2 -minus3 'Hello' 'World'`.
		This message will return a citionary with: 
			minus1 --> #(Hello)
			minus2 --> #()
			minus3 --> #('Hello' 'World')"
			
	| args dict str previous previousWithMinus index |
	
	dict := Dictionary new.
	args := OrderedCollection new.
	previous := previousWithMinus := ''.

	str := self getSystemAttribute: (index := 2).
	[ str isNil ] whileFalse: [ args add: str. str := self getSystemAttribute: (index := index + 1) ].

	args ifEmpty: [ ^ dict ].

	previous := previousWithMinus := args detect: [ :each | each first = $- ].
	[ (args first = previous) & (args isEmpty not) ] whileFalse: [ args removeFirst ].
	
	args do: [ :each |
		(each first = $-)
			ifTrue: [ 
				(previous first = $-) ifTrue: [ dict at: previous allButFirst put: #() ].
				previousWithMinus := each ]
			ifFalse: [ dict at: previousWithMinus allButFirst put: ((dict at: previousWithMinus allButFirst ifAbsent: [ #() ]), (Array with: each)) ].
				
		previous := each.
	].

	(previous first = $-) ifTrue: [ dict at: previous allButFirst put: #() ].

	^ dict! !

!SmalltalkImage methodsFor: 'system attribute' stamp: ''!
extractParameters

	| pName value index dict |
	dict := Dictionary new.
	index := 3. "Muss bei 3 starten, da 2 documentName ist"
	[pName := self  getSystemAttribute: index.
	pName isEmptyOrNil] whileFalse:[
		index := index + 1.
		value := self getSystemAttribute: index.
		value ifNil: [value := ''].
 		dict at: pName asUppercase put: value.
		index := index + 1].
	^dict! !

!SmalltalkImage methodsFor: 'housekeeping' stamp: ''!
fixObsoleteReferences
	"SmalltalkImage current fixObsoleteReferences.
	SystemNavigation default obsoleteBehaviors size > 0
		ifTrue: [ SystemNavigation default obsoleteBehaviors inspect.
			self error:'Still have obsolete behaviors. See inspector']"

	| informee obsoleteBindings obsName realName realClass |
	Smalltalk
		garbageCollect;
		garbageCollect.
	CompiledMethod allInstances
		do: [ :method | 
			obsoleteBindings := method literals
				select: [ :literal | literal isVariableBinding 
									and: [ literal value isBehavior  and: [ literal value isObsolete ]]].
			obsoleteBindings
				do: [ :binding | 
					obsName := binding value name.
					Transcript
						show: 'Binding: ';
						show: obsName;
						cr.
					realName := obsName copyReplaceAll: 'AnObsolete' with: ''.
					realClass := Smalltalk globals at: realName asSymbol ifAbsent: [ UndefinedObject ].
					binding key: binding key value: realClass ] ].
	Behavior flushObsoleteSubclasses.
	Smalltalk
		garbageCollect;
		garbageCollect! !

!SmalltalkImage methodsFor: 'housekeeping' stamp: ''!
fixUpProblemsWithAllCategory
	"Moves all methods that are in formally classified a category named '-- all --' into the default 'as yet unclassified' category"

	"Smalltalk fixUpProblemsWithAllCategory"

	| methodCount classCount |
	self flag: #ShouldBeMovedInClassOrganization.
	methodCount := 0.
	classCount := 0.
	self systemNavigation allBehaviorsDo:
		[:aClass | | org any | org := aClass organization.
			(org categories includes: #'-- all --') ifTrue:
				[any := false.
				aClass selectorsDo:
					[:aSelector | | aCategory |
						aCategory := org categoryOfElement: aSelector.
						aCategory = #'-- all --' ifTrue:
							[org classify: aSelector under: ClassOrganizer default suppressIfDefault: false.
							Transcript cr; show: aClass name, ' >> ', aSelector.
							methodCount := methodCount + 1.
							any := true]].
			any ifTrue: [classCount := classCount + 1].
			org removeEmptyCategories]].
	Transcript cr; show: methodCount printString, ' methods in ', classCount printString, ' classes moved from "-- all --" to "as yet unclassified"'
! !

!SmalltalkImage methodsFor: 'system attribute' stamp: ''!
flagInterpretedMethods: aBoolean
	"The Cog VM can be instructed to set the flag bit of CompiledMethods that it executes but will only interpret.  This can be used e.g. to profile startup. See CompiledMethod>>#flag & CompiledMethod>>#clearFlag. This flag persists across snapshots, stored in the image header."
	
	self vmParameterAt: 48 put: ((self vmParameterAt: 48) bitClear: 2) + (aBoolean ifTrue: [2] ifFalse: [0])! !

!SmalltalkImage methodsFor: 'to clean later' stamp: ''!
flushClassNameCache
	"Smalltalk flushClassNameCache"
	"Force recomputation of the cached list of class names."

	globals flushClassNameCache.! !

!SmalltalkImage methodsFor: 'sources, change log' stamp: ''!
forceChangesToDisk
	"Ensure that the changes file has been fully written to disk by closing and re-opening it. This makes the system more robust in the face of a power failure or hard-reboot."

	| changesFile |
	changesFile := SourceFiles at: 2.
	(changesFile isKindOf: FileStream) ifTrue: [
		changesFile flush.
		SecurityManager default hasFileAccess ifTrue:[
			changesFile close.
			changesFile open: changesFile name forWrite: true].
		changesFile setToEnd.
	].
! !

!SmalltalkImage methodsFor: 'memory space' stamp: ''!
forceTenure
	"Primitive. Tell the GC logic to force a tenure on the next increment GC."
	<primitive: 'primitiveForceTenure'>
	^self primitiveFailed! !

!SmalltalkImage methodsFor: 'to clean later' stamp: ''!
forgetClass: aClass logged: aBool 
	"Delete the class, aClass, from the system."

	^globals forgetClass: aClass logged: aBool! !

!SmalltalkImage methodsFor: 'modules' stamp: ''!
forgetModule: aString
	"Primitive. If the module named aString is loaded, unloaded. If not, and it is marked an unloadable, unmark it so the VM will try to load it again next time. See comment for #unloadModule:."
	<primitive: 571>
	^self primitiveFailed! !

!SmalltalkImage methodsFor: 'image, changes names' stamp: ''!
fullNameForChangesNamed: aName
	| imgName |
	imgName := self fullNameForImageNamed: aName.
	^FileDirectory fileName: (FileDirectory baseNameFor: imgName) extension: FileDirectory changeSuffix.! !

!SmalltalkImage methodsFor: 'image, changes names' stamp: ''!
fullNameForImageNamed: aName
	| imgDir |
	imgDir := FileDirectory on: self imagePath.
	^FileDirectory fileName: (imgDir fullNameFor: aName) extension: FileDirectory imageSuffix.! !

!SmalltalkImage methodsFor: 'memory space' stamp: ''!
garbageCollect
	"Primitive. Reclaims all garbage and answers the number of bytes of available space."
	Object flushDependents.
	Object flushEvents.
	^self primitiveGarbageCollect! !

!SmalltalkImage methodsFor: 'housekeeping' stamp: 'AlainPlantec 12/1/2010 09:54'!
garbageCollectAndReport
	"Do a garbage collection, and report results to the user."

	"SmalltalkImage current garbageCollectAndReport"

	self inform: self bytesLeftString! !

!SmalltalkImage methodsFor: 'memory space' stamp: ''!
garbageCollectMost
	"Primitive. Reclaims recently created garbage (which is usually most of it) fairly quickly and answers the number of bytes of available space."

	<primitive: 131>
	^ self primBytesLeft! !

!SmalltalkImage methodsFor: 'vm parameters' stamp: 'slave 01/17/2011 11:31'!
gcBiasToGrow: aBool
	"Tell the VM to grow after tenures instead of running in a tight loop where it does IGCs over and over. For some weird reason the primitive takes an integer not a bool but oh well..."
	
	self primitiveGCBiasToGrow: (aBool ifTrue: [1] ifFalse: [0])! !

!SmalltalkImage methodsFor: 'vm parameters' stamp: 'slave 01/17/2011 11:31'!
gcBiasToGrowLimit: arg
	"Tell the VM the grow limit if the GC logic has bias to grow."
	
	<primitive: 'primitiveSetGCBiasToGrowGCLimit'>
	^self primitiveFailed! !

!SmalltalkImage methodsFor: 'snapshot and quit' stamp: ''!
getFileNameFromUser

	| newName |
	newName := UIManager default
		request: 'New File Name?' translated
		initialAnswer: (FileDirectory localNameFor: self imageName).
	newName isEmptyOrNil ifTrue: [^nil].
	((FileDirectory default fileOrDirectoryExists: (self fullNameForImageNamed: newName)) or:
	 [FileDirectory default fileOrDirectoryExists: (self fullNameForChangesNamed: newName)]) ifTrue: [
		(self confirm: ('{1} already exists. Overwrite?' translated format: {newName})) ifFalse: [^nil]].
	^newName
! !

!SmalltalkImage methodsFor: 'system attribute' stamp: ''!
getSystemAttribute: attributeID 
	"Optional. Answer the string for the system attribute with the given 
	integer ID. Answer nil if the given attribute is not defined on this 
	platform. On platforms that support invoking programs from command 
	lines (e.g., Unix), this mechanism can be used to pass command line 
	arguments to programs written in Pharo.

	By convention, the first command line argument that is not a VM
	configuration option is considered a 'document' to be filed in. Such a
	document can add methods and classes, can contain a serialized object,
	can include code to be executed, or any combination of these.

	Currently defined attributes include: 
	-1000...-1 - command line arguments that specify VM options 
	0 - the full path name for currently executing VM 
	(or, on some platforms, just the path name of the VM's directory) 
	1 - full path name of this image 
	2 - a Pharo document to open, if any 
	3...1000 - command line arguments for Pharo programs 
	1001 - this platform's operating system 
	1002 - operating system version 
	1003 - this platform's processor type
	1004 - vm version"

	<primitive: 149>
	^ nil! !

!SmalltalkImage methodsFor: 'vm parameters' stamp: ''!
getVMParameters	
	"Answer an Array containing the current values of the VM's internal
	parameter/metric registers.  Each value is stored in the array at the
	index corresponding to its VM register.  (See #vmParameterAt: and
	#vmParameterAt:put:.)"
	"SmalltalkImage current getVMParameters"
	
	<primitive: 254>
	self primitiveFailed! !

!SmalltalkImage methodsFor: 'accessing' stamp: ''!
globals
	"Answer the global SystemDictionary"
	^globals! !

!SmalltalkImage methodsFor: 'private' stamp: ''!
globals: aSystemDictionary
	"Sets the system-wide globals"
	
	globals ifNotNil: [self error: 'Cannot overwrite existing globals'].
	globals := aSystemDictionary! !

!SmalltalkImage methodsFor: 'to clean later' stamp: ''!
hasBindingThatBeginsWith: aString

	^ self globals hasBindingThatBeginsWith: aString! !

!SmalltalkImage methodsFor: 'class and trait names' stamp: ''!
hasClassNamed: aString
	"Answer whether there is a class of the given name, but don't intern aString if it's not alrady interned."

	^ globals hasClassNamed: aString! !

!SmalltalkImage methodsFor: 'testing' stamp: ''!
hasDisplay

	| arg i |
	i := -1.
	[(arg := self getSystemAttribute: i) isNil]
		whileFalse: [(#('-nodisplay' '-headless') includes: arg) ifTrue: [^ false].
			i := i - 1].
	^ true! !

!SmalltalkImage methodsFor: 'special objects' stamp: ''!
hasSpecialSelector: aLiteral ifTrueSetByte: aBlock
	"Some basic messages such as #+ have their own bytecode for speed.
	To gain some space, the literal array of compiled methods does not contain such message selectors.
	This implies that the cross referencer and other tools have to check if a selector is of the optimized one 
	or not by calling this method. This is an optimization when machines were slow and small and it 
	may be removed in the future to gain simplicity."
	
	1 to: self specialSelectorSize do:
		[:index | 
		(self specialSelectorAt: index) == aLiteral
			ifTrue: [aBlock value: index + 16rAF. ^true]].
	^false! !

!SmalltalkImage methodsFor: 'image' stamp: ''!
image
	"Answer the object to query about the current object memory and execution environment."
	
	^self! !

!SmalltalkImage methodsFor: 'image' stamp: ''!
imageFormatVersion
	"Answer an integer identifying the type of image. The image version number may
	identify the format of the image (e.g. 32 or 64-bit word size) or specific requirements
	of the image (e.g. block closure support required). This invokes an optional primitive
	that may not be available on all virtual machines."

	"Smalltalk image imageFormatVersion"

	<primitive: 'primitiveImageFormatVersion'>
	self notify: 'This virtual machine does not support the optional primitive #primitiveImageFormatVersion' translated.
	^''! !

!SmalltalkImage methodsFor: 'image, changes names' stamp: ''!
imageName
	"Answer the full path name for the current image."
	"SmalltalkImage current imageName"

	| str |
	str := self primImageName.
	^ (FilePath pathName: str isEncoded: true) asSqueakPathName.
! !

!SmalltalkImage methodsFor: 'image, changes names' stamp: ''!
imageName: newName
	"Set the the full path name for the current image.  All further snapshots will use this."

	| encoded |
	encoded := (FilePath pathName: newName isEncoded: false) asVmPathName.
	self primImageName: encoded.
! !

!SmalltalkImage methodsFor: 'image, changes names' stamp: ''!
imagePath
	"Answer the path for the directory containing the image file."
	"SmalltalkImage current imagePath"

	^ FileDirectory dirPathFor: self imageName
! !

!SmalltalkImage methodsFor: 'to clean later' stamp: ''!
includes: element
	"Obsoleted."

	self flag: #tobeprecated.
	"self deprecated: 'Use Smalltalk globals'."
	^globals includes: element! !

!SmalltalkImage methodsFor: 'to clean later' stamp: ''!
includesKey: key
	"Answer whether the receiver has a key equal to the argument, key."

	^globals includesKey: key! !

!SmalltalkImage methodsFor: 'memory space' stamp: ''!
installLowSpaceWatcher
	"Start a process to watch for low-space conditions."
	"Smalltalk installLowSpaceWatcher"

	self primSignalAtBytesLeft: 0.  "disable low-space interrupts"
	LowSpaceProcess == nil ifFalse: [LowSpaceProcess terminate].
	LowSpaceProcess := [self lowSpaceWatcher] newProcess.
	LowSpaceProcess priority: Processor lowIOPriority.
	LowSpaceProcess resume.

! !

!SmalltalkImage methodsFor: 'sources, change log' stamp: ''!
internalizeChangeLog    
		"Smalltalk internalizeChangeLog"
	"Bring the changes file into a memory-resident filestream, for faster access and freedom from external file system.  1/31/96 sw"

	| reply aName aFile |
	reply := self confirm:  'CAUTION -- do not undertake this lightly!!
If you have backed up your system and
are prepared to face the consequences of
the requested internalization of sources,
hit Yes.  If you have any doubts, hit No
to back out with no harm done.'.

	(reply ==  true) ifFalse:
		[^ self inform: 'Okay - abandoned'].

	aName := Smalltalk changesName.
	(aFile := SourceFiles last) == nil ifTrue:
		[(FileDirectory default fileExists: aName)
			ifFalse: [^ self halt: 'Cannot locate ', aName, ' so cannot proceed.'].
		aFile := FileStream readOnlyFileNamed: aName].
	SourceFiles at: 2 put: (ReadWriteStream with: aFile contentsOfEntireFile).

	self inform: 'Okay, changes file internalized'! !

!SmalltalkImage methodsFor: 'sources, change log' stamp: ''!
internalizeSources    
		"Smalltalk internalizeSources"
	"Bring the sources and changes files into memory-resident filestreams, for faster access and freedom from file-system interface.  1/29/96 sw"

	| reply aName aFile |
	reply := self confirm:  'CAUTION -- do not undertake this lightly!!
If you have backed up your system and
are prepared to face the consequences of
the requested internalization of sources,
hit Yes.  If you have any doubts, hit No
to back out with no harm done.'.

	(reply ==  true) ifFalse:
		[^ self inform: 'Okay - abandoned'].

	aName := self sourcesName.
	(aFile := SourceFiles first) == nil ifTrue:
		[(FileDirectory default fileExists: aName)
			ifFalse: [^ self halt: 'Cannot locate ', aName, ' so cannot proceed.'].
		aFile := FileStream readOnlyFileNamed: aName].
	SourceFiles at: 1 put: (ReadWriteStream with: aFile contentsOfEntireFile).

	aName := self changesName.
	(aFile := SourceFiles last) == nil ifTrue:
		[(FileDirectory default fileExists: aName)
			ifFalse: [^ self halt: 'Cannot locate ', aName, ' so cannot proceed.'].
		aFile := FileStream readOnlyFileNamed: aName].
	SourceFiles at: 2 put: (ReadWriteStream with: aFile contentsOfEntireFile).

	self inform: 'Okay, sources internalized'! !

!SmalltalkImage methodsFor: 'vm' stamp: ''!
interpreterSourceVersion
	"Answer a string corresponding to the version of the interpreter source.
	This represents the version level of the Smalltalk source code (interpreter
	and various plugins) that is translated to C by a CCodeGenerator, as distinct
	from the external platform source code, typically written in C and managed
	separately for each platform. An optional primitive is invoked that may not
	be available on all virtual machines."

	"Smalltalk vm interpreterSourceVersion"

	<primitive: 'primitiveInterpreterSourceVersion'>
	self notify: 'This virtual machine does not support the optional primitive #primitiveInterpreterSourceVersion' translated.
	^''! !

!SmalltalkImage methodsFor: 'endian' stamp: ''!
isBigEndian
	^self endianness == #big! !

!SmalltalkImage methodsFor: 'endian' stamp: ''!
isLittleEndian
	^self endianness == #little! !

!SmalltalkImage methodsFor: 'system attribute' stamp: ''!
isMacOS

	self deprecated: 'Use OSPlatform'.
	^ OSPlatform isMacOS! !

!SmalltalkImage methodsFor: 'memory space' stamp: ''!
isRoot: oop
	"Primitive. Answer whether the object is currently a root for youngSpace."
	<primitive: 'primitiveIsRoot'>
	^self primitiveFailed! !

!SmalltalkImage methodsFor: 'system attribute' stamp: ''!
isRunningCog
	"Returns true if we're running on a Cog VM (JIT or StackInterpreter)
	 (vmParameterAt: 42 is the number of stack pages)"

	^[(self vmParameterAt: 42) > 0] on: Error do:[:ex| ex return: false]! !

!SmalltalkImage methodsFor: 'system attribute' stamp: ''!
isRunningCogit
	"Returns true if we're running on the Cog JIT
	 (vmParameterAt: 46 is the size of the machine code zone)"

	^[(self vmParameterAt: 46) > 0] on: Error do:[:ex| ex return: false]! !

!SmalltalkImage methodsFor: 'printing' stamp: ''!
isSelfEvaluating
	self == Smalltalk ifTrue: [^true].
	^super isSelfEvaluating! !

!SmalltalkImage methodsFor: 'testing' stamp: ''!
isTrunk
	"Trunk image file names use several letter combinations e.g. 'a' for alpha, 'rc' for release candidates"

	^ (self version allButFirst: 5) anySatisfy: [ :char | char isLetter ]! !

!SmalltalkImage methodsFor: 'system attribute' stamp: ''!
isUnix

	self deprecated: 'Use OSPlatform'.
	^ OSPlatform isUnix ! !

!SmalltalkImage methodsFor: 'system attribute' stamp: ''!
isWin32

	self deprecated: 'Use OSPlatform'.
	^ OSPlatform isWin32 ! !

!SmalltalkImage methodsFor: 'system attribute' stamp: ''!
isWin32CE

	self deprecated: 'Use OSPlatform'.
	^ OSPlatform isWin32CE! !

!SmalltalkImage methodsFor: 'system attribute' stamp: ''!
isX11

	self deprecated: 'Use OSPlatform'.
	^ OSPlatform isX11! !

!SmalltalkImage methodsFor: 'memory space' stamp: ''!
isYoung: oop
	"Primitive. Answer whether the object currently resides in youngSpace."
	<primitive: 'primitiveIsYoung'>
	^self primitiveFailed! !

!SmalltalkImage methodsFor: 'to clean later' stamp: ''!
keyAtIdentityValue: anObject ifAbsent: aBlock
	"Obsoleted."

	self flag: #tobeprecated.
	"self deprecated: 'Use Smalltalk globals'."
	^globals keyAtIdentityValue: anObject ifAbsent: aBlock! !

!SmalltalkImage methodsFor: 'to clean later' stamp: ''!
keyAtValue: anObject
	"Obsoleted."
	self flag: #tobeprecated.
	"self deprecated: 'Use Smalltalk globals'."
	^globals keyAtValue: anObject! !

!SmalltalkImage methodsFor: 'to clean later' stamp: ''!
keys
	"Obsoleted."

	self flag: #tobeprecated.
	"self deprecated: 'Use Smalltalk globals'."
	^globals keys! !

!SmalltalkImage methodsFor: 'to clean later' stamp: ''!
keysAndValuesDo: aBlock
	"Obsoleted."
	self flag: #tobeprecated.
	"self deprecated: 'Use Smalltalk globals'."
	^globals keysAndValuesDo: aBlock
! !

!SmalltalkImage methodsFor: 'sources, change log' stamp: ''!
lastQuitLogPosition
	^ LastQuitLogPosition! !

!SmalltalkImage methodsFor: 'sources, change log' stamp: ''!
lastQuitLogPosition: aNumber
	"should be only use to ensure the transition from SystemDictionary to SmalltalkImage, then  	be removed"
		
	LastQuitLogPosition := aNumber! !

!SmalltalkImage methodsFor: 'shrinking' stamp: ''!
lastRemoval
	"Smalltalk lastRemoval"
	"Some explicit removals - add unwanted methods keeping
	other methods."
	| oldDicts newDicts |
	#(#abandonSources )
		do: [:each | self class removeSelector: each].
	"Get rid of all unsent methods."
	"Shrink method dictionaries."
	self garbageCollect.
	oldDicts := MethodDictionary allInstances.
	newDicts := Array new: oldDicts size.
	oldDicts
		withIndexDo: [:d :index | newDicts at: index put: d rehashWithoutBecome].
	oldDicts elementsExchangeIdentityWith: newDicts.
	oldDicts := newDicts := nil.
	self
		allClassesDo: [:c | c zapOrganization].
	SystemOrganization := nil.
	ChangeSet current initialize! !

!SmalltalkImage methodsFor: 'sources, change log' stamp: ''!
lastUpdateString
	"SmalltalkImage current lastUpdateString"
	^'Latest update: #' translated, SystemVersion current highestUpdate printString! !

!SmalltalkImage methodsFor: 'sources, change log' stamp: ''!
licenseString
	^ 'LICENSE
	
Licensed under the MIT License with parts under the Apache License.

Copyright (c) Pharo Project, and Contributors Copyright (c) 1996-2008 Viewpoints
Research Institute, and Contributors Copyright (c) 1996 Apple Computer, Inc.

Permission is hereby granted, free of charge, to any person obtaining a copy of this
software and associated documentation files (the "Software"), to deal in the Software
without restriction, including without limitation the rights to use, copy, modify, merge,
publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons
to whom the Software is furnished to do so, subject to the following conditions: The
above copyright notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS
OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

You may obtain a copy of the Apache License at
http://www.apache.org/licenses/LICENSE-2.0'! !

!SmalltalkImage methodsFor: 'modules' stamp: ''!
listBuiltinModule: index
	"Return the name of the n-th builtin module.
	This list is not sorted!!"
	<primitive: 572>
	^self primitiveFailed! !

!SmalltalkImage methodsFor: 'modules' stamp: ''!
listBuiltinModules
	"SmalltalkImage current listBuiltinModules"
	"Return a list of all builtin modules (e.g., plugins). Builtin plugins are those that are 	compiled with the VM directly, as opposed to plugins residing in an external shared library. 	The list will include all builtin plugins regardless of whether they are currently loaded 
	or not. Note that the list returned is not sorted!!"

	| modules index name |
	modules := Array new writeStream.
	index := 1.
	[true] whileTrue:[
		name := self listBuiltinModule: index.
		name ifNil:[^modules contents].
		modules nextPut: name.
		index := index + 1.
	].! !

!SmalltalkImage methodsFor: 'modules' stamp: ''!
listLoadedModule: index
	"Return the name of the n-th loaded module.
	This list is not sorted!!"
	<primitive: 573>
	^self primitiveFailed! !

!SmalltalkImage methodsFor: 'modules' stamp: ''!
listLoadedModules
	"SmalltalkImage current listLoadedModules"
	"Return a list of all currently loaded modules (e.g., plugins). Loaded modules are those that currently in use (e.g., active). The list returned will contain all currently active modules regardless of whether they're builtin (that is compiled with the VM) or external (e.g., residing in some external shared library). Note that the returned list is not sorted!!"
	| modules index name |
	modules := Array new writeStream.
	index := 1.
	[true] whileTrue:[
		name := self listLoadedModule: index.
		name ifNil:[^modules contents].
		modules nextPut: name.
		index := index + 1.
	].! !

!SmalltalkImage methodsFor: 'sources, change log' stamp: ''!
logChange: aStringOrText 
	"Write the argument, aString, onto the changes file."
	| aString changesFile |
	(SourceFiles isNil or: [(SourceFiles at: 2) == nil]) ifTrue: [^ self].
	self assureStartupStampLogged.

	aString := aStringOrText asString.
	(aString findFirst: [:char | char isSeparator not]) = 0
		ifTrue: [^ self].  "null doits confuse replay"
	(changesFile := SourceFiles at: 2).
	changesFile isReadOnly ifTrue:[^self].
	changesFile setToEnd; cr; cr.
	changesFile nextChunkPut: aString.
		"If want style changes in DoIt, use nextChunkPutWithStyle:, and allow Texts to get here"
	self forceChangesToDisk.! !

!SmalltalkImage methodsFor: 'miscellaneous' stamp: ''!
logError: errMsg inContext: aContext to: aFilename
	"Log the error message and a stack trace to the given file."

	"wks - 9-09 - do not delete existing errors.  Note that this could be made
	conditional or left here in this state for those who want it.
	FileDirectory default deleteFileNamed: aFilename ifAbsent: []."
	
	FileStream fileNamed: aFilename do: [:ff |
		ff wantsLineEndConversion: true.
		ff setToEnd.
		ff nextPutAll: 'THERE_BE_DRAGONS_HERE'; cr. 
	  	ff nextPutAll: errMsg; cr.
		aContext errorReportOn: ff.
		"wks 9-09 - write some type of separator"
		ff nextPutAll: (String new: 60 withAll: $- ); cr; cr].
! !

!SmalltalkImage methodsFor: 'memory space' stamp: ''!
lowSpaceThreshold 
	"Return the low space threshold. When the amount of free memory (after garbage collection) falls below this limit, the system is in serious danger of completely exhausting memory and crashing. This limit should be made high enough to allow the user open a debugger to diagnose a problem or to save the image."

	thisContext isPseudoContext
		ifTrue: [^ 400000  "Enough for JIT compiler"]
		ifFalse: [^ 200000  "Enough for interpreter"]! !

!SmalltalkImage methodsFor: 'memory space' stamp: ''!
lowSpaceWatcher
	"Wait until the low space semaphore is signalled, then take appropriate
	actions. "
	| free preemptedProcess |
	self garbageCollectMost <= self lowSpaceThreshold
		ifTrue: [self garbageCollect <= self lowSpaceThreshold
				ifTrue: ["free space must be above threshold before
					starting low space watcher"
					^ Beeper beep]].
	self specialObjectsArray at: 23 put: nil.
	"process causing low space will be saved here"
	LowSpaceSemaphore := Semaphore new.
	self primLowSpaceSemaphore: LowSpaceSemaphore.
	self primSignalAtBytesLeft: self lowSpaceThreshold.
	"enable low space interrupts"
	LowSpaceSemaphore wait.
	"wait for a low space condition..."
	self primSignalAtBytesLeft: 0.
	"disable low space interrupts"
	self primLowSpaceSemaphore: nil.
	LowSpaceProcess := nil.
	"The process that was active at the time of the low space interrupt."
	preemptedProcess := Smalltalk specialObjectsArray at: 23.
	Smalltalk specialObjectsArray at: 23 put: nil.
	"Note: user now unprotected until the low space watcher is re-installed"
	self memoryHogs isEmpty
		ifFalse: [free := self bytesLeft.
			self memoryHogs
				do: [:hog | hog freeSomeSpace].
			self bytesLeft > free
				ifTrue: [^ self installLowSpaceWatcher]].
	UIManager default lowSpaceWatcherDefaultAction: preemptedProcess! !

!SmalltalkImage methodsFor: 'memory space' stamp: ''!
lowSpaceWatcherProcess
	^LowSpaceProcess! !

!SmalltalkImage methodsFor: 'memory space' stamp: ''!
memoryHogs
	"Answer the list of objects to notify with #freeSomeSpace if memory gets full."

	^ MemoryHogs ifNil: [MemoryHogs := OrderedCollection new]! !

!SmalltalkImage methodsFor: 'objects from disk' stamp: 'tk 9/28/2000 15:50'!
objectForDataStream: refStrm
	| dp |
	"I am about to be written on an object file.  Write a reference to Smalltalk instead."

	dp := DiskProxy global: #Smalltalk selector: #yourself
			args: #().
	refStrm replace: self with: dp.
	^ dp! !

!SmalltalkImage methodsFor: 'memory space' stamp: ''!
okayToProceedEvenIfSpaceIsLow
	"Return true if either there is enough memory to do so safely or if the user gives permission after being given fair warning."

	self garbageCollectMost > self lowSpaceThreshold ifTrue: [^ true].  "quick"
	self garbageCollect > self lowSpaceThreshold ifTrue: [^ true].  "work harder"

	^ self confirm:
'WARNING: There is not enough space to start the low space watcher.
If you proceed, you will not be warned again, and the system may
run out of memory and crash. If you do proceed, you can start the
low space notifier when more space becomes available simply by
opening and then closing a debugger (e.g., by hitting Cmd-period.)
Do you want to proceed?'
! !

!SmalltalkImage methodsFor: 'sources, change log' stamp: ''!
openSourceFiles

	self imageName = LastImageName 
		ifFalse: ["Reset the author full name to blank when the image gets moved"
			LastImageName := self imageName.
			Author fullName: ''].
	self checkAndOpenSourcesAndChanges.
	StandardSourceFileArray install! !

!SmalltalkImage methodsFor: 'system attribute' stamp: ''!
optionAt: i
	"Answer the i-th option of the command line, or nil if not so many options."
	
	^self getSystemAttribute: i negated! !

!SmalltalkImage methodsFor: 'system attribute' stamp: ''!
options
	"Answer an array with all the command line options."
	
	"Smalltalk commandLine options"
	
	^Array streamContents: [:str |
		| arg i |
		i := 1.
		[i > 1000 or: [(arg := self optionAt: i) == nil]]
			whileFalse:
				[str nextPut: arg.
				i := i + 1]].! !

!SmalltalkImage methodsFor: 'class and trait names' stamp: ''!
organization
	"Return the organizer for the receiver"
	^globals organization! !

!SmalltalkImage methodsFor: 'os' stamp: ''!
os
	"Answer the object to query about os."
	
	^ OSPlatform! !

!SmalltalkImage methodsFor: 'os' stamp: ''!
osVersion
	"Return the version number string of the platform we're running on"
	"SmalltalkImage current osVersion"

	self deprecated: 'Use OSPlatform'.
	^ OSPlatform osVersion! !

!SmalltalkImage methodsFor: 'os' stamp: ''!
platform
	"Answer the object to query about os."
	
	^ OSPlatform! !

!SmalltalkImage methodsFor: 'os' stamp: ''!
platformName
	"Return the name of the platform we're running on"

	self deprecated: 'Use Smalltalk os platformName'.
	^self os platformName! !

!SmalltalkImage methodsFor: 'vm' stamp: ''!
platformSourceVersion
	"Answer a string corresponding to the version of the external platform source
	code, typically written in C and managed separately for each platform. This
	invokes an optional primitive that may not be available on all virtual machines."

	"Smalltalk vm platformSourceVersion"

	<primitive: 'primitivePlatformSourceVersion'>
	self notify: 'This virtual machine does not support the optional primitive #primitivePlatformSourceVersion' translated.
	^''! !

!SmalltalkImage methodsFor: 'os' stamp: ''!
platformSubtype
	"Return the subType of the platform we're running on"

	self deprecated: 'Use OSPlatform'.
	^ OSPlatform platformSubtype ! !

!SmalltalkImage methodsFor: 'to clean later' stamp: ''!
poolUsers
	"Answer a dictionary of pool name -> classes that refer to it.
	Also includes any globally know dictionaries (such as
	Smalltalk, Undeclared etc) which although not strictly
	accurate is potentially useful information"
	"Smalltalk poolUsers"
	| poolUsers |
	poolUsers := Dictionary new.
	self keys
		do: [:k | | pool refs | "yes, using isKindOf: is tacky but for reflective code like
			this it is very useful. If you really object you can:-
			a) go boil your head.
			b) provide a better answer.
			your choice."
			(((pool := self at: k) isKindOf: Dictionary)
					or: [pool isKindOf: SharedPool class])
				ifTrue: [refs := self systemNavigation allClasses
								select: [:c | c sharedPools identityIncludes: pool]
								thenCollect: [:c | c name].
					refs
						add: (self systemNavigation
								allReferencesTo: (self associationAt: k)).
					poolUsers at: k put: refs]].
	^ poolUsers! !

!SmalltalkImage methodsFor: 'memory space' stamp: ''!
primBytesLeft
	"Primitive. Answer the number of bytes available for new object data.
	Not accurate unless preceded by
		Smalltalk garbageCollectMost (for reasonable accuracy), or
		Smalltalk garbageCollect (for real accuracy).
	See Object documentation whatIsAPrimitive."

	<primitive: 112>
	^ 0! !

!SmalltalkImage methodsFor: 'image, changes name' stamp: ''!
primImageName
	"Answer the full path name for the current image."
	"Smalltalk imageName"

	<primitive: 121>
	self primitiveFailed! !

!SmalltalkImage methodsFor: 'image, changes name' stamp: ''!
primImageName: newName
	"Set the the full path name for the current image.  All further snapshots will use this."

	<primitive: 121>
	^ self primitiveFailed! !

!SmalltalkImage methodsFor: 'memory space' stamp: ''!
primLowSpaceSemaphore: aSemaphore
	"Primitive. Register the given Semaphore to be signalled when the
	number of free bytes drops below some threshold. Disable low-space
	interrupts if the argument is nil."

	<primitive: 124>
	self primitiveFailed! !

!SmalltalkImage methodsFor: 'memory space' stamp: ''!
primSignalAtBytesLeft: numBytes
	"Tell the interpreter the low-space threshold in bytes. When the free
	space falls below this threshold, the interpreter will signal the low-space
	semaphore, if one has been registered.  Disable low-space interrupts if the
	argument is zero.  Fail if numBytes is not an Integer."

	<primitive: 125>
	self primitiveFailed! !

!SmalltalkImage methodsFor: 'image, changes name' stamp: ''!
primVmPath
	"Answer the path for the directory containing the Smalltalk virtual machine. Return the empty string if this primitive is not implemented."
	"Smalltalk vmPath"

	<primitive: 142>
	^ ''! !

!SmalltalkImage methodsFor: 'vm parameters' stamp: 'slave 01/17/2011 11:31'!
primitiveGCBiasToGrow: arg
	"Tell the VM to grow after tenures instead of running in a tight loop where it does IGCs over and over. For some weird reason the primitive takes an integer not a bool but oh well..."
	
	<primitive: 'primitiveSetGCBiasToGrow'>
	^self primitiveFailed! !

!SmalltalkImage methodsFor: 'memory space' stamp: ''!
primitiveGarbageCollect
	"Primitive. Reclaims all garbage and answers the number of bytes of available space."

	<primitive: 130>
	^ self primBytesLeft! !

!SmalltalkImage methodsFor: 'printing' stamp: ''!
printElementsOn: aStream
	aStream nextPutAll:'(lots of globals)'! !

!SmalltalkImage methodsFor: 'printing' stamp: ''!
printOn: aStream
	self == Smalltalk ifFalse: [^super printOn: aStream].
	aStream nextPutAll: 'Smalltalk'! !

!SmalltalkImage methodsFor: 'system attribute' stamp: ''!
processHasThreadIdInstVar: aBoolean
	"The threaded VM needs to know if the 4th inst var of Process is threadId which it uses to control process-to-thread binding. This flag persists across snapshots, stored in the image header."
	aBoolean ifTrue: [self assert: (Process instVarNames at: 4) ='threadId'].
	self vmParameterAt: 48 put: ((self vmParameterAt: 48) bitClear: 1) + (aBoolean ifTrue: [1] ifFalse: [0])! !

!SmalltalkImage methodsFor: 'system attribute' stamp: ''!
processPreemptionYields
	"Answer whether the VM causes a process to yield on process preemption, i.e. to put a preempted process at the back of its run queue.  If the parameter is unavailable (non-Cog VMs) or bit 2 (4) is 0 then preemption yields."

	^(([self vmParameterAt: 48]
			on: Error
			do: [:ex| ^true]) allMask: 4) not! !

!SmalltalkImage methodsFor: 'system attribute' stamp: ''!
processPreemptionYields: aBoolean
	"The Cog VM can be instructed not to yield on process preemption, i.e. not to put a preempted process at the back of its run queue.  By default preempting a process causes it to yield (Blue Book semantics) which can have unfortunate effects. This flag persists across snapshots, stored in the image header."
	
	self vmParameterAt: 48 put: ((self vmParameterAt: 48) bitClear: 4) + (aBoolean ifTrue: [0] ifFalse: [4])! !

!SmalltalkImage methodsFor: 'snapshot and quit' stamp: ''!
processShutDownList: quitting
	"Send #shutDown to each class that needs to wrap up before a snapshot.
	Also void the endianness chace; this can't safely be done on start-up because
	Smalltalk is too late in the start-up sequence."

	EndianCache := nil.
	self send: #shutDown: toClassesNamedIn: ShutDownList with: quitting.
! !

!SmalltalkImage methodsFor: 'snapshot and quit' stamp: ''!
processStartUpList: resuming
	"Send #startUp to each class that needs to run initialization after a snapshot."

	self send: #startUp: toClassesNamedIn: StartUpList with: resuming.
! !

!SmalltalkImage methodsFor: 'snapshot and quit' stamp: ''!
quitPrimitive
	"Primitive. Exit to another operating system on the host machine, if one
	exists. All state changes in the object space since the last snapshot are lost.
	Essential. See Object documentation whatIsAPrimitive."

	<primitive: 113>
	self primitiveFailed! !

!SmalltalkImage methodsFor: 'snapshot and quit' stamp: ''!
recordStartupStamp
	
	StartupStamp := '----STARTUP----', Time dateAndTimeNow printString, ' as ', self imageName.
! !

!SmalltalkImage methodsFor: 'sources, change log' stamp: ''!
recover: nCharacters
	"Schedule an editable text view on the last n characters of changes."
	self writeRecentCharacters: nCharacters toFileNamed: 'st80.recent'! !

!SmalltalkImage methodsFor: 'special objects' stamp: ''!
recreateSpecialObjectsArray
	"Smalltalk recreateSpecialObjectsArray"
	
	"To external package developers:
	**** DO NOT OVERRIDE THIS METHOD.  *****
	If you are writing a plugin and need additional special object(s) for your own use, 
	use addGCRoot() function and use own, separate special objects registry "
	
	"The Special Objects Array is an array of objects used by the Squeak virtual machine.
	 Its contents are critical and accesses to it by the VM are unchecked, so don't even
	 think of playing here unless you know what you are doing."
	| newArray |
	newArray := Array new: 56.
	"Nil false and true get used throughout the interpreter"
	newArray at: 1 put: nil.
	newArray at: 2 put: false.
	newArray at: 3 put: true.
	"This association holds the active process (a ProcessScheduler)"
	newArray at: 4 put: (self associationAt: #Processor).
	"Numerous classes below used for type checking and instantiation"
	newArray at: 5 put: Bitmap.
	newArray at: 6 put: SmallInteger.
	newArray at: 7 put: ByteString.
	newArray at: 8 put: Array.
	newArray at: 9 put: Smalltalk.
	newArray at: 10 put: Float.
	newArray at: 11 put: MethodContext.
	newArray at: 12 put: BlockContext.
	newArray at: 13 put: Point.
	newArray at: 14 put: LargePositiveInteger.
	newArray at: 15 put: Display.
	newArray at: 16 put: Message.
	newArray at: 17 put: CompiledMethod.
	newArray at: 18 put: (self specialObjectsArray at: 18).
	"(low space Semaphore)"
	newArray at: 19 put: Semaphore.
	newArray at: 20 put: Character.
	newArray at: 21 put: #doesNotUnderstand:.
	newArray at: 22 put: #cannotReturn:.
	newArray at: 23 put: nil. "This is the process signalling low space."
	"An array of the 32 selectors that are compiled as special bytecodes,
	 paired alternately with the number of arguments each takes."
	newArray at: 24 put: #(	#+ 1 #- 1 #< 1 #> 1 #<= 1 #>= 1 #= 1 #~= 1
							#* 1 #/ 1 #\\ 1 #@ 1 #bitShift: 1 #// 1 #bitAnd: 1 #bitOr: 1
							#at: 1 #at:put: 2 #size 0 #next 0 #nextPut: 1 #atEnd 0 #== 1 #class 0
							#blockCopy: 1 #value 0 #value: 1 #do: 1 #new 0 #new: 1 #x 0 #y 0 ).
	"An array of the 255 Characters in ascii order.
	 Cog inlines table into machine code at: prim so do not regenerate it."
	newArray at: 25 put: (self specialObjectsArray at: 25).
	newArray at: 26 put: #mustBeBoolean.
	newArray at: 27 put: ByteArray.
	newArray at: 28 put: Process.
	"An array of up to 31 classes whose instances will have compact headers"
	newArray at: 29 put: self compactClassesArray.
	newArray at: 30 put: (self specialObjectsArray at: 30).
	"(delay Semaphore)"
	newArray at: 31 put: (self specialObjectsArray at: 31).
	"(user interrupt Semaphore)"
	"Prototype instances that can be copied for fast initialization"
	newArray at: 32 put: (Float new: 2).
	newArray at: 33 put: (LargePositiveInteger new: 4).
	newArray at: 34 put: Point new.
	newArray at: 35 put: #cannotInterpret:.
	"Note: This must be fixed once we start using context prototypes (yeah, right)"
	"(MethodContext new: CompiledMethod fullFrameSize)."
	newArray at: 36 put: (self specialObjectsArray at: 36). "Is the prototype MethodContext (unused by the VM)"
	newArray at: 37 put: BlockClosure.
	"(BlockContext new: CompiledMethod fullFrameSize)."
	newArray at: 38 put: (self specialObjectsArray at: 38). "Is the prototype BlockContext (unused by the VM)"
	"array of objects referred to by external code"
	newArray at: 39 put: (self specialObjectsArray at: 39).	"preserve external semaphores"
	newArray at: 40 put: nil. "Reserved for Mutex in Cog VMs"
	newArray at: 41 put: nil. "Reserved for a LinkedList instance for overlapped calls in CogMT"
	"finalization Semaphore"
	newArray at: 42 put: ((self specialObjectsArray at: 42) ifNil: [Semaphore new]).
	newArray at: 43 put: LargeNegativeInteger.
	"External objects for callout.
	 Note: Written so that one can actually completely remove the FFI."
	newArray at: 44 put: (self at: #ExternalAddress ifAbsent: []).
	newArray at: 45 put: (self at: #ExternalStructure ifAbsent: []).
	newArray at: 46 put: (self at: #ExternalData ifAbsent: []).
	newArray at: 47 put: (self at: #ExternalFunction ifAbsent: []).
	newArray at: 48 put: (self at: #ExternalLibrary ifAbsent: []).
	newArray at: 49 put: #aboutToReturn:through:.
	newArray at: 50 put: #run:with:in:.
	"51 reserved for immutability message"
	"newArray at: 51 put: #attemptToAssign:withIndex:."
	newArray at: 52 put: #(nil "nil => generic error" #'bad receiver'
							#'bad argument' #'bad index'
							#'bad number of arguments'
							#'inappropriate operation'  #'unsupported operation'
							#'no modification' #'insufficient object memory'
							#'insufficient C memory' #'not found' #'bad method'
							#'internal error in named primitive machinery'
							#'object may move').
	"53 to 55 are for Alien"
	newArray at: 53 put: (self at: #Alien ifAbsent: []).
	newArray at: 54 put: #invokeCallback:stack:registers:jmpbuf:.
	newArray at: 55 put: (self at: #UnsafeAlien ifAbsent: []).

	"Weak reference finalization"
	newArray at: 56 put: (self at: #WeakFinalizationList ifAbsent: []).

	"Now replace the interpreter's reference in one atomic operation"
	self specialObjectsArray become: newArray
	! !

!SmalltalkImage methodsFor: 'special objects' stamp: ''!
registerExternalObject: anObject
	"Register the given object in the external objects array and return its index. If it is already there, just return its index."

	^ExternalSemaphoreTable registerExternalObject: anObject! !

!SmalltalkImage methodsFor: 'housekeeping' stamp: ''!
removeAllLineFeeds
	"Smalltalk removeAllLineFeeds"
	"Scan all methods for source code with lineFeeds.
	Replaces all occurrences of <CR><LF> or <LF> by <CR>.
	When done, offers to display an Inspector containing the message
	names grouped by author initials.
	In this dictionary, the key 'OK' contains the methods that had literals that contained <LF> characters."
	| totalStripped totalOK authors |
	'Scanning sources for LineFeeds.
This will take a few minutes...'
		displayProgressAt: Sensor cursorPoint
		from: 0
		to: CompiledMethod instanceCount
		during: [:bar | | n | 
			n := 0.
			authors := self
						removeAllLineFeedsQuietlyCalling: [(n := n + 1) \\ 100 = 0
								ifTrue: [bar value: n]]].
	totalStripped := authors
				inject: 1
				into: [:sum :set | sum + set size].
	totalOK := (authors at: 'OK') size.
	totalStripped := totalStripped - totalOK.
	Transcript cr; show: totalStripped printString , ' methods stripped of LFs.'.
	Transcript cr; show: totalOK printString , ' methods still correctly contain LFs.'.
	(self confirm: 'Do you want to see the affected methods?')
		ifTrue: [authors inspect]! !

!SmalltalkImage methodsFor: 'housekeeping' stamp: ''!
removeAllLineFeedsQuietly
	"Smalltalk removeAllLineFeedsQuietly"
	"Scan all methods for source code with lineFeeds.
	Replaces all occurrences of <CR><LF> or <LF> by <CR>.
	Answer a Dictionary keyed by author name containing sets of affected method names,
	as well as (at the key 'OK') a list of methods that still contain LF characters inside literal strings or characters."
	^self removeAllLineFeedsQuietlyCalling: [].! !

!SmalltalkImage methodsFor: 'housekeeping' stamp: ''!
removeAllLineFeedsQuietlyCalling: aBlock
	"Smalltalk removeAllLineFeedsQuietly"
	"Scan all methods for source code with lineFeeds.
	Replaces all occurrences of <CR><LF> or <LF> by <CR>.
	Answer a Dictionary keyed by author name containing sets of affected method names,
	as well as (at the key 'OK') a list of methods that still contain LF characters inside literal strings or characters.
	Evaluate aBlock for each method so that status can be updated."
	| authors |
	authors := Dictionary new.
	authors at: 'OK' put: Set new.
	self systemNavigation
		allBehaviorsDo: [:cls | cls selectorsAndMethodsDo: [:selector :method |
					| oldStamp newCodeString oldCodeString oldCategory nameString | 
					aBlock cull: cls cull: selector.
					oldCodeString := cls sourceCodeAt: selector.
					(oldCodeString includes: Character lf)
						ifTrue: [
							newCodeString := oldCodeString withSqueakLineEndings.
							nameString := cls name , '>>' , selector.
							((cls compiledMethodAt: selector) hasLiteralSuchThat: [ :lit | lit asString includes: Character lf ])
								ifTrue: [(authors at: 'OK')
										add: nameString]
								ifFalse: [oldStamp := (method timeStamp)
												copy replaceAll: Character cr
												with: Character space.
									(authors
										at: (oldStamp copyFrom: 1 to: (oldStamp findFirst: [ :c | c isAlphaNumeric not ]))
										ifAbsentPut: [Set new])
										add: nameString.
									oldCategory := cls whichCategoryIncludesSelector: selector.
									cls
										compile: newCodeString
										classified: oldCategory
										withStamp: oldStamp
										notifying: nil ]]]].
	^ authors! !

!SmalltalkImage methodsFor: 'class and trait names' stamp: ''!
removeClassNamed: aName
	"Invoked from fileouts:  if there is currently a class in the system named aName, then remove it.  If anything untoward happens, report it in the Transcript.  "

	self globals removeClassNamed: aName! !

!SmalltalkImage methodsFor: 'housekeeping' stamp: ''!
removeEmptyMessageCategories
	"Smalltalk removeEmptyMessageCategories"
	self garbageCollect.
	(ClassOrganizer allInstances copyWith: SystemOrganization)
		do: [:org | org removeEmptyCategories]! !

!SmalltalkImage methodsFor: 'snapshot and quit' stamp: ''!
removeFromShutDownList: aClass

	ShutDownList remove: aClass name ifAbsent: []! !

!SmalltalkImage methodsFor: 'snapshot and quit' stamp: ''!
removeFromStartUpList: aClass

	StartUpList remove: aClass name ifAbsent: []! !

!SmalltalkImage methodsFor: 'to clean later' stamp: ''!
removeKey: key
	"Obsoleted."

	self flag: #tobeprecated.
	"self deprecated: 'Use Smalltalk globals'."
	^globals removeKey: key! !

!SmalltalkImage methodsFor: 'to clean later' stamp: ''!
removeKey: key ifAbsent: aBlock
	"Obsoleted."

	self flag: #tobeprecated.
	"self deprecated: 'Use Smalltalk globals'."
	^globals removeKey: key ifAbsent: aBlock! !

!SmalltalkImage methodsFor: 'shrinking' stamp: ''!
removeSelector: descriptor 
	"Safely remove a selector from a class (or metaclass). If the
	class or the method doesn't exist anymore, never mind and
	answer nil.
	This method should be used instead of 'Class removeSelector:
	#method' to omit global class references."
	| class sel |
	class := self
				at: descriptor first
				ifAbsent: [^ nil].
	(descriptor size > 2 and: [descriptor second == #class])
		ifTrue: 
			[class := class class.
			sel := descriptor third]
		ifFalse: 
			[sel := descriptor second].
	^ class removeSelector: sel! !

!SmalltalkImage methodsFor: 'class and trait names' stamp: ''!
renameClass: aClass as: newName 
	"Rename the class, aClass, to have the title newName."

	^globals renameClass: aClass as: newName! !

!SmalltalkImage methodsFor: 'class and trait names' stamp: ''!
renameClass: aClass from: oldName
	"Rename the class, aClass, to have the title newName."

	^globals renameClass: aClass from: oldName! !

!SmalltalkImage methodsFor: 'class and trait names' stamp: ''!
renameClassNamed: oldName as: newName
	"Invoked from fileouts:  if there is currently a class in the system named oldName, then rename it to newName.  If anything untoward happens, report it in the Transcript.  "

	| oldClass |
	(oldClass := self at: oldName asSymbol ifAbsent: [nil]) == nil
		ifTrue:
			[Transcript cr; show: 'Class-rename for ', oldName, ' ignored because ', oldName, ' does not exist.'.
			^ self].

	oldClass rename: newName! !

!SmalltalkImage methodsFor: 'class and trait names' stamp: ''!
renamedClass: aClass from: oldName to: newName
	"Inform Smalltalk that aClass has been renamed"
	
	(Array with: StartUpList with: ShutDownList) 
		do: [ :list | | index |  
			index := list indexOf: oldName ifAbsent: [0].
			index > 0 ifTrue: [list at: index put: newName]].
! !

!SmalltalkImage methodsFor: 'vm statistics' stamp: 'StephaneDucasse 2/16/2010 13:34'!
reportCPUandRAM
	"Write several text files with useful analysis for profiling purposes.
	Overwrites any existing report.
	SmalltalkImage current reportCPUandRAM
	"	

	| stream |
	"VM statistics (Memory use and GC, mainly)"
	FileStream 
		forceNewFileNamed: 'Log-MemoryStats.txt'
		do: [:aStream | aStream nextPutAll: self vmStatisticsReportString ].
	
	"Process list"
	stream := FileStream 
				forceNewFileNamed: 'Log-ProcessList.txt' 
				do: [ :aStream |  ProcessBrowser new processNameList 
								do: [ :each | aStream nextPutAll: each; cr ]].

"Fork all these, so they run in sequence, as the system is back running"
[
	"Process taking most CPU"
	stream := FileStream forceNewFileNamed: 'Log-ThePig.txt'.
	ProcessBrowser dumpPigStackOn: stream andClose: true.
	
	"Tally of all processes"
	stream := FileStream forceNewFileNamed: 'Log-FullTally.txt'.
	[ | tally |
		tally := MessageTally new.
		tally reportOtherProcesses: true.
		tally spyAllEvery: 1 on: [ (Delay forMilliseconds: 1000) wait ].
		tally report: stream ] ensure: [ stream close ].

	"Memory Analysis"
	stream := FileStream forceNewFileNamed: 'Log-MemoryAnalysis.txt'.
	[ SpaceTally new printSpaceAnalysis: 1 on: stream ]
		ensure: [ stream close ]

] forkNamed: 'CPU usage analysis'! !

!SmalltalkImage methodsFor: 'memory space' stamp: ''!
rootTable
	"Primitive. Answer a snapshot of the VMs root table. 
	Keep in mind that the primitive may itself cause GC."
	<primitive: 'primitiveRootTable'>
	^self primitiveFailed! !

!SmalltalkImage methodsFor: 'memory space' stamp: ''!
rootTableAt: index
	"Primitive. Answer the nth element of the VMs root table"
	<primitive: 'primitiveRootTableAt'>
	^nil! !

!SmalltalkImage methodsFor: 'saving' stamp: ''!
saveAs

	"Put up the 'saveAs' prompt, obtain a name, and save the image  under that new name."

	self saveAs: self getFileNameFromUser! !

!SmalltalkImage methodsFor: 'saving' stamp: ''!
saveAs: newName
	newName ifNil: [ ^ self ].
	(SourceFiles at: 2) ifNotNil:
		[self closeSourceFiles; "so copying the changes file will always work"
			 saveChangesInFileNamed: (self fullNameForChangesNamed: newName)].
	^self saveImageInFileNamed: (self fullNameForImageNamed: newName)! !

!SmalltalkImage methodsFor: 'saving' stamp: ''!
saveAsEmbeddedImage
	"Save the current state of the system as an embedded image"

	| dir newName newImageName oldImageSegDir haveSegs previousFullImageName |
	dir := FileDirectory default.
	previousFullImageName := self imageName.
	newName := UIManager default request: 'Select existing VM file'
				initialAnswer: (FileDirectory localNameFor: '').
	newName isEmptyOrNil ifTrue: [^Smalltalk].
	newName := FileDirectory baseNameFor: newName asFileName.
	newImageName := newName.
	(dir includesKey: newImageName) 
		ifFalse: 
			[^self 
				inform: 'Unable to find name ' , newName , ' Please choose another name.'].
	haveSegs := self areThereImageSegments.

	self logChange: '----SAVEAS (EMBEDDED) ' , newName , '----' 
				, Date dateAndTimeNow printString.
	self imageName: (dir fullNameFor: newImageName) asSqueakPathName.
	LastImageName := self imageName.
	self closeSourceFiles.
	self saveImageSegments: previousFullImageName.

	self 
		snapshot: true
		andQuit: true
		embedded: true! !

!SmalltalkImage methodsFor: 'saving' stamp: ''!
saveAsNewVersion
	"Save the image/changes using the next available version number."
	"SmalltalkImage current saveAsNewVersion"
	
	| newName changesName aName anIndex |
	aName := FileDirectory baseNameFor: (FileDirectory default localNameFor: self imageName).
	anIndex := aName lastIndexOf: FileDirectory extensionDelimiter ifAbsent: [nil].
	(anIndex notNil and: [(aName copyFrom: anIndex + 1 to: aName size) isAllDigits])
		ifTrue:
			[aName := aName copyFrom: 1 to: anIndex - 1].

	newName := FileDirectory default nextNameFor: aName extension: FileDirectory imageSuffix.
	changesName := self fullNameForChangesNamed: newName.

	"Check to see if there is a .changes file that would cause a problem if we saved a new .image file with the new version number"
	(FileDirectory default fileOrDirectoryExists: changesName)
		ifTrue:
			[^ self inform:
'There is already .changes file of the desired name,
', newName, '
curiously already present, even though there is
no corresponding .image file.   Please remedy
manually and then repeat your request.'].

	^self saveAs: newName


! !

!SmalltalkImage methodsFor: 'saving' stamp: ''!
saveChangesInFileNamed: aString
	| fullChangesName |
	fullChangesName := (FileDirectory default fullNameFor: aString).
	(FileDirectory default directoryNamed:(FileDirectory dirPathFor: fullChangesName )) assureExistence.
	FileDirectory default 
		copyFileWithoutOverwriteConfirmationNamed: Smalltalk changesName 
		toFileNamed: fullChangesName.
	self setMacFileInfoOn: fullChangesName.! !

!SmalltalkImage methodsFor: 'saving' stamp: ''!
saveImageInFileNamed: aString
	| fullImageName previousFullImageName |
	fullImageName := (FileDirectory default fullNameFor: aString).
	(FileDirectory default directoryNamed:(FileDirectory dirPathFor: fullImageName )) assureExistence.
	previousFullImageName := self imageName.
	^self
		changeImageNameTo: fullImageName;
		closeSourceFiles;
		openSourceFiles;  "so SNAPSHOT appears in new changes file"
		saveImageSegments: previousFullImageName;
		snapshot: true andQuit: false! !

!SmalltalkImage methodsFor: 'saving' stamp: ''!
saveImageSegments
	self saveImageSegments: self imageName! !

!SmalltalkImage methodsFor: 'saving' stamp: ''!
saveImageSegments: previousFullImageName
	"This method is used for example when you save an image with another name. In this case, if there are segments written to files and there are ImageSegment instances, then it copies the file from the previous image location, to the new one. It is like a backup of the segment files when renaming an image. "

	| haveSegs oldImageSegDir theClass newImageSegDir |
	"If the class does not exist or has no instances, it has no sense to backUp the segments."
	self areThereImageSegments
		ifFalse: [ ^ self ].
	theClass := Smalltalk globals at: #ImageSegment.
	oldImageSegDir := theClass segmentDirectoryFrom: previousFullImageName.
	newImageSegDir := theClass segmentDirectory.	"create the folder"
	oldImageSegDir fileNames
		do: [ :theName | 
			| imageSegmentName "copy all segment files" |
			imageSegmentName := oldImageSegDir pathName , FileDirectory slash , theName.
			newImageSegDir copyFileWithoutOverwriteConfirmationNamed: imageSegmentName toFileNamed: theName ]! !

!SmalltalkImage methodsFor: 'saving' stamp: 'StephaneDucasse 11/9/2010 18:55'!
saveSession
	"Save the current image without quitting and return it resuming status (whether it was restarted or in the same flow)."
	
	^ self snapshot: true andQuit: false! !

!SmalltalkImage methodsFor: 'deprecated' stamp: ''!
scopeFor: varName from: lower envtAndPathIfFound: envtAndPathBlock
	"Obsoleted."

	self flag: #tobeprecated.
	"self deprecated: 'Use Smalltalk globals'."
	(globals includesKey: varName)
		ifTrue: [^ envtAndPathBlock value: self value: String new]
		ifFalse: [^ nil]! !

!SmalltalkImage methodsFor: 'to clean later' stamp: ''!
select: aBlock
	"Obsoleted."
	self flag: #tobeprecated.
	"self deprecated: 'Use Smalltalk globals'."
	^globals select: aBlock
! !

!SmalltalkImage methodsFor: 'snapshot and quit' stamp: ''!
send: startUpOrShutDown toClassesNamedIn: startUpOrShutDownList with: argument
	"Send the message #startUp: or #shutDown: to each class named in the list.
	The argument indicates if the system is about to quit (for #shutDown:) or if
	the image is resuming (for #startUp:).
	If any name cannot be found, then remove it from the list."

	| removals |
	removals := OrderedCollection new.
	startUpOrShutDownList
		do: [ :name | 
			| class |
			class := self at: name ifAbsent: [ nil ].
			class == nil
				ifTrue: [ removals add: name ]
				ifFalse: [ 
							[ class perform: startUpOrShutDown with: argument ]
								on: Exception
								do: [ :ex | 
									Smalltalk hasDisplay
										ifTrue: [ ex pass ]
										ifFalse: [ 
											globals at: #Console ifPresent: [ :console | console printNl: ex description ].
											self logError: ex printString inContext: thisContext to: 'PharoDebug.log' ] ] ] ].	"Remove any obsolete entries, but after the iteration"
	startUpOrShutDownList removeAll: removals! !

!SmalltalkImage methodsFor: 'memory space' stamp: ''!
setGCBiasToGrow: aNumber
	"Primitive. Indicate that the GC logic should be bias to grow"
	<primitive: 'primitiveSetGCBiasToGrow'>
	^self primitiveFailed
"Example:
	Smalltalk setGCBiasToGrowGCLimit: 16*1024*1024.
	Smalltalk setGCBiasToGrow: 1.
"! !

!SmalltalkImage methodsFor: 'memory space' stamp: ''!
setGCBiasToGrowGCLimit: aNumber
	"Primitive. Indicate that the bias to grow logic should do a GC after aNumber Bytes"
	<primitive: 'primitiveSetGCBiasToGrowGCLimit'>
	^self primitiveFailed
"Example:
	Smalltalk setGCBiasToGrowGCLimit: 16*1024*1024.
"! !

!SmalltalkImage methodsFor: 'snapshot and quit' stamp: ''!
setGCParameters
	"Adjust the VM's default GC parameters to avoid premature tenuring."

	self vmParameterAt: 5 put: 4000.  "do an incremental GC after this many allocations"
	self vmParameterAt: 6 put: 2000.  "tenure when more than this many objects survive the GC"
	self setGCBiasToGrowGCLimit: 16*1024*1024.
	self  setGCBiasToGrow: 1.
! !

!SmalltalkImage methodsFor: 'memory space' stamp: ''!
setGCSemaphore: semaIndex
	"Primitive. Indicate the GC semaphore index to be signaled on GC occurance."
	<primitive: 'primitiveSetGCSemaphore'>
	^self primitiveFailed
"Example:

	| index sema process |
	sema := Semaphore new.
	index := Smalltalk registerExternalObject: sema.
	Smalltalk setGCSemaphore: index.
	process := [
		[[true] whileTrue:[
			sema wait.
			Smalltalk beep.
		]] ensure:[
			Smalltalk setGCSemaphore: 0.
			Smalltalk unregisterExternalObject: sema.
		].
	] fork.
	process inspect.
"! !

!SmalltalkImage methodsFor: 'miscellaneous' stamp: ''!
setMacFileInfoOn: aString
	"On Mac, set the file type and creator (noop on other platforms)"
	FileDirectory default
		setMacFileNamed: aString
		type: 'STch'
		creator: 'FAST'.! !

!SmalltalkImage methodsFor: 'image, changes names' stamp: 'StephaneDucasse 11/8/2010 21:56'!
shortImageName
	"Answer the current image name but only that name and not the full path as with imageName."
	"SmalltalkImage current shortImageName"

	^ FileDirectory baseNameFor: (FileDirectory default localNameFor: Smalltalk imageName).! !

!SmalltalkImage methodsFor: 'snapshot and quit' stamp: ''!
shutDown
	^ Smalltalk closeSourceFiles! !

!SmalltalkImage methodsFor: 'snapshot and quit' stamp: ''!
shutDownSound
	"No longer used in the release, but retained for backward compatibility."

	SoundService default shutDown
! !

!SmalltalkImage methodsFor: 'memory space' stamp: ''!
signalLowSpace
	"Signal the low-space semaphore to alert the user that space is running low."

	LowSpaceSemaphore signal.! !

!SmalltalkImage methodsFor: 'to clean later' stamp: ''!
size
	"Obsoleted."

	self flag: #tobeprecated.
	"self deprecated: 'Use Smalltalk globals'."
	^globals size! !

!SmalltalkImage methodsFor: 'snapshot and quit' stamp: ''!
snapshot: save andQuit: quit
	^self snapshot: save andQuit: quit embedded: false! !

!SmalltalkImage methodsFor: 'snapshot and quit' stamp: ''!
snapshot: save andQuit: quit embedded: embeddedFlag
	"Mark the changes file and close all files as part of #processShutdownList.
	If save is true, save the current state of this Smalltalk in the image file.
	If quit is true, then exit to the outer OS shell.
	The latter part of this method runs when resuming a previously saved image. This resume logic checks for a document file to process when starting up."

	| resuming msg |
	Object flushDependents.
	Object flushEvents.
	(SourceFiles at: 2)
		ifNotNil: [ 
			msg := String
				streamContents: [ :s | 
					s
						nextPutAll: '----';
						nextPutAll:
								(save
										ifTrue: [ 
											quit
												ifTrue: [ 'QUIT' ]
												ifFalse: [ 'SNAPSHOT' ] ]
										ifFalse: [ 
											quit
												ifTrue: [ 'QUIT/NOSAVE' ]
												ifFalse: [ 'NOP' ] ]);
						nextPutAll: '----';
						print: Date dateAndTimeNow;
						space;
						nextPutAll: (FileDirectory default localNameFor: self imageName);
						nextPutAll: ' priorSource: ';
						print: LastQuitLogPosition ].
			self assureStartupStampLogged.
			save
				ifTrue: [ 
					LastQuitLogPosition := (SourceFiles at: 2)
						setToEnd;
						position ].
			self logChange: msg.
			Transcript
				cr;
				show: msg ].
	self processShutDownList: quit.
	Cursor write show.
	save
		ifTrue: [ 
			resuming := embeddedFlag
				ifTrue: [ self snapshotEmbeddedPrimitive ]
				ifFalse: [ self snapshotPrimitive ].	"<-- PC frozen here on image file"
			resuming == false
				ifTrue: [ 
					"Time to reclaim segment files is immediately after a save"
					Smalltalk globals at: #ImageSegment ifPresent: [ :theClass | theClass reclaimObsoleteSegmentFiles ] ]	"guard against failure" ]
		ifFalse: [ resuming := false ].
	quit & (resuming == false)
		ifTrue: [ self quitPrimitive ].
	Cursor normal show.
	self setGCParameters.
	resuming == true
		ifTrue: [ Smalltalk clearExternalObjects ].
	self processStartUpList: resuming == true.
	resuming == true
		ifTrue: [ self recordStartupStamp ].
	UIManager default onSnapshot.	"Now it's time to raise an error"
	resuming == nil
		ifTrue: [ self error: 'Failed to write image file (disk full?)' ].
	^ resuming! !

!SmalltalkImage methodsFor: 'snapshot and quit' stamp: ''!
snapshotEmbeddedPrimitive
	"This is from experimental code to create single-file applications. 
	It saves a copy of the VM and append an image."
	<primitive: 247>
	^nil "indicates error writing embedded image file"! !

!SmalltalkImage methodsFor: 'snapshot and quit' stamp: ''!
snapshotPrimitive
	"Primitive. Write the current state of the object memory on a file in the
	same format as the Smalltalk-80 release. The file can later be resumed,
	returning you to this exact state. Return normally after writing the file.
	Essential. See Object documentation whatIsAPrimitive."

	<primitive: 97>
	^nil "indicates error writing image file"! !

!SmalltalkImage methodsFor: 'image, changes names' stamp: ''!
sourceFileVersionString

	^ SourceFileVersionString! !

!SmalltalkImage methodsFor: 'private' stamp: ''!
sourceFileVersionString: aString

	SourceFileVersionString := aString! !

!SmalltalkImage methodsFor: 'image, changes names' stamp: ''!
sourcesName
	"Answer the full path to the version-stable source code"
	^ self vmPath , SourceFileVersionString , FileDirectory dot , 'sources'! !

!SmalltalkImage methodsFor: 'special objects' stamp: ''!
specialNargsAt: anInteger 
	"Answer the number of arguments for the special selector at: anInteger."

	^ (self specialObjectsArray at: 24) at: anInteger * 2! !

!SmalltalkImage methodsFor: 'special objects' stamp: ''!
specialObjectsArray  "Smalltalk specialObjectsArray at: 1"
	<primitive: 129>
	^ self primitiveFailed! !

!SmalltalkImage methodsFor: 'special objects' stamp: ''!
specialSelectorAt: anInteger 
	"Answer the special message selector from the interleaved specialSelectors array."

	^ (self specialObjectsArray at: 24) at: anInteger * 2 - 1! !

!SmalltalkImage methodsFor: 'special objects' stamp: ''!
specialSelectorSize
	"Answer the number of special selectors in the system."

	^ (self specialObjectsArray at: 24) size // 2! !

!SmalltalkImage methodsFor: 'special objects' stamp: ''!
specialSelectors
	"Used by SystemTracer only."

	^SpecialSelectors! !

!SmalltalkImage methodsFor: 'objects from disk' stamp: 'tk 3/7/2000 18:40'!
storeDataOn: aDataStream
	"I don't get stored.  Use a DiskProxy"

	self error: 'use a DiskProxy to store me'! !

!SmalltalkImage methodsFor: 'sources, change log' stamp: ''!
systemInformationString
	^ String cr, SystemVersion current version, String cr, self lastUpdateString, String cr, String cr, String cr, self licenseString! !

!SmalltalkImage methodsFor: 'housekeeping' stamp: ''!
testFormatter
	"Smalltalk testFormatter"

	"Reformats the source for every method in the system, and
	then compiles that source and verifies that it generates
	identical code"

	| badOnes |
	badOnes := OrderedCollection new.
	'Formatting all classes...' 
		displayProgressAt: Sensor cursorPoint
		from: 0
		to: CompiledMethod instanceCount
		during: 
			[:bar | | n | 
			n := 0.
			self systemNavigation allBehaviorsDo: 
					[:cls | 
					"Transcript cr; show: cls name."

					cls selectorsAndMethodsDo: 
							[:selector :oldMethod |
							| newCodeString newMethod methodNode | 
							(n := n + 1) \\ 100 = 0 ifTrue: [bar value: n].
							newCodeString := cls prettyPrinterClass 
										format: (cls sourceCodeAt: selector)
										in: cls
										notifying: nil.
							methodNode := cls compilerClass new 
										compile: newCodeString
										in: cls
										notifying: nil
										ifFail: [].
							newMethod := methodNode generate.
							oldMethod = newMethod 
								ifFalse: 
									[Transcript
										cr;
										show: '***' , cls name , ' ' , selector.
									badOnes add: cls name , ' ' , selector]]]].
	self systemNavigation browseMessageList: badOnes asSortedCollection
		name: 'Formatter Discrepancies'! !

!SmalltalkImage methodsFor: 'housekeeping' stamp: ''!
testFormatter2
	"Smalltalk testFormatter2"

	"Reformats the source for every method in the system, and
	then verifies that the order of source tokens is unchanged."

	| badOnes |
	badOnes := OrderedCollection new.
	'Formatting all classes...' 
		displayProgressAt: Sensor cursorPoint
		from: 0
		to: CompiledMethod instanceCount
		during: 
			[:bar | | n | 
			n := 0.
			self systemNavigation allBehaviorsDo: 
					[:cls | 
					"Transcript cr; show: cls name."

					cls selectorsDo: 
							[:selector | | newCodeString oldTokens oldCodeString newTokens | 
							(n := n + 1) \\ 100 = 0 ifTrue: [bar value: n].
							oldCodeString := (cls sourceCodeAt: selector) asString.
							newCodeString := cls prettyPrinterClass 
										format: oldCodeString
										in: cls
										notifying: nil.
							oldTokens := oldCodeString findTokens: Character separators.
							newTokens := newCodeString findTokens: Character separators.
							oldTokens = newTokens 
								ifFalse: 
									[Transcript
										cr;
										show: '***' , cls name , ' ' , selector.
									badOnes add: cls name , ' ' , selector]]]].
	self systemNavigation browseMessageList: badOnes asSortedCollection
		name: 'Formatter Discrepancies'! !

!SmalltalkImage methodsFor: 'vm statistics' stamp: ''!
textMarkerForShortReport

	^  'Since last view	'! !

!SmalltalkImage methodsFor: 'sources, change log' stamp: ''!
timeStamp: aStream 
	"Writes system version and current time on stream aStream."

	| dateTime |
	dateTime := Time dateAndTimeNow.
	aStream nextPutAll: 'From ', self datedVersion, ' [', self  lastUpdateString, '] on ', (dateTime at: 1) printString,
						' at ', (dateTime at: 2) printString! !

!SmalltalkImage methodsFor: 'class and trait names' stamp: ''!
traitNames
	"Answer a SortedCollection of all traits (not including class-traits) names."
	
	^ globals traitNames! !

!SmalltalkImage methodsFor: 'external' stamp: ''!
unbindExternalPrimitives
	"Primitive. Force all external primitives to be looked up again afterwards. Since external primitives that have not found are bound for fast failure this method will force the lookup of all primitives again so that after adding some plugin the primitives may be found."
	<primitive: 570>
	"Do nothing if the primitive fails for compatibility with older VMs"
! !

!SmalltalkImage methodsFor: 'modules' stamp: ''!
unloadModule: aString
	"Primitive. Unload the given module.
	This primitive is intended for development only since some
	platform do not implement unloading of DLL's accordingly.
	Also, the mechanism for unloading may not be supported
	on all platforms."
	<primitive: 571>
	^self primitiveFailed! !

!SmalltalkImage methodsFor: 'special objects' stamp: ''!
unregisterExternalObject: anObject
	"Unregister the given object in the external objects array. Do nothing if it isn't registered."

	ExternalSemaphoreTable unregisterExternalObject: anObject! !

!SmalltalkImage methodsFor: 'shrinking' stamp: ''!
unusedClasses
	"Enumerates all classes in the system and returns a list of those that are 
	apparently unused. A class is considered in use if it (a) has subclasses 
	or (b) is referred to by some method or (c) has its name in use as a 
	literal. "
	"Smalltalk unusedClasses asSortedCollection"
	^ self systemNavigation allUnusedClassesWithout: {{}. {}}! !

!SmalltalkImage methodsFor: 'memory space' stamp: ''!
useUpMemory
	"For testing the low space handler..."
	"Smalltalk installLowSpaceWatcher; useUpMemory"

	| lst |
	lst := nil.
	[true] whileTrue: [
		lst := Link nextLink: lst.
	].! !

!SmalltalkImage methodsFor: 'memory space' stamp: ''!
useUpMemoryWithArrays 
	"For testing the low space handler..."
	"Smalltalk installLowSpaceWatcher; useUpMemoryWithArrays"

	| b |  "First use up most of memory."
	b := String new: self bytesLeft - self lowSpaceThreshold - 100000.
	b := b.  "Avoid unused value warning"
	(1 to: 10000) collect: [:i | Array new: 10000]! !

!SmalltalkImage methodsFor: 'memory space' stamp: ''!
useUpMemoryWithContexts 
	"For testing the low space handler..."
	"Smalltalk installLowSpaceWatcher; useUpMemoryWithContexts"

	self useUpMemoryWithContexts! !

!SmalltalkImage methodsFor: 'memory space' stamp: ''!
useUpMemoryWithTinyObjects 
	"For testing the low space handler..."
	"Smalltalk installLowSpaceWatcher; useUpMemoryWithTinyObjects"

	| b |  "First use up most of memory."
	b := String new: self bytesLeft - self lowSpaceThreshold - 100000.
	b := b.  "Avoid unused value warning"
	(1 to: 10000) collect: [:i | BitBlt new]! !

!SmalltalkImage methodsFor: 'housekeeping' stamp: ''!
verifyChanges		
	"Smalltalk verifyChanges"
	"Recompile all methods in the changes file."
	self systemNavigation allBehaviorsDo: [:class | class recompileChanges].
! !

!SmalltalkImage methodsFor: 'sources, change log' stamp: ''!
version
	"Answer the version of this release."

	^SystemVersion current version! !

!SmalltalkImage methodsFor: 'vm' stamp: ''!
versionLabel
	"Answer a string corresponding to the version of virtual machine. This
	represents the version level of the Smalltalk source code (interpreter
	and various plugins) that is translated to C by a CCodeGenerator,  in
	addition to the external platform source code, typically written in C and
	managed separately for each platform.
	
	This invokes an optional primitive that may not be available on all virtual
	machines. See also vmVersion, which answers a string identifying the image
	from which virtual machine sources were generated."

	"Smalltalk vm versionLabel"

	<primitive: 'primitiveVMVersion'>
	self notify: 'This virtual machine does not support the optional primitive #primitiveVMVersion' translated.
	^''! !

!SmalltalkImage methodsFor: 'copying' stamp: ''!
veryDeepCopyWith: deepCopier
	"Return self.  I can't be copied.  Do not record me."! !

!SmalltalkImage methodsFor: 'vm' stamp: ''!
vm
	"Answer the object to query about virtual machine."
	
	^self! !

!SmalltalkImage methodsFor: 'system attribute' stamp: ''!
vmFullPath
	"Return the full vm path as a string"
	"SmalltalkImage current vmFullPath"

	^self getSystemAttribute: 0! !

!SmalltalkImage methodsFor: 'vm parameters' stamp: ''!
vmParameterAt: parameterIndex
	"parameterIndex is a positive integer corresponding to one of the VM's internal
	parameter/metric registers.  Answer with the current value of that register.
	Fail if parameterIndex has no corresponding register.
	VM parameters are numbered as follows:
		1	end of old-space (0-based, read-only)
		2	end of young-space (read-only)
		3	end of memory (read-only)
		4	allocationCount (read-only)
		5	allocations between GCs (read-write)
		6	survivor count tenuring threshold (read-write)
		7	full GCs since startup (read-only)
		8	total milliseconds in full GCs since startup (read-only)
		9	incremental GCs since startup (read-only)
		10	total milliseconds in incremental GCs since startup (read-only)
		11	tenures of surving objects since startup (read-only)
		12-20 specific to the translating VM
		21	root table size (read-only)
		22	root table overflows since startup (read-only)
		23	bytes of extra memory to reserve for VM buffers, plugins, etc.

		24	memory threshold above which shrinking object memory (rw)
		25	memory headroom when growing object memory (rw)
		26  interruptChecksEveryNms - force an ioProcessEvents every N milliseconds, in case the image  is not calling getNextEvent often (rw)
		27	number of times mark loop iterated for current IGC/FGC (read-only) includes ALL marking
		28	number of times sweep loop iterated  for current IGC/FGC (read-only)
		29	number of times make forward loop iterated for current IGC/FGC (read-only)
		30	number of times compact move loop iterated for current IGC/FGC (read-only)
		31	number of grow memory requests (read-only)
		32	number of shrink memory requests (read-only)
		33	number of root table entries used for current IGC/FGC (read-only)
		34	number of allocations done before current IGC/FGC (read-only)
		35	number of survivor objects after current IGC/FGC (read-only)
		36  millisecond clock when current IGC/FGC completed (read-only)
		37  number of marked objects for Roots of the world, not including Root Table entries for current IGC/FGC (read-only)
		38  milliseconds taken by current IGC  (read-only)
		39  Number of finalization signals for Weak Objects pending when current IGC/FGC completed (read-only)
		40  VM word size - 4 or 8 (read-only)"

	<primitive: 254>
	self primitiveFailed! !

!SmalltalkImage methodsFor: 'vm parameters' stamp: ''!
vmParameterAt: parameterIndex put: newValue
	"parameterIndex is a positive integer corresponding to one of the VM's internal
	parameter/metric registers.  Store newValue (a positive integer) into that
	register and answer with the previous value that was stored there.
	Fail if newValue is out of range, if parameterIndex has no corresponding
	register, or if the corresponding register is read-only."

	<primitive: 254>
	self primitiveFailed! !

!SmalltalkImage methodsFor: 'image, changes names' stamp: ''!
vmPath
	"Answer the path for the directory containing the Smalltalk virtual machine. Return the 	empty string if this primitive is not implemented."
	"SmalltalkImage current vmPath"

	^ (FilePath pathName: (self primVmPath) isEncoded: true) asSqueakPathName.
! !

!SmalltalkImage methodsFor: 'vm statistics' stamp: ''!
vmStatisticsReportString
	"StringHolderView open: (StringHolder new contents:
		SmalltalkImage current vmStatisticsReportString) label: 'VM Statistics'"

	| params oldSpaceEnd youngSpaceEnd memoryEnd fullGCs fullGCTime incrGCs incrGCTime tenureCount mcMisses mcHits icHits upTime sendCount upTime2 fullGCs2 fullGCTime2 incrGCs2 incrGCTime2 tenureCount2 str |
	params := self getVMParameters.
	oldSpaceEnd			:= params at: 1.
	youngSpaceEnd		:= params at: 2.
	memoryEnd			:= params at: 3.
	fullGCs				:= params at: 7.
	fullGCTime			:= params at: 8.
	incrGCs				:= params at: 9.
	incrGCTime			:= params at: 10.
	tenureCount			:= params at: 11.
	upTime := Time millisecondClockValue.
	str := (String new: 1000) writeStream.
	str	nextPutAll: 'uptime			';
		print: (upTime / 1000 / 60 // 60); nextPut: $h;
		print: (upTime / 1000 / 60 \\ 60) asInteger; nextPut: $m;
		print: (upTime / 1000 \\ 60) asInteger; nextPut: $s; cr.

	str	nextPutAll: 'memory			';
		nextPutAll: memoryEnd asStringWithCommas; nextPutAll: ' bytes'; cr.
	str	nextPutAll:	'	old			';
		nextPutAll: oldSpaceEnd asStringWithCommas; nextPutAll: ' bytes (';
		print: ((oldSpaceEnd / memoryEnd * 100) roundTo: 0.1); nextPutAll: '%)'; cr.
	str	nextPutAll: '	young		';
		nextPutAll: (youngSpaceEnd - oldSpaceEnd) asStringWithCommas; nextPutAll: ' bytes (';
		print: ((youngSpaceEnd - oldSpaceEnd / memoryEnd * 100) roundTo: 0.1); nextPutAll: '%)'; cr.
	str	nextPutAll: '	used		';
		nextPutAll: youngSpaceEnd asStringWithCommas; nextPutAll: ' bytes (';
		print: ((youngSpaceEnd / memoryEnd * 100) roundTo: 0.1); nextPutAll: '%)'; cr.
	str	nextPutAll: '	free		';
		nextPutAll: (memoryEnd - youngSpaceEnd) asStringWithCommas; nextPutAll: ' bytes (';
		print: ((memoryEnd - youngSpaceEnd / memoryEnd * 100) roundTo: 0.1); nextPutAll: '%)'; cr.

	str	nextPutAll: 'GCs				';
		nextPutAll: (fullGCs + incrGCs) asStringWithCommas.
	fullGCs + incrGCs > 0 ifTrue: [
		str
			nextPutAll: ' ('; 
			print: ((upTime / (fullGCs + incrGCs)) roundTo: 1); 
			nextPutAll: 'ms between GCs)'
	].
	str cr.
	str	nextPutAll: '	full			';
		print: fullGCs; nextPutAll: ' totalling '; nextPutAll: fullGCTime asStringWithCommas; nextPutAll: 'ms (';
		print: ((fullGCTime / upTime * 100) roundTo: 0.1);
		nextPutAll: '% uptime)'.
	fullGCs = 0 ifFalse:
		[str	nextPutAll: ', avg '; print: ((fullGCTime / fullGCs) roundTo: 0.1); nextPutAll: 'ms'].
	str	cr.
	str	nextPutAll: '	incr		';
		print: incrGCs; nextPutAll: ' totalling '; nextPutAll: incrGCTime asStringWithCommas; nextPutAll: 'ms (';
		print: ((incrGCTime / upTime * 100) roundTo: 0.1);
		nextPutAll: '% uptime), avg '; print: ((incrGCTime / incrGCs) roundTo: 0.1); nextPutAll: 'ms'; cr.
	str	nextPutAll: '	tenures		';
		nextPutAll: tenureCount asStringWithCommas.
	tenureCount = 0 ifFalse:
		[str nextPutAll: ' (avg '; print: (incrGCs / tenureCount) asInteger; nextPutAll: ' GCs/tenure)'].
	str	cr.

LastStats ifNil: [LastStats := Array new: 6]
ifNotNil: [
	upTime2 := upTime - (LastStats at: 1).
	fullGCs2 := fullGCs - (LastStats at: 2).
	fullGCTime2 := fullGCTime - (LastStats at: 3).
	incrGCs2 := incrGCs - (LastStats at: 4).
	incrGCTime2 := incrGCTime - (LastStats at: 5).
	tenureCount2 := tenureCount - (LastStats at: 6).

	str	nextPutAll: self textMarkerForShortReport ;
		nextPutAll: (fullGCs2 + incrGCs2) asStringWithCommas.
	fullGCs2 + incrGCs2 > 0 ifTrue: [
		str
			nextPutAll: ' ('; 
			print: ((upTime2 / (fullGCs2 + incrGCs2)) roundTo: 1); 
			nextPutAll: 'ms between GCs)'.
	].
	str cr.
	str	nextPutAll: '	uptime		'; print: ((upTime2 / 1000.0) roundTo: 0.1); nextPutAll: 's'; cr.
	str	nextPutAll: '	full			';
		print: fullGCs2; nextPutAll: ' totalling '; nextPutAll: fullGCTime2 asStringWithCommas; nextPutAll: 'ms (';
		print: ((fullGCTime2 / upTime2 * 100) roundTo: 0.1);
		nextPutAll: '% uptime)'.
	fullGCs2 = 0 ifFalse:
		[str	nextPutAll: ', avg '; print: ((fullGCTime2 / fullGCs2) roundTo: 0.1); nextPutAll: 'ms'].
	str	cr.
	str	nextPutAll: '	incr		';
		print: incrGCs2; nextPutAll: ' totalling '; nextPutAll: incrGCTime2 asStringWithCommas; nextPutAll: 'ms (';
		print: ((incrGCTime2 / upTime2 * 100) roundTo: 0.1);
		nextPutAll: '% uptime), avg '.
	incrGCs2 > 0 ifTrue: [
		 str print: ((incrGCTime2 / incrGCs2) roundTo: 0.1); nextPutAll: 'ms'
	].
	str cr.
	str	nextPutAll: '	tenures		';
		nextPutAll: tenureCount2 asStringWithCommas.
	tenureCount2 = 0 ifFalse:
		[str nextPutAll: ' (avg '; print: (incrGCs2 / tenureCount2) asInteger; nextPutAll: ' GCs/tenure)'].
	str	cr.
].
	LastStats at: 1 put: upTime.
	LastStats at: 2 put: fullGCs.
	LastStats at: 3 put: fullGCTime.
	LastStats at: 4 put: incrGCs.
	LastStats at: 5 put: incrGCTime.
	LastStats at: 6 put: tenureCount.

	^ str contents
! !

!SmalltalkImage methodsFor: 'vm statistics' stamp: ''!
vmStatisticsShortString
	"Convenience item for access to recent statistics only"
	"StringHolderView open: (StringHolder new contents: SmalltalkImage current vmStatisticsShortString)
		label: 'VM Recent Statistics'"
	^ self vmStatisticsReportString readStream
		upToAll: 'Since';
		nextLine;
		upToEnd! !

!SmalltalkImage methodsFor: 'system attribute' stamp: ''!
vmVersion	
	"Return a string identifying the interpreter version"
	"SmalltalkImage current vmVersion"

	^self getSystemAttribute: 1004! !

!SmalltalkImage methodsFor: 'os' stamp: ''!
windowSystemName			
	"Return the name of the window system currently being used for display."
	"Smalltalk os windowSystemName"

	self deprecated: 'Use OSPlatform'.
	^ OSPlatform windowSystemName ! !

!SmalltalkImage methodsFor: 'image' stamp: ''!
wordSize
	"Answer the size in bytes of an object pointer or word in the object memory.
	The value does not change for a given image, but may be modified by a SystemTracer
	when converting the image to another format. The value is cached in WordSize to
	avoid the performance overhead of repeatedly consulting the VM."

	"Smalltalk wordSize"

	^ WordSize ifNil: [WordSize := [SmalltalkImage current vmParameterAt: 40] on: Error do: [4]]! !

!SmalltalkImage methodsFor: 'sources, change log' stamp: ''!
writeRecentCharacters: nCharacters toFileNamed: aFilename
	"Schedule an editable text view on the last n characters of changes."
	| changes |
	changes := SourceFiles at: 2.
	changes setToEnd; skip: nCharacters negated.
	(FileStream newFileNamed: aFilename) nextPutAll: (changes next: nCharacters); close; open; edit! !

Object subclass: #SourceReference
	instanceVariableNames: 'timeStamp timeStampString sourcePointer'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'System-Support'!

SourceReference subclass: #CommentReference
	instanceVariableNames: 'class classSymbol stringVersion'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'System-Support'!

!CommentReference commentStamp: 'Ben 10/23/2010 14:53' prior: 0!
A CommentReference is xxxxxxxxx.

dfg

Instance Variables
	class:		<Object>
	classSymbol:		<Object>
	stringVersion:		<Object>

class
	- xxxxx

classSymbol
	- xxxxx

stringVersion
	- xxxxx
!

!CommentReference class methodsFor: 'instance creation' stamp: ''!
class: aClass
	^ self new class: aClass! !

!CommentReference methodsFor: 'accessing' stamp: ''!
actualClass
	^ class! !

!CommentReference methodsFor: 'accessing' stamp: 'BenjaminVanRyseghem 11/2/2010 16:52'!
browse

	ToolSet browse: self actualClass selector: nil! !

!CommentReference methodsFor: 'initialization' stamp: 'JohanBrichau 12/30/2010 12:47'!
class: aClass
	classSymbol := aClass theNonMetaClass name.
	sourcePointer := aClass organization commentRemoteStr sourcePointer.
	stringVersion := aClass name , ' ' , #comment.
	class  := aClass.! !

!CommentReference methodsFor: 'accessing' stamp: ''!
classSymbol 
	^ classSymbol! !

!CommentReference methodsFor: 'accessing' stamp: 'Sd 12/8/2010 21:09'!
classSymbol: aSymbol
	
	classSymbol := aSymbol! !

!CommentReference methodsFor: 'testing' stamp: 'JohanBrichau 1/21/2011 12:08'!
isCommentReference 
	^ true! !

!CommentReference methodsFor: 'accessing' stamp: 'JohanBrichau 1/21/2011 12:08'!
methodSymbol 
	self flag: 'do not want that here :-)'.
	^ #Comment! !

!CommentReference methodsFor: 'private' stamp: ''!
preambleString 
	^ 'commentStamp:'! !

!CommentReference methodsFor: 'printing' stamp: ''!
printOn: aStream
	aStream 
		nextPutAll: self classSymbol asString;
		nextPutAll: ' >> #';
		nextPutAll: 'class comment'.! !

!CommentReference methodsFor: 'printing' stamp: ''!
printStringForRML
	^'Comment'! !

!CommentReference methodsFor: 'polymorphism' stamp: 'BenjaminVanRyseghem 11/29/2010 11:40'!
selector
	^#Comment! !

!CommentReference methodsFor: 'private' stamp: ''!
stampSymbol
	^ self preambleString asSymbol! !

!CommentReference methodsFor: 'accessing' stamp: ''!
stringVersion
	^ stringVersion! !

SourceReference subclass: #MethodReference
	instanceVariableNames: 'classIsMeta classSymbol methodSymbol stringVersion category package'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'System-Support'!

!MethodReference commentStamp: 'Ben 10/23/2010 15:13' prior: 0!
A MethodReference is is a lightweight proxy for a CompiledMethod.  Has methods for pointed to the CompileMethod's source statements, byte codes. Is heavily used my Tools.

Instance Variables
	classIsMeta:		     Boolean class vs. instance
	classSymbol:		Symbol for method's class (without class keyword if meta)
	methodSymbol:		Symbol for method's selector
	stringVersion:		'Class>>selector:' format

!

!MethodReference class methodsFor: 'instance creation' stamp: ''!
class: aClass selector: aSelector
	"Changed to support inheritance"
	^ self new setStandardClass: aClass methodSymbol: aSelector! !

!MethodReference class methodsFor: 'Package' stamp: ''!
class: aClass selector: aSelector package: aPackage

	^ (self class: aClass selector: aSelector ) package: aPackage ; yourself! !

!MethodReference methodsFor: 'comparisons' stamp: ''!
<= anotherMethodReference

	classSymbol < anotherMethodReference classSymbol ifTrue: [^true].
	classSymbol > anotherMethodReference classSymbol ifTrue: [^false].
	classIsMeta = anotherMethodReference classIsMeta ifFalse: [^classIsMeta not].
	^methodSymbol <= anotherMethodReference methodSymbol
! !

!MethodReference methodsFor: 'comparisons' stamp: ''!
= anotherMethodReference 
	"Answer whether the receiver and the argument represent the 
	same object."
	^ self species == anotherMethodReference species
		and: [(self classSymbol = anotherMethodReference classSymbol)
				and: [(self classIsMeta = anotherMethodReference classIsMeta)
						and: [self methodSymbol = anotherMethodReference methodSymbol]]]! !

!MethodReference methodsFor: 'accessing' stamp: ''!
actualClass
	| actualClass traitName |
	('*classTrait' match: classSymbol)
		ifTrue: [ 
			traitName := classSymbol copyUpTo: Character space.
			^ Smalltalk globals at: traitName asSymbol ifAbsent: [ nil ] ].
	actualClass := Smalltalk globals at: classSymbol ifAbsent: [ ^ nil ].
	classIsMeta
		ifTrue: [ ^ actualClass classSide ].
	^ actualClass! !

!MethodReference methodsFor: 'queries' stamp: ''!
asStringOrText
	| text |
	self actualClass ifNil: [ ^stringVersion ].
	text := Text fromString: self stringVersion,' {',self category asString,'} '.
	self isLocalSelector ifFalse: [ text addAttribute: TextEmphasis italic ].
	^ text! !

!MethodReference methodsFor: 'queries' stamp: ''!
category
	^ category ifNil: [category := self actualClass organization categoryOfElement: methodSymbol]! !

!MethodReference methodsFor: 'string version' stamp: ''!
category: aString
	
	 category := aString! !

!MethodReference methodsFor: 'queries' stamp: ''!
classIsMeta

	^classIsMeta! !

!MethodReference methodsFor: 'queries' stamp: ''!
classSymbol

	^classSymbol! !

!MethodReference methodsFor: 'accessing' stamp: 'BenjaminVanRyseghem 11/30/2010 13:55'!
classSymbol: aSymbol

	classSymbol := aSymbol! !

!MethodReference methodsFor: 'accessing' stamp: ''!
compiledMethod
	^ self actualClass compiledMethodAt: methodSymbol ! !

!MethodReference methodsFor: 'comparisons' stamp: ''!
hash
	"Answer a SmallInteger whose value is related to the receiver's  
	identity."
	^ (self species hash bitXor: self classSymbol hash)
		bitXor: self methodSymbol hash! !

!MethodReference methodsFor: 'testing' stamp: ''!
isLocalSelector
	^self actualClass
		includesLocalSelector: self methodSymbol! !

!MethodReference methodsFor: 'queries' stamp: ''!
isValid
	"Answer whether the receiver represents a current selector or Comment"

	| aClass |
	methodSymbol isDoIt ifTrue: [^ false].
	(aClass := self actualClass) ifNil: [^ false].
	^ (aClass includesSelector: methodSymbol) or:
		[methodSymbol == #Comment]! !

!MethodReference methodsFor: 'accessing' stamp: ''!
methodClass

	^ self actualClass ! !

!MethodReference methodsFor: 'queries' stamp: ''!
methodSymbol

	^methodSymbol! !

!MethodReference methodsFor: 'accessing' stamp: ''!
package
	"we will have to change that in the future since we will not rely on * convention"
	
	^ package! !

!MethodReference methodsFor: 'accessing' stamp: ''!
package: aPackage
	
	package := aPackage! !

!MethodReference methodsFor: 'queries' stamp: ''!
printOn: aStream
	"Print the receiver on a stream"

	super printOn: aStream.
	aStream nextPutAll: ' ', self actualClass name, ' >> #', methodSymbol! !

!MethodReference methodsFor: 'accessing' stamp: ''!
selector

	^ methodSymbol! !

!MethodReference methodsFor: 'setting' stamp: ''!
setClass: aClass methodSymbol: methodSym

	classSymbol := aClass theNonMetaClass name.
	classIsMeta := aClass isMeta.
	methodSymbol := methodSym.
	stringVersion := aClass name , ' ' , methodSym.! !

!MethodReference methodsFor: 'setting' stamp: ''!
setClass: aClass methodSymbol: methodSym stringVersion: aString 

	classSymbol := aClass theNonMetaClass name.
	classIsMeta := aClass isMeta.
	methodSymbol := methodSym.
	stringVersion := aString.! !

!MethodReference methodsFor: 'setting' stamp: ''!
setClassAndSelectorIn: csBlock

	^csBlock value: self actualClass value: methodSymbol! !

!MethodReference methodsFor: 'setting' stamp: ''!
setClassSymbol: classSym classIsMeta: isMeta methodSymbol: methodSym stringVersion: aString 

	classSymbol := classSym.
	classIsMeta := isMeta.
	methodSymbol := methodSym.
	stringVersion := aString.! !

!MethodReference methodsFor: 'setting' stamp: 'Sd 11/2/2010 12:49'!
setStandardClass: aClass methodSymbol: methodSym

	classSymbol := aClass theNonMetaClass name.
	classIsMeta := aClass isMeta.
	methodSymbol := methodSym.
	stringVersion := aClass name , ' ' , methodSym.! !

!MethodReference methodsFor: 'accessing' stamp: ''!
source
	^ (self actualClass sourceCodeAt: methodSymbol) asString withSqueakLineEndings! !

!MethodReference methodsFor: 'queries' stamp: ''!
sourceString
	^ (self actualClass sourceCodeAt: self methodSymbol) asString! !

!MethodReference methodsFor: 'string version' stamp: ''!
stringVersion	
	^stringVersion.! !

!MethodReference methodsFor: 'string version' stamp: ''!
stringVersion: aString

	stringVersion := aString! !

!MethodReference methodsFor: 'string version' stamp: ''!
stringVersionDefault

	^classSymbol, (classIsMeta ifTrue: [ ' class ' ] ifFalse: [' ']), methodSymbol  ! !

!MethodReference methodsFor: 'accessing' stamp: ''!
timeStamp
	^ self compiledMethod timeStamp! !

MethodReference subclass: #SourcedMethodReference
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'System-Support'!

!SourcedMethodReference commentStamp: 'Sd 12/8/2010 21:09' prior: 0!
A SourcedMethodReference is a hack that should be eliminated when we will have a correct and decent
MethodDefinition class.
For now the inheritance decomposition (sourceReference, methodReference, SourcedMethodReference) is just to make sure that we can migrate the system to MethodReference supporting source. !

!SourcedMethodReference methodsFor: 'comparisons' stamp: 'JohanBrichau 1/21/2011 12:07'!
<= aSourceReference
	"Sort source references according to: 1. class - 2. class comment - 3. class/instance methods - 4. method selector - 5. timestamp"
	self flag: 'temporary override as long as MethodReference is in this hierarchy'.
	
	classSymbol < aSourceReference classSymbol ifTrue: [^true].
	classSymbol > aSourceReference classSymbol ifTrue: [^false].
	aSourceReference isCommentReference ifTrue:[^false].
	self isCommentReference ifTrue:[^ true].
	self isClassMethodReference = aSourceReference isClassMethodReference ifFalse: [^self isClassMethodReference not].
	^ self methodSymbol = aSourceReference methodSymbol 
		ifTrue:[true]
		ifFalse:[self time <= aSourceReference time]
		! !

!SourcedMethodReference methodsFor: 'comparisons' stamp: 'BenjaminVanRyseghem 12/13/2010 17:40'!
= aMethodReference

	^(super = aMethodReference) and: [self sourcePointer = aMethodReference sourcePointer].! !

!SourcedMethodReference methodsFor: 'accessing' stamp: 'BenjaminVanRyseghem 11/2/2010 16:49'!
browse
	ToolSet browse: self actualClass selector: self selector.! !

!SourcedMethodReference methodsFor: 'comparisons' stamp: 'StephaneDucasse 1/14/2011 15:39'!
hash

	^ super hash bitXor: self sourcePointer hash
! !

!SourcedMethodReference methodsFor: 'private' stamp: ''!
preambleString
	^ 'methodsFor:'! !

!SourcedMethodReference methodsFor: 'printing' stamp: ''!
printOn: aStream
	aStream 
		nextPutAll: self classSymbol asString;
		nextPutAll: ' >> #';
		nextPutAll: self methodSymbol asString.! !

!SourcedMethodReference methodsFor: 'initialization' stamp: ''!
printStringForRML
	|suffix|
	suffix := ''.
	(self actualClass canUnderstand: self selector)
		ifFalse:[suffix := ' (removed)'].
	^self selector, suffix! !

!SourcedMethodReference methodsFor: 'initialization' stamp: 'StephaneDucasse 1/7/2011 19:00'!
setClass: aClass methodSymbol: methodSym

	super setClass: aClass methodSymbol: methodSym.
	sourcePointer := (aClass >> methodSym) sourcePointer.! !

!SourcedMethodReference methodsFor: 'initialization' stamp: 'StephaneDucasse 1/7/2011 19:00'!
setClass: aClass methodSymbol: methodSym stringVersion: aString 

	super setClass: aClass methodSymbol: methodSym stringVersion: aString.
	sourcePointer := (aClass >> methodSym) sourcePointer.! !

!SourcedMethodReference methodsFor: 'initialization' stamp: 'StephaneDucasse 1/7/2011 19:03'!
setClassSymbol: classSym classIsMeta: isMeta methodSymbol: methodSym stringVersion: aString 

	| aClass |
	aClass := isMeta 
		ifTrue: [(Smalltalk globals at:  classSym) class]
		ifFalse: [(Smalltalk globals at:  classSym) ].
	self  setStandardClass: aClass methodSymbol: methodSym! !

!SourcedMethodReference methodsFor: 'initialization' stamp: 'JohanBrichau 12/30/2010 12:46'!
setStandardClass: aClass methodSymbol: methodSym

	super setStandardClass: aClass methodSymbol: methodSym.
	sourcePointer := (aClass >> methodSym) sourcePointer.! !

!SourcedMethodReference methodsFor: 'accessing' stamp: ''!
sourceCode
	| fileIndex filePosition |
	self flag: 'should be removed'.
	fileIndex := SourceFiles fileIndexFromSourcePointer: sourcePointer.
	filePosition := SourceFiles filePositionFromSourcePointer: sourcePointer.
	^ filePosition = 0 
		ifTrue: [nil]
		ifFalse:[(RemoteString newFileNumber: fileIndex position: filePosition) text]

! !

!SourcedMethodReference methodsFor: 'private' stamp: ''!
stampSymbol
	^ #stamp:! !

!SourcedMethodReference methodsFor: 'accessing' stamp: 'BenjaminVanRyseghem 1/8/2011 15:02'!
timeStamp
	self flag: 'ugly inheritance/overridding for temporary compatibility'.
	
	^ self time asString! !

!SourceReference methodsFor: 'comparing' stamp: 'JohanBrichau 1/21/2011 12:07'!
<= aSourceReference
	"Sort source references according to: 1. class - 2. class comment - 3. class/instance methods - 4. method selector - 5. timestamp"
	
	self classSymbol < aSourceReference classSymbol ifTrue: [^true].
	self classSymbol > aSourceReference classSymbol ifTrue: [^false].
	aSourceReference isCommentReference ifTrue:[^false].
	self isCommentReference ifTrue:[^ true].
	self isClassMethodReference = aSourceReference isClassMethodReference ifFalse: [^self isClassMethodReference not].
	^ self methodSymbol = aSourceReference methodSymbol 
		ifTrue:[true]
		ifFalse:[self timeStamp <= aSourceReference timeStamp]! !

!SourceReference methodsFor: 'accessing' stamp: 'BenjaminVanRyseghem 11/2/2010 16:52'!
browse
	^self subclassResponsibility! !

!SourceReference methodsFor: 'accessing' stamp: ''!
date
	"return the date when the method was modified"
	^ self timeStamp asDate! !

!SourceReference methodsFor: 'private' stamp: ''!
fileIndex
	^SourceFiles fileIndexFromSourcePointer: self sourcePointer! !

!SourceReference methodsFor: 'private' stamp: ''!
filePosition
	^SourceFiles filePositionFromSourcePointer: self sourcePointer! !

!SourceReference methodsFor: 'private' stamp: 'BenjaminVanRyseghem 11/12/2010 00:07'!
findTimestampInSource	
		"Answer the authoring time-stamp for the given method, retrieved from the sources or changes file. Answer the empty string if no time stamp is available."	
			
	| position file preamble stamp tokens stampPosition |

	self fileIndex == 0 ifTrue: [^ String new].  "no source pointer for this method"
	position :=self filePosition.
	file := SourceFiles at: self fileIndex.
	file ifNil: [^ String new].  "sources file not available"
	"file does not exist happens in secure mode"
	"Transcript show: file ; cr."
	file := [file readOnlyCopy] 
			on: FileDoesNotExistException 
			do:[:ex| ^ String new].
	self flag: 'I do not know why 3 is a good number'.
	preamble :=  self getPreambleFrom: file at: (0 max: position - 3).
	(preamble findString: self preambleString startingAt: 1) > 0
			ifTrue: [tokens := Scanner new scanTokens: preamble.
				
				     	stampPosition := tokens indexOf: self stampSymbol.
					stampPosition = 0
						ifTrue:[stamp:= ''.]
						ifFalse: [
							"New format gives change stamp and unified prior pointer"
							stamp := tokens at: stampPosition + 1]]
			ifFalse: [" Name is already defined ->The stamp get lost, maybe after a condenseChanges"
					stamp := '' ].
	file close.
 	^ stamp! !

!SourceReference methodsFor: 'private' stamp: ''!
getPreambleFrom: aFileStream at: position
	|  writeStream |
	writeStream := String new writeStream.
	position to: 0 by: -1 do: [ :p | 
			| c | 
			aFileStream position: p.
			c := aFileStream basicNext.
			c == $!!
				ifTrue: [^ writeStream contents reversed]
				ifFalse: [writeStream nextPut: c]].
	self halt: 'should not arrive here'.! !

!SourceReference methodsFor: 'testing' stamp: 'JohanBrichau 1/21/2011 12:08'!
isClassMethodReference 
	^ false! !

!SourceReference methodsFor: 'testing' stamp: 'JohanBrichau 1/21/2011 12:08'!
isCommentReference 
	^ false! !

!SourceReference methodsFor: 'private' stamp: 'JohanBrichau 12/30/2010 15:05'!
parseTimestampFrom: aString
	"Parse a date-time from a timestamp-string that is extracted from a source file. If there is no timestamp, or we cannot make sense of it, we return the epoch."
	| tokens dateStartIndex |
	"The following timestamp strings are supported (source: squeak sources archeological survey):
		<authorname><date><time>. 
		<authorname><date>
		<date><time>
		<date><time><authorname>
		<date><authorname>
		<historical>
	All fields can be separated by spaces or line ends but a separator between authorname and date can be missing as well"
	
	dateStartIndex := (aString indexOf: $/) - 1. 
	"If there is no / character in the timestamp, we cannot parse a date and return the epoch"
	dateStartIndex = -1 ifTrue:[^ DateAndTime epoch].
	^ [	"Go the start of the date string (there can be 1 or 2 digits and a space separator can be missing at the front!!!!)"
		((dateStartIndex < 2) not and:[(aString at: dateStartIndex - 1) isDigit]) ifTrue:[dateStartIndex := dateStartIndex - 1].
		"Extract only those tokens that possibly represent date or time"
		tokens := (aString copyFrom: dateStartIndex to: aString size) subStrings select: [:token | (token occurrencesOf: $/) = 2 or:[(token occurrencesOf: $:) = 1]].
		"2 tokens is a datetime"
		tokens size = 2
			ifTrue:[(tokens joinUsing: Character space) asDateAndTime]
			ifFalse:[(tokens first) asDate asDateAndTime]
		] 
		on: Exception 
		do:[:e | DateAndTime epoch]! !

!SourceReference methodsFor: 'private' stamp: ''!
preambleString
	self subclassResponsibility! !

!SourceReference methodsFor: 'printing' stamp: ''!
printStringForRML
	^self subclassResponsibility! !

!SourceReference methodsFor: 'tool compatibility' stamp: 'JohanBrichau 1/21/2011 12:08'!
setClassAndSelectorIn: csBlock

	^csBlock value: self actualClass value: self methodSymbol! !

!SourceReference methodsFor: 'accessing' stamp: ''!
sourceCode
	| fileIndex filePosition |
	fileIndex := SourceFiles fileIndexFromSourcePointer: sourcePointer.
	filePosition := SourceFiles filePositionFromSourcePointer: sourcePointer.
	^ filePosition = 0 
		ifTrue: [nil]
		ifFalse:[(RemoteString newFileNumber: fileIndex position: filePosition) text]

! !

!SourceReference methodsFor: 'accessing' stamp: ''!
sourcePointer 
	^ sourcePointer! !

!SourceReference methodsFor: 'private' stamp: ''!
stampSymbol
	self subclassResponsibility! !

!SourceReference methodsFor: 'accessing' stamp: 'BenjaminVanRyseghem 1/8/2011 15:02'!
time
	"I'm actually the real accessor to timestamp as a TimeStamp"
	
	^ timeStamp ifNil:[timeStamp := self parseTimestampFrom: self timeStampString]! !

!SourceReference methodsFor: 'accessing' stamp: 'BenjaminVanRyseghem 1/8/2011 15:02'!
timeStamp
	"Should be a String"
	"If you want a Date, use date instead"
	
	^ self time asString! !

!SourceReference methodsFor: 'accessing' stamp: 'JohanBrichau 12/30/2010 12:51'!
timeStampString 
	^ timeStampString ifNil:[timeStampString := self findTimestampInSource]! !

Object subclass: #SystemNavigation
	instanceVariableNames: 'browserClass hierarchyBrowserClass environment'
	classVariableNames: 'Default'
	poolDictionaries: ''
	category: 'System-Support'!

!SystemNavigation commentStamp: 'LucFabresse 10/10/2010 19:54' prior: 0!
I support the navigation of the system. 
I act as a facade but as I could require some state or different way of navigating the system all my behavior are on the instance side.
I should (it is not complety done yet) be parametrized by an environment (aSystemDictionary) that scopes my queries.!

!SystemNavigation class methodsFor: 'accessing' stamp: 'StephaneDucasse 10/23/2010 15:40'!
default
	^ self new! !

!SystemNavigation methodsFor: 'message sends' stamp: ''!
addSelectorsReferingTo: aSymbol in: class to: sortedSenders special: special byte: byte 
	{class. class class} do: [:behavior| (behavior 
		thoroughWhichSelectorsReferTo: aSymbol
		special: special
		byte: byte) do: [ :sel | 
			sortedSenders add: (SourcedMethodReference 
				class: behavior
				selector: sel) ]]! !

!SystemNavigation methodsFor: 'query' stamp: 'AlexandreBergel 12/29/2010 11:39'!
allBehaviorsDo: aBlock 
	"Execute a block on each class, metaclass, trait and trait class"
	
	self environment allClassesDo: [ :aClass | 
		aBlock value: aClass ; 
			value: aClass class
	].
	
	self environment allTraitsDo: [ :aTrait |
		aBlock value:  aTrait ;
			value: aTrait classTrait
	]
! !

!SystemNavigation methodsFor: 'query' stamp: ''!
allCallsOn: aSymbol 
	"Answer a Collection of all the methods that call on the provided symbol even deeply embedded in literal array."
	"self new allCallsOn: #allCallsOn:"
	
	^ self allReferencesTo: aSymbol! !

!SystemNavigation methodsFor: 'query' stamp: ''!
allCallsOn: firstLiteral and: secondLiteral
	"Answer a SortedCollection of all the methods that call on both aLiteral 
	and secondLiteral."

	| aCollection firstSpecial secondSpecial firstByte secondByte |
	self flag: #ShouldUseAllCallsOn:. "sd"
	aCollection := SortedCollection new.
	firstSpecial := Smalltalk hasSpecialSelector: firstLiteral ifTrueSetByte: [:b | firstByte := b].
	secondSpecial := Smalltalk hasSpecialSelector: secondLiteral ifTrueSetByte: [:b | secondByte := b].
	self allBehaviorsDo: [:class | | secondArray |
			secondArray := class whichSelectorsReferTo: secondLiteral special: secondSpecial byte: secondByte.
			((class whichSelectorsReferTo: firstLiteral special: firstSpecial byte: firstByte) select: [:aSel | 
				(secondArray includes: aSel)]) do: [:sel | 
					aCollection add: (SourcedMethodReference class: class selector: sel)]].
	^aCollection! !

!SystemNavigation methodsFor: 'query' stamp: ''!
allCallsOn: aSymbol from: aClass
	"Answer a SortedCollection of all the methods that call on aSymbol."

	| aSortedCollection special byte |
	aSortedCollection := SortedCollection new.
	special := Smalltalk hasSpecialSelector: aSymbol ifTrueSetByte: [:b | byte := b ].
	aClass withAllSubclassesDo: [ :class |
		(class whichSelectorsReferTo: aSymbol special: special byte: byte) do: [:sel |
				aSortedCollection add: (SourcedMethodReference class: class selector: sel)]].
	^aSortedCollection! !

!SystemNavigation methodsFor: 'query' stamp: ''!
allCallsOn: aSymbol localTo: aClass
	"Answer a Set of SourcedMethodReferences for all the methods
	 that call on aSymbol in, above or below the given class."

	| aSet special byte enum |
	aSet := Set new.
	special := Smalltalk hasSpecialSelector: aSymbol ifTrueSetByte: [:b | byte := b].
	enum := [:behavior|
			(behavior whichSelectorsReferTo: aSymbol special: special byte: byte) do:
				[:sel | aSet add: (SourcedMethodReference new setClass: behavior  methodSymbol: sel)]].
	aClass theNonMetaClass withAllSuperAndSubclassesDoGently: enum.
		aClass theNonMetaClass class withAllSuperAndSubclassesDoGently: enum.
	^aSet! !

!SystemNavigation methodsFor: 'query' stamp: ''!
allCallsOn: aLiteral localToPackage: packageNameOrInfo
	"Answer a Set of MethodReferences for all the methods
	 that call on aSymbol in the given package."

	| aSet special byte |
	aSet := Set new.
	special := Smalltalk hasSpecialSelector: aLiteral ifTrueSetByte: [:b | byte := b].
	Cursor wait showWhile:
		[(self packageInfoFor: packageNameOrInfo) actualMethodsDo:
			[:method |
			((method hasLiteral: aLiteral) or: [special and: [method scanFor: byte]]) ifTrue:
				[((aLiteral isVariableBinding) not
					or: [method literals allButLast includes: aLiteral])
						ifTrue: [aSet add: method methodReference]]].].
	^aSet! !

!SystemNavigation methodsFor: 'query' stamp: 'LucFabresse 10/11/2010 19:57'!
allClasses
	"currently returns all the classes in the current environment"

	^ self environment allClasses

	! !

!SystemNavigation methodsFor: 'query' stamp: 'AlexandreBergel 12/29/2010 11:38'!
allClassesAndTraits
	"Answer all the classes and traits present in the image"
	"Note that metaclasses are not part of the answer"
	
	^ Smalltalk globals allClassesAndTraits

	! !

!SystemNavigation methodsFor: 'query' stamp: 'StephaneDucasse 11/9/2010 15:21'!
allClassesDefiningClassVarNamed: aString  
	"Answer an Array of all classes that define the class variables named aString."

	| aCollection |
	aCollection := ReadWriteStream on: Array new.
	self allClassesDo: [:class | (class hasClassVarNamed: aString) ifTrue: [aCollection nextPut: class]].
	^ aCollection contents! !

!SystemNavigation methodsFor: 'query' stamp: ''!
allClassesDo: aBlock
	"currently returns all the classes defined in Smalltalk but could be customized 
	for dealing with environments and  in such a case would work on really all the classes"

	^ Smalltalk allClassesDo: aBlock

	! !

!SystemNavigation methodsFor: 'query' stamp: ''!
allClassesImplementing: aSelector  
	"Answer an Array of all classes that implement the message aSelector."

	| aCollection |
	aCollection := ReadWriteStream on: Array new.
	self allBehaviorsDo: [:class | (class includesSelector: aSelector) ifTrue: [aCollection nextPut: class]].
	^ aCollection contents! !

!SystemNavigation methodsFor: 'packages' stamp: ''!
allClassesInPackageNamed: packageName
	"Returns the list of classes that belongs to a Monticello package."
	
	"It would be probably better to delegate that query to the package itself"
	| classes organization |
	organization := Smalltalk globals organization.
	classes := (organization classesInCategory: packageName) asSet.
	(organization categoriesMatching: packageName, '-*') do: [:category| 
		classes addAll: (organization classesInCategory: category)].
	^classes asArray! !

!SystemNavigation methodsFor: 'query' stamp: 'StephaneDucasse 11/9/2010 14:48'!
allClassesUsingSharedPool: aString  
	"Answer an Array of all classes that uses the shared pool named aString."

	| aCollection |
	aCollection := ReadWriteStream on: Array new.
	self allClassesDo: [:class | (class includesSharedPoolNamed: aString) ifTrue: [aCollection nextPut: class]].
	^ aCollection contents! !

!SystemNavigation methodsFor: 'query' stamp: ''!
allClassesWithUnimplementedCalls
	"Answer an Array of classes that have messages with calls to methods that aren't implemented
	anywhere in the system"
	| all dict |
	dict := Dictionary new.
	all := self systemNavigation allImplementedMessages.
	self systemNavigation allBehaviorsDo: [:cl | 
		cl selectorsAndMethodsDo: [:sel :meth | 
			meth primitive = 0 ifTrue: [
				meth messages do: [:m | 
					(all includes: m) ifFalse: [
						((dict at: cl ifAbsentPut: [ Dictionary new ])
							at: sel ifAbsentPut: [ OrderedCollection new])
								add: m
					]
				]
			]
		]
	].
	^ dict! !

!SystemNavigation methodsFor: 'author' stamp: 'AlexandreBergel 12/29/2010 11:38'!
allContributors
	"Return the list of initials and names of method authors."
	
	"I am wondering why a bag is used instead of a set. AlexandreBergel 12/29/2010 11:37"

	"SystemNavigation default allContributors"
	| contributorsNames |
	contributorsNames := Bag new.
	self allBehaviorsDo: [ :behavior |
		behavior methodsDo: [ :compiledMethod | | stamp |
			stamp := compiledMethod timeStamp.
			stamp notEmpty ifTrue: [
				contributorsNames add: compiledMethod timeStamp substrings first ]]].
	^contributorsNames! !

!SystemNavigation methodsFor: 'query' stamp: ''!
allGlobalRefs
	"Answer a set of symbols that may be refs to Global names.  In some sense we should only need the associations, but this will also catch, eg, HTML tag types."

	^ self allGlobalRefsWithout: {{}. {}}! !

!SystemNavigation methodsFor: 'query' stamp: ''!
allGlobalRefsWithout: classesAndMessagesPair
	"Answer a set of symbols that may be refs to Global names. In some  
	sense we should only need the associations, but this will also catch, eg,  
	HTML tag types. This method computes its result in the absence of  
	specified classes and messages."

	"may be a problem if namespaces are introduced as for the moment  
	only Smalltalk is queried. sd 29/4/03"

	| globalRefs absentClasses absentSelectors |
	globalRefs := IdentitySet new: CompiledMethod instanceCount.
	absentClasses := classesAndMessagesPair first.
	absentSelectors := classesAndMessagesPair second.
	self flag: #shouldBeRewrittenUsingSmalltalkAllClassesDo:.	"sd 29/04/03"
	Cursor execute
		showWhile: [ 
			Smalltalk classNames
				do: [ :cName | 
					((absentClasses includes: cName)
						ifTrue: [ {} ]
						ifFalse: [ 
							{(Smalltalk globals at: cName).
							((Smalltalk globals at: cName) class)} ])
						do: [ :cl | 
							(absentSelectors isEmpty
								ifTrue: [ cl selectors ]
								ifFalse: [ cl selectors copyWithoutAll: absentSelectors ])
								do: [ :sel | 
									"Include all capitalized symbols for good 
									measure"
									(cl compiledMethodAt: sel)
										literalsDo: [ :m | 
											(m isSymbol and: [ m size > 0 and: [ m first canBeGlobalVarInitial ] ])
												ifTrue: [ globalRefs add: m ].
											(m isMemberOf: Array)
												ifTrue: [ 
													m
														do: [ :x | 
															(x isSymbol and: [ x size > 0 and: [ x first canBeGlobalVarInitial ] ])
																ifTrue: [ globalRefs add: x ] ] ].
											m isVariableBinding
												ifTrue: [ m key ifNotNil: [ globalRefs add: m key ] ] ] ] ] ] ].
	^ globalRefs! !

!SystemNavigation methodsFor: 'query' stamp: ''!
allImplementedMessages
	"Answer a Set of all the messages that are implemented in the system."
	^ self allImplementedMessagesWithout: {{}. {}}! !

!SystemNavigation methodsFor: 'query' stamp: ''!
allImplementedMessagesWithout: classesAndMessagesPair
	"Answer a Set of all the messages that are implemented in the system,  
	computed in the absence of the supplied classes and messages. Note this  
	reports messages that are in the absent selectors set."

	| messages absentClasses |
	messages := IdentitySet new: CompiledMethod instanceCount.
	absentClasses := classesAndMessagesPair first.
	self flag: #shouldBeRewrittenUsingSmalltalkAllClassesDo:.	"sd 29/04/03"
	Smalltalk classNames , Smalltalk traitNames
		do: [ :name | 
			((absentClasses includes: name)
				ifTrue: [ {} ]
				ifFalse: [ 
					{(Smalltalk globals at: name).
					((Smalltalk globals at: name) classSide)} ]) do: [ :each | messages addAll: each selectors ] ].
	^ messages! !

!SystemNavigation methodsFor: 'query' stamp: ''!
allImplementorsOf: aSelector 
	"Answer a SortedCollection of all the methods that implement the message 
	aSelector."
	| aCollection |
	aCollection := SortedCollection new.
	self allBehaviorsDo: [:class | 
		(class includesSelector: aSelector)
			ifTrue: [aCollection add: (SourcedMethodReference new setClass: class methodSymbol: aSelector)]].
	^ aCollection! !

!SystemNavigation methodsFor: 'query' stamp: ''!
allImplementorsOf: aSelector localTo: aClass 
	"Answer a SortedCollection of all the methods that implement the message  
	aSelector in, above, or below the given class."
	| cls aCollection |
	aCollection := SortedCollection new.
	cls := aClass theNonMetaClass.
	cls
				withAllSuperAndSubclassesDo: [:class | (class includesSelector: aSelector)
						ifTrue: [aCollection
								add: (SourcedMethodReference new setClass: class methodSymbol: aSelector)]].
			cls class
				withAllSuperAndSubclassesDo: [:class | (class includesSelector: aSelector)
						ifTrue: [aCollection
								add: (SourcedMethodReference new setClass: class methodSymbol: aSelector)]].
	^ aCollection! !

!SystemNavigation methodsFor: 'query' stamp: ''!
allImplementorsOf: aSelector  localToPackage: packageNameOrInfo
	"Answer a SortedCollection of all the methods that implement the message 
	 aSelector in the given package."

	| aSet |
	aSet := Set new.
	Cursor wait showWhile:
		[(self packageInfoFor: packageNameOrInfo) actualMethodsDo:
			[:m |
			(m selector = aSelector) ifTrue:
				[aSet add: m methodReference]]].
	^aSet! !

!SystemNavigation methodsFor: 'query' stamp: ''!
allLocalCallsOn: aSymbol ofClass: aClass
	"Answer a SortedCollection of all the methods that call on aSymbol, anywhere in my class hierarchy."

	| aSet special byte cls |
	aSet := Set new.
	cls := aClass theNonMetaClass.
	special := Smalltalk hasSpecialSelector: aSymbol
					ifTrueSetByte: [:b | byte := b ].
	cls withAllSuperAndSubclassesDo: [ :class |
		(class whichSelectorsReferTo: aSymbol special: special byte: byte)
			do: [:sel |
				aSet add: (SourcedMethodReference new
							setClass: class  			
							methodSymbol: sel 
							stringVersion: (class name asString , ' ', sel asString)).]].
		
	cls class withAllSuperAndSubclassesDo: [ :class |
		(class whichSelectorsReferTo: aSymbol special: special byte: byte)
			do: [:sel |
				aSet add: (SourcedMethodReference new
							setClass: class  			
							methodSymbol: sel 
							stringVersion: (class name asString , ' ', sel asString)).]].
	^aSet! !

!SystemNavigation methodsFor: 'accessing' stamp: 'AlexandreBergel 12/29/2010 11:35'!
allMethodsInCategory: category 
	"Answer the collection of methods that are contained in a particular method category"
	
	" SystemNavigation default allMethodsInCategory: 'printing'"

	| aCollection |
	aCollection := Set new.
	Cursor wait showWhile: [
			self allBehaviorsDo: [:x | ((category = ClassOrganizer allCategory
					ifTrue: [x organization allMethodSelectors]
					ifFalse: [x organization listAtCategoryNamed: category])) do:
				[:sel | aCollection add: (SourcedMethodReference new setClass: x 
methodSymbol: sel)]]].
	^aCollection.
! !

!SystemNavigation methodsFor: 'query' stamp: ''!
allMethodsSelect: aBlock 
	"Answer a SortedCollection of each method that, when used as the block  
	argument to aBlock, gives a true result."
	| aCollection |
	aCollection := SortedCollection new.
	self allBehaviorsDo: [:class | class
						selectorsAndMethodsDo: [:sel :m | (aBlock value: m)
								ifTrue: [aCollection
										add: (SourcedMethodReference new setClass: class methodSymbol: sel)]]].
	^ aCollection! !

!SystemNavigation methodsFor: 'query' stamp: ''!
allMethodsSelect: aBlock localToPackage: packageNameOrInfo
	"Answer a SortedCollection of each method in the given package
	 for which the evaluation of aBlock with the metnod answers true."

	| aSet |
	aSet := Set new.
	Cursor wait showWhile:
		[(self packageInfoFor: packageNameOrInfo) actualMethodsDo:
			[:aMethod |
			(aBlock value: aMethod) ifTrue:
				[aSet add: aMethod methodReference]]].
	^aSet! !

!SystemNavigation methodsFor: 'query' stamp: 'BenjaminVanRyseghem 11/10/2010 13:25'!
allMethodsWithSourceString: aString matchCase: caseSensitive
	"Answer a SortedCollection of all the methods that contain, in source code, aString as a substring.  Search the class comments also"

	| list addMethod addComment |
	list := Set new.
	
	addMethod := [ :mrClass :mrSel | list add: (SourcedMethodReference new
											setClass: mrClass
											methodSymbol: mrSel)].									
	addComment := [ :mrClass | list add: (CommentReference class: mrClass)].
										
	'Searching all source code...'
		displayProgressAt: Sensor cursorPoint
		from: 0 to: Smalltalk globals classNames size
		during: [:bar | | count |
			count := 0.
			self allBehaviorsDo: [:each |
				bar value: (count := count + 1).
					each selectorsDo: [:sel | 
						((each sourceCodeAt: sel) findString: aString 
							startingAt: 1 caseSensitive: caseSensitive) > 0 ifTrue: [
								sel isDoIt ifFalse: [addMethod value: each value: sel]]].
					(each organization classComment asString findString: aString 
							startingAt: 1 caseSensitive: caseSensitive) > 0 ifTrue: [
								addComment value: each]	]].
			^ list asSortedCollection! !

!SystemNavigation methodsFor: 'query' stamp: ''!
allObjectsDo: aBlock 
	"Evaluate the argument, aBlock, for each object in the system 
	excluding SmallIntegers."
	| object endMarker |
	object := self someObject.
	endMarker := Object new.
	[endMarker == object]
		whileFalse: [aBlock value: object.
			object := object nextObject]! !

!SystemNavigation methodsFor: 'query' stamp: ''!
allObjectsSelect: aBlock 
	"Evaluate the argument, aBlock, for each object in the system excluding 
	SmallIntegers. Return a collection af all objects for whom the value is 
	true. "
	^ Array
		streamContents: [:s | self
				allObjectsDo: [:object | (aBlock value: object)
						ifTrue: [s nextPut: object]]]! !

!SystemNavigation methodsFor: 'query' stamp: ''!
allPrimitiveMethods
	"Answer an OrderedCollection of all the methods that are implemented by primitives."
	| aColl |
	aColl := OrderedCollection new: 200.
	self allBehaviorsDo: [:class | class
		selectorsAndMethodsDo: [:sel :method | 
			method primitive ~= 0
				ifTrue: [aColl addLast: class name , ' ' , sel , ' ' , method primitive printString]]].
	^ aColl! !

!SystemNavigation methodsFor: 'query' stamp: 'AlexandreBergel 12/29/2010 11:34'!
allPrimitiveMethodsInCategories: aList 
	"Answer an OrderedCollection of all the methods that are implemented by 
	primitives in the given categories. 1/26/96 sw"
	"SystemNavigation new allPrimitiveMethodsInCategories:  
		#('Collections-Streams' 'Files-Streams' 'Files-Abstract' 'Files-Macintosh')"

	| aColl |
	aColl := OrderedCollection new: 200.
	self
		allBehaviorsDo: [:aClass | (aList includes: (SystemOrganization categoryOfElement: aClass theNonMetaClass name asString) asString)
				ifTrue: [aClass
						selectorsAndMethodsDo: [:sel :method | 
							method primitive ~= 0
								ifTrue: [aColl addLast: aClass name , ' ' , sel , ' ' , method primitive printString]]]].
	^ aColl! !

!SystemNavigation methodsFor: 'query' stamp: 'AlexandreBergel 12/29/2010 11:33'!
allReferencesTo: aLiteral 
	"Answer a Collection of all the methods that refer to aLiteral even deeply embedded in literal array."

	"SystemNavigation default allReferencesTo: #+"
	"SystemNavigation default allReferencesTo: Point binding"
	
	| result special byte symbol |
	result := OrderedCollection new.
	special := Smalltalk
		hasSpecialSelector: aLiteral
		ifTrueSetByte: [ :value | byte := value ].
	symbol := aLiteral isSymbol.
	self allBehaviorsDo: [ :class |
		| selectors |
		selectors := symbol
			ifFalse: [ class whichSelectorsReferTo: aLiteral special: special byte: byte ]
			ifTrue: [ class thoroughWhichSelectorsReferTo: aLiteral special: special byte: byte ].
		selectors do: [ :selector |
			result addLast: (SourcedMethodReference class: class selector: selector) ] ].
	^ result! !

!SystemNavigation methodsFor: 'query' stamp: ''!
allReferencesToPool: aPool
	"Answer all the references to variable aPool"
	| list |
	list := OrderedCollection new.
	self allClassesDo: [:cls|
				cls selectorsAndMethodsDo:[:sel :meth| | ref |
					ref := meth literals detect:
							[:lit|
								(lit isVariableBinding 
								and: [(lit key) notNil])
								and: [(aPool bindingOf: lit key) notNil	]
							] ifNone:[nil].
					ref ifNotNil:[
						list add:(SourcedMethodReference new setClass: cls methodSymbol: sel)
					].
				].
			].
	^list! !

!SystemNavigation methodsFor: 'query' stamp: ''!
allReferencesToPool: aPool from: aClass
	"Answer all the references to variables from aPool"
	| list |
	list := OrderedCollection new.
	aClass withAllSubclassesDo:[:cls|
		cls selectorsAndMethodsDo:[:sel :meth| | ref |
			ref := meth literals detect:[:lit|
				lit isVariableBinding and:[(aPool bindingOf: lit key) notNil]
			] ifNone:[nil].
			ref ifNotNil:[
				list add:(SourcedMethodReference new setClass: cls methodSymbol: sel)
			].
		].
	].
	^list! !

!SystemNavigation methodsFor: 'query' stamp: ''!
allSelectorsWithAnyImplementorsIn: selectorList 
	"Answer the subset of the given list which represent method selectors 
	which have at least one implementor in the system."
	| good |
	good := OrderedCollection new.
	self allBehaviorsDo: [:class | 
			selectorList do: [:aSelector | (class includesSelector: aSelector) ifTrue: [good add: aSelector]]].
	^ good asSet asArray sort! !

!SystemNavigation methodsFor: 'message sends' stamp: ''!
allSendersOf: selector 
	|  sortedSenders special byte |
	sortedSenders := SortedCollection new.
	special := Smalltalk 
				hasSpecialSelector: selector
				ifTrueSetByte: [ :b | byte := b ].
	self allBehaviorsDo: 
		[ :behavior | 
		self 
			addSelectorsReferingTo: selector
			in: behavior
			to: sortedSenders
			special: special
			byte: byte ].
	^ sortedSenders! !

!SystemNavigation methodsFor: 'message sends' stamp: ''!
allSendersOf: selector inClass: aClass 
	| sortedSenders special byte |
	sortedSenders := SortedCollection new.
	special := Smalltalk hasSpecialSelector: selector ifTrueSetByte: [:b | byte := b ].
	self 
		addSelectorsReferingTo: selector
		in: aClass
		to: sortedSenders
		special: special
		byte: byte.
	^sortedSenders! !

!SystemNavigation methodsFor: 'message sends' stamp: 'LucFabresse 10/12/2010 08:23'!
allSendersOf: selector inClassCategory: category 
	| classes sortedSenders special byte |
	classes := self environment organization classesInCategory: category.
	sortedSenders := SortedCollection new.
	classes ifEmpty: [ ^ sortedSenders ].
	special := Smalltalk
				hasSpecialSelector: selector
				ifTrueSetByte: [ :b | byte := b ].
	classes do: 
		[ :class | 
		self 
			addSelectorsReferingTo: selector
			in: class
			to: sortedSenders
			special: special
			byte: byte ].
	^ sortedSenders! !

!SystemNavigation methodsFor: 'message sends' stamp: ''!
allSendersOf: selector inPackageNamed: packageName
	^(self categoriesInPackageNamed: packageName) inject: SortedCollection new into: [:sortedSenders :category|
		sortedSenders, (self allSendersOf: selector inClassCategory: category)]! !

!SystemNavigation methodsFor: 'query' stamp: ''!
allSentMessages
	"Answer the set of selectors which are sent somewhere in the system."
	^ self  allSentMessagesWithout: {{}. {}}! !

!SystemNavigation methodsFor: 'query' stamp: ''!
allSentMessagesWithout: classesAndMessagesPair 
	"Answer the set of selectors which are sent somewhere in the system,  
	computed in the absence of the supplied classes and messages."
	| sent absentClasses absentSelectors |
	sent := IdentitySet new: CompiledMethod instanceCount.
	absentClasses := classesAndMessagesPair first.
	absentSelectors := classesAndMessagesPair second.
	self allBehaviorsDo: [:class |
		((absentClasses includes: class)
			ifTrue: [{}]
			ifFalse: [{class. class classSide}]) do: [:each | 
					(absentSelectors isEmpty ifTrue: [each selectors] ifFalse: [each selectors copyWithoutAll: absentSelectors])
					do: [:sel | "Include all sels, but not if sent by self"
						(each compiledMethodAt: sel) literalsDo: [:m | 
								(m isSymbol) ifTrue: ["might be sent"
										m == sel ifFalse: [sent add: m]].
								(m isMemberOf: Array) ifTrue: ["might be performed"
										m do: [:x | (x isSymbol) ifTrue: [x == sel ifFalse: [sent add: x]]]]]]]].
		"The following may be sent without being in any literal frame"
		1
			to: Smalltalk specialSelectorSize
			do: [:index | sent
					add: (Smalltalk specialSelectorAt: index)].
	^ sent! !

!SystemNavigation methodsFor: 'query' stamp: ''!
allUnSentMessagesWithout: classesAndMessagesPair 
	"Answer the set of selectors that are implemented but not sent, computed  
	in the absence of the supplied classes and messages."
	^ (self  allImplementedMessagesWithout: classesAndMessagesPair)
		copyWithoutAll: (self  allSentMessagesWithout: classesAndMessagesPair)! !

!SystemNavigation methodsFor: 'query' stamp: ''!
allUnimplementedCalls
	"Answer an Array of each message that is sent by an expression in a  
	method but is not implemented by any object in the system."
	| aStream all |
	all := self allImplementedMessages.
	aStream := (Array new: 50) writeStream.
	self
		allBehaviorsDo: [:cl | cl
				selectorsAndMethodsDo: [:sel :meth |
					| secondStream | 
					secondStream := (String new: 5) writeStream.
					meth messages
						do: [:m | (all includes: m)
								ifFalse: [secondStream nextPutAll: m;
										 space]].
					secondStream position = 0
						ifFalse: [aStream nextPut: cl name , ' ' , sel , ' calls: ' , secondStream contents]]].
	^ aStream contents! !

!SystemNavigation methodsFor: 'query' stamp: ''!
allUnimplementedNonPrimitiveCalls
	"Answer an Array of each message that is sent by an expression in a  
	method but is not implemented by any object in the system."
	| aStream all |
	all := self systemNavigation allImplementedMessages.
	aStream := (Array new: 50) writeStream.
	self systemNavigation
		allBehaviorsDo: [:cl | cl
				selectorsAndMethodsDo: [:sel :meth |
					| secondStream | 
					secondStream := (String new: 5) writeStream.
					meth primitive = 0 ifTrue: [
						meth messages
							do: [:m | (all includes: m)
									ifFalse: [secondStream nextPutAll: m;
											 space]].
						secondStream position = 0
							ifFalse: [aStream nextPut: cl name , ' ' , sel , ' calls: ' , secondStream contents]]]].
	^ aStream contents! !

!SystemNavigation methodsFor: 'query' stamp: ''!
allUnreferencedClassVariablesOf: aClass
	"Answer a list of the names of all the receiver's unreferenced class  
	vars, including those defined in superclasses"
	| aList |
	aList := OrderedCollection new.
	aClass withAllSuperclasses
		reverseDo: [:aSuperClass | aSuperClass classVarNames
				do: [:var | (self allReferencesTo: (aSuperClass classPool associationAt: var)) isEmpty
						ifTrue: [aList add: var]]].
	^ aList! !

!SystemNavigation methodsFor: 'query' stamp: ''!
allUnsentMessages
	"SystemNavigation new allUnSentMessages"
	"Answer the set of selectors that are implemented by some object in the  
	system but not sent by any."
	^ self allUnSentMessagesWithout: {{}. {}}! !

!SystemNavigation methodsFor: 'query' stamp: 'LucFabresse 10/12/2010 08:39'!
allUnsentMessagesIn: selectorSet 
	"Answer the subset of selectorSet which are not sent anywhere in the 
	system. As allSentMessages does not recurse into nested Arrays, filter
	with #allCallsOn, which does"
	
	^ (selectorSet copyWithoutAll: self allSentMessages) select: [:each | 			(self allCallsOn: each) isEmpty]   ! !

!SystemNavigation methodsFor: 'query' stamp: ''!
allUnusedClassesWithout: classesAndMessagesPair
	"Enumerates all classes in the system and returns a list of those that are 
	apparently unused. A class is considered in use if it (a) has subclasses  
	or (b) is referred to by some method or (c) has its name in use as a  
	literal."

	"SystemNavigation new unusedClasses"

	| unused |
	self flag: #rewriteMePlease.
	unused := Smalltalk globals classNames asIdentitySet copyWithoutAll: (self allGlobalRefsWithout: classesAndMessagesPair).
	^ unused
		reject: [ :cName | 
			| cl |
			cl := Smalltalk globals at: cName.
			cl subclasses isEmpty not or: [ cl inheritsFrom: FileDirectory ] ]! !

!SystemNavigation methodsFor: 'accessing' stamp: 'LucFabresse 10/11/2010 21:31'!
categoriesInPackageNamed: packageName
	^(self environment organization categoriesMatching: packageName), (self environment organization categoriesMatching: (packageName, '-*'))! !

!SystemNavigation methodsFor: 'author' stamp: ''!
contributionsOf: aString
	"SystemNavigation default contributionsOf: 'alain.plantec'"
	
	| answer |
	answer := OrderedCollection new.
	self allBehaviorsDo: [ :behavior |
		behavior methodsDo: [ :compiledMethod | | stamp initials |
			stamp := compiledMethod timeStamp.
			stamp notEmpty ifTrue: [
				initials := compiledMethod timeStamp substrings first.
				aString = initials ifTrue: [
					answer add: (compiledMethod selector -> compiledMethod methodClass)]]]].
	^answer! !

!SystemNavigation methodsFor: 'author' stamp: ''!
contributorsNotSignatories
	"SystemNavigation default contributorsNotSignatories"
	
	| signatories answer |
	answer := Dictionary new.
	signatories := self signatories.
	self allBehaviorsDo: [ :behavior |
		behavior methodsDo: [ :compiledMethod | | stamp initials |
			stamp := compiledMethod timeStamp.
			stamp notEmpty ifTrue: [
				initials :=  compiledMethod timeStamp substrings first.
				(signatories includes: initials) ifFalse: [
					(answer at: initials ifAbsentPut: [OrderedCollection new])
						add: (compiledMethod selector -> compiledMethod methodClass)]]]].
	^answer! !

!SystemNavigation methodsFor: 'accessing' stamp: 'LucFabresse 10/11/2010 19:09'!
environment
	^environment ! !

!SystemNavigation methodsFor: 'accessing' stamp: 'LucFabresse 10/11/2010 19:08'!
environment: aSystemDictionary
	^environment :=  aSystemDictionary! !

!SystemNavigation methodsFor: 'private' stamp: ''!
headingAndAutoselectForLiteral: aLiteral do: binaryBlock
	"Evaluate aBlock with either Users of ... or Senders of ... plus the auto-select string
	 for the given literal.  aLiteral can be a Symbol, a VariableBinding or an arbitrary object."

	| autoSelect |
	^aLiteral isSymbol
		ifTrue: [binaryBlock value: 'Senders of ', aLiteral value: aLiteral keywords first]
		ifFalse:
			[autoSelect := aLiteral isVariableBinding
							ifTrue: [aLiteral key]
							ifFalse: [aLiteral printString].
			binaryBlock value: 'Users of ', autoSelect value: autoSelect]! !

!SystemNavigation methodsFor: 'query' stamp: ''!
hierarchyOfClassesSurrounding: aClass
	"Answer a list of classes in the hierarchy both above and below the given class"
	"SystemNavigation default hierarchyOfClassesSurrounding: StringHolder"
	
	| list aClassNonMeta isMeta |
	aClass ifNil: [^ OrderedCollection new].
	aClass ifNil: [^ self].
	aClassNonMeta := aClass theNonMetaClass.
	isMeta := aClassNonMeta ~~ aClass.
	list := OrderedCollection new.
	aClass allSuperclasses reverseDo:
		[:cl | list addLast: cl].
	aClassNonMeta allSubclassesWithLevelDo:
		[:cl :level | | theClassOrMeta |
		theClassOrMeta := isMeta ifTrue: [cl class] ifFalse: [cl].
		list addLast: theClassOrMeta]
	 	startingLevel: 0.
	^ list

! !

!SystemNavigation methodsFor: 'query' stamp: ''!
hierarchyOfImplementorsOf: aSelector forClass: aClass
	"Answer a list of classes in the hierarchy both above and below the given class which implement the given selector."
	"SystemNavigation default hierarchyOfImplementorsOf: #contents forClass: StringHolder"

	^ (self hierarchyOfClassesSurrounding: aClass) select:
		[:cl | cl includesSelector: aSelector]
! !

!SystemNavigation methodsFor: 'initialize' stamp: 'StephaneDucasse 10/23/2010 15:14'!
initialize

	super initialize.
	self environment: Smalltalk globals! !

!SystemNavigation methodsFor: 'message sends' stamp: ''!
isMessage: selector sentInClass: aClass 
	^(self allSendersOf: selector inClass: aClass) notEmpty ! !

!SystemNavigation methodsFor: 'message sends' stamp: ''!
isMessage: selector sentInClassCategory: category 
	^(self allSendersOf: selector inClassCategory: category) notEmpty ! !

!SystemNavigation methodsFor: 'message sends' stamp: ''!
isMessage: selector sentInPackageNamed: packageName 
	^(self allSendersOf: selector inPackageNamed: packageName) notEmpty ! !

!SystemNavigation methodsFor: 'query' stamp: ''!
isThereAnImplementorOf: aSelector 
	"Answer true if there is at least one implementor of the selector found  
	in the system, false if there are no implementors"
	"self new isThereAnImplementorOf: #contents.  
	self new isThereAnImplementorOf: #nobodyImplementsThis."
	self 
		allBehaviorsDo: [:class | (class includesSelector: aSelector)
				ifTrue: [^ true]].
	^ false! !

!SystemNavigation methodsFor: 'message sends' stamp: ''!
isUnsentMessage: selector

	| special byte |
	special := Smalltalk 
				hasSpecialSelector: selector
				ifTrueSetByte: [ :b | byte := b ].
	self allBehaviorsDo: 
		[ :behavior | 
			(behavior 
				thoroughHasSelectorReferringTo: selector
				special: special
				byte: byte) ifTrue: [^false] ].
	^true! !

!SystemNavigation methodsFor: 'utilities' stamp: ''!
methodsWithInitials: targetInitials
	"Based on a do-it contributed to the Squeak mailing list by Goran Hultgen:
 Browse methods whose initials (in the time-stamp, as logged to disk) match the given initials.
 Print out the complete time-stamp table to the Transcript.
 Answer a list of (initials -> count) associations.

CAUTION: It may take several minutes for this to complete."

	"Time millisecondsToRun: [SystemNavigation default methodsWithInitials: 'bf'] "

	| initials timeStamp  allSubmitters |
	initials := ''.
	timeStamp := ''.
	allSubmitters := Bag new.
	self browseAllSelect:
			[:cm |
				timeStamp := cm timeStamp.
				initials := timeStamp isEmpty
					ifTrue:
						['']
					ifFalse:
						[timeStamp substrings first].
				initials := initials isEmpty
					ifTrue:
						['<no time stamp>']
					ifFalse:
						[initials first isDigit
							ifTrue:
								['<date>']
							ifFalse:
								[initials]].
				allSubmitters add: initials.
				(initials = targetInitials)]
		name: ('Methods with initials ', targetInitials)
		autoSelect: nil.

	allSubmitters sortedCounts do: [:elem | Transcript cr; show: elem asString].
	^ allSubmitters
! !

!SystemNavigation methodsFor: 'query' stamp: ''!
numberOfImplementorsOf: aSelector 
	"Answer a count of the implementors of the given selector found in the  
	system"
	"self new numberOfImplementorsOf: #contents.  
	self new numberOfImplementorsOf: #nobodyImplementsThis.  
	self new numberOfimplementorsOf: #numberOfImplementorsOf:."
	| aCount |
	aCount := 0.
	self
		allBehaviorsDo: [:class | (class includesSelector: aSelector)
				ifTrue: [aCount := aCount + 1]].
	^ aCount! !

!SystemNavigation methodsFor: 'query' stamp: ''!
obsoleteBehaviors
	"SystemNavigation default obsoleteBehaviors inspect"
	"Find all obsolete behaviors including meta classes"

	| obs |
	obs := OrderedCollection new.
	Smalltalk garbageCollect.
	self 
		allObjectsDo: [:cl | (cl isBehavior
					and: [cl isObsolete])
				ifTrue: [obs add: cl]].
	^ obs asArray! !

!SystemNavigation methodsFor: 'query' stamp: ''!
obsoleteClasses   

	"SystemNavigation default obsoleteClasses inspect"
	"NOTE:  Also try inspecting comments below"
	| obs |
	obs := OrderedCollection new.  Smalltalk garbageCollect.
	Metaclass allInstancesDo:
		[:m | | c | c := m soleInstance.
		(c ~~ nil and: ['AnOb*' match: c name asString])
			ifTrue: [obs add: c]].
	^ obs asArray

"Likely in a ClassDict or Pool...
(Association allInstances select: [:a | (a value isKindOf: Class) and: ['AnOb*' match: a value name]]) asArray
"
"Obsolete class refs or super pointer in last lit of a method...
| n l found |
Smalltalk browseAllSelect:
	[:m | found := false.
	1 to: m numLiterals do:
		[:i | (((l := m literalAt: i) isMemberOf: Association)
				and: [(l value isKindOf: Behavior)
				and: ['AnOb*' match: l value name]])
			ifTrue: [found := true]].
	found]
"! !

!SystemNavigation methodsFor: 'query' stamp: ''!
obsoleteMethodReferences
	"SystemNavigation default obsoleteMethodReferences"

	"Open a browser on all referenced behaviors that are obsolete"

	| obsClasses references |
	references := Array new writeStream.
	obsClasses := self obsoleteBehaviors.
	'Scanning for methods referencing obsolete classes' 
		displayProgressAt: Sensor cursorPoint
		from: 1
		to: obsClasses size
		during: 
			[:bar | 
			obsClasses keysAndValuesDo: 
					[:index :each | | obsRefs | 
					bar value: index.
					obsRefs := each pointersToExcept: obsClasses.
					obsRefs do: 
							[:ref | 
							"Figure out if it may be a global"

							(ref isVariableBinding and: [ref key isString	"or Symbol"]) 
								ifTrue: 
									[(ref pointersTo) do: 
											[:meth | 
											(meth isKindOf: CompiledMethod) 
												ifTrue: [meth methodReference ifNotNil: [:mref | references nextPut: mref]]]]]]].
	^references contents! !

!SystemNavigation methodsFor: 'private' stamp: ''!
packageInfoFor: packageInfoOrString
	^packageInfoOrString isString
		ifTrue: [PackageInfo named: packageInfoOrString]
		ifFalse: [packageInfoOrString]! !

!SystemNavigation methodsFor: 'packages' stamp: ''!
removeAllButPackages: packagesList
	"Remove all classes and methods that do not belong to some of listed packages. Skip #unload. Do not use this method if you can use Monticello!! This method is used by some low-level shrinking process. Do not use it. This is quite hackish"

	| keep remove oldRemove extensionCategories log |
	keep := Set new.      
	packagesList do: [:pkg | keep addAll: (self allClassesInPackageNamed: pkg)].
	oldRemove := Set new.
 	[ 
		remove := Smalltalk globals allClassesAndTraits reject: [:c | keep includes: c].
		remove := remove collect: [:c | c name].
		(remove asSortedCollection: [:a :b | a  <= b ]) do: [:className |
			| class |
			class := (Smalltalk globals classOrTraitNamed: className).
			class allSubclasses ifEmpty: [
				SystemChangeNotifier uniqueInstance noMoreNotificationsFor: class.
				class isTrait ifFalse: [ 
					(class class includesSelector: #unload) ifTrue: [ class removeSelector: #unload ] ].
				class removeFromSystem. 
				] ].

		
		oldRemove ifNotEmpty: [
			oldRemove = remove ifTrue: [					
				 self error: 'Some classes still have subclasses' ] ].
		oldRemove := remove.
		remove notEmpty] whileTrue.

	keep do: [: class | 
		extensionCategories := (class organization categories select: [:category | 
			category first = $*]) asOrderedCollection.
		extensionCategories removeAllSuchThat: [:category |
			packagesList anySatisfy: [:package |
				((category asLowercase allButFirst = package asLowercase) or: [package asLowercase, '-*' match: category asLowercase allButFirst])]].
		extensionCategories do: [:category | 
			class removeCategory: category ]. 
		class isTrait ifFalse: [
			extensionCategories := (class class organization categories select: [:category | 
				category first = $*]) asOrderedCollection.
			extensionCategories removeAllSuchThat: [:category |
				packagesList anySatisfy: [:package |
					((category asLowercase allButFirst = package asLowercase) or: [package asLowercase, '-*' match: category asLowercase allButFirst])]].
			extensionCategories do: [:category | 
				class class removeCategory: category ]]].
			
! !

!SystemNavigation methodsFor: 'query' stamp: ''!
reportSenderCountsFor: selectorList 
	"Produce a report on the number of senders of each of the selectors in 
	the list. 1/27/96 sw"
	"SystemNavigation default reportSenderCountsFor: #(foo: printString)"
	| total report |
	total := 0.
	report := '
'.
	selectorList
		do: [:selector | | thisSize | 
			thisSize := (self allCallsOn: selector) size.
			report := report , thisSize printString , String tab , selector printString , String cr.
			total := total + thisSize].
	report := report , '--- ------------------
'.
	report := report , total printString , String tab , 'TOTAL
'.
	^ report! !

!SystemNavigation methodsFor: 'query' stamp: ''!
selectAllMethods: aBlock 
	"Answer a SortedCollection of each method that, when used as the block  
	argument to aBlock, gives a true result."
	| aCollection |
	aCollection := SortedCollection new.
	self
		allBehaviorsDo: [:class | class
				selectorsAndMethodsDo: [:sel :m | (aBlock value: m)
						ifTrue: [aCollection
								add: (SourcedMethodReference new setClass: class methodSymbol: sel)]]].
	^ aCollection! !

!SystemNavigation methodsFor: 'author' stamp: 'AlexandreBergel 12/29/2010 11:41'!
signatories
	"Answer the initials corresponding to the people who have signed the license"
	"SystemNavigation default signatories"
	
	^(self signatoriesString subStrings: {Character cr})
		collect: [ :each | each substrings first]! !

!SystemNavigation methodsFor: 'author' stamp: 'AlexandreBergel 12/29/2010 11:40'!
signatoriesString
	^'AB               Alexandre Bergel
ack              Alan Kay
ads              Adam Spitz
ajh              Anthony Hannan
aka              Mark Guzdial
al               Adrian Lienhard
aoy              Andres Otaduy
apb              Andrew P. Black
apl              Alain Plantec
ar               Andreas Raab
asm              Alejandro Magistrello
avi              Avi Bryant
bf               Bert Freudenberg
BG               Boris Gaertner
bh               Bob Hartwig
BJP              Bijan Parsia
bkv              Brent Vukmer
bolot            Bolot Kerimbaev
bootstrap        Pavel Krivanek
BP               Bijan Parsia
brp              Brent Pinkney
btr              Brian Rice
cbc              Chris Cunningham
ccn              Chris Norton
CdG              Cees de Groot
cds              C. David Shaffer
cmm              Chris Muller
crl              Craig Latta
cwp              Colin Putney
daf              Dave Faught
dao              danil osipchuk
DAS              David A Smith
dc               Damien Cassou
dd               Dominique Dutoit
dew              Doug Way
DF               Diego Fernandez
dgd              Diego Gomez Deck
dhhi             Dan Ingalls
di               Dan Ingalls
DM               Duane Maxwell
DSM              Duane Maxwell
dtl              Dave Lewis
dvf              Daniel Vainsencher
edc              Edgar DeCleene
efc              Eddie Cottongim
efo              Emilio Oca
em               Ernest Micklei?
emm              Ernest Micklei
es               Enrico Spinielli
FBS              Frank Shearar
fc               Frank Caggiano
fcs              Frank Sergeant
fm               Florin Mateoc
gh               Goran Krampe (nee Hultgren)
gk               Goran Krampe (nee Hultgren)
gm               German Morales
go               Georg Gollmann
gsa              German Arduino
HEG              Henrik Ekenberg
HilaireFernandes Hilaire Fernandes
HK               Herbert Konig
hmm              Hans-Martin Mosner
hpt              Hernan Tylim
huma             Lyndon Tremblay
ich.             Yuji Ichikawa
ikp              Ian Piumarta
jaf              Jan Fietz
jam              Javier Musa
jcg              Joshua Gargus
jdf              David Farber
jdr              Javier Diaz-Reinoso
je               Jorn Eyrich
je77             Jochen Rick
JF               Julian Fitzell
jhm              John Maloney
jlb              Jim Benson
jmb              Hans Baveco
JMM              John McIntosh
JMV              Juan Manuel Vuletich
jon              Jon Hylands
jrm              John-Reed Maffeo
jrp              John Pierce
jsp              Jeff Pierce
ka               Kazuhiro Abe
kfr              Karl Ramberg
KLC              Ken Causey
KR               korakurider
KTT              Kurt Thams
kwl              Klaus D. Witzel
ky               Koji Yokokawa
laza             Alexander Lazarevic
LB               Leo Burd
LC               Leandro Caniglia
LEG              Gerald Leeb
len              Luciano Esteban Notarfrancesco
lr               Lukas Renggli
lrs              Lorenzo Schiavina
ls               Lex Spoon
m3r              Maurice Rabb
mas              Mark Schwenk
MD               Markus Denker
mdr              Mike Rutenberg
mga              Markus Galli
miki             Mikael Kindborg
mikki            Mikael Kindborg
mir              Michael Rueger
mist             Michal Starke
mjg              Mark Guzdial
mjr              Mike Roberts
mjt              Mike Thomas
mk               Matej Kosik
MPW              Marcel Weiher
mrm              Martin McClure
ms               math
MU               Masashi Umezawa
mw               Martin Wirblat
nb               Naala Brewer
nice             Nicolas Cellier
nk               Ned Konz
Noury            Noury Bouraqadi
NS               Nathanael Schaerli
PHK              Peter Keeler
pk               Pavel Krivanek
pm               Patrick Mauritz
pnm              Paul McDonough
RAA              Bob Arning
raok             Richard A. O"Keefe
rbb              Brian Brown
rca              Russell Allen
reThink          Paul McDonough
rew              Roger Whitney
rhi              Robert Hirschfeld
Rik              Rik Fischer SmOOdy
RJT              Ron Teitelbaum
rr               Romain Robbes
rw               Roel Wuyts
rww              Robert Withers
sbw              Stephan B. Wessels
SD               Stephane Ducasse
sge              Steve Elkins
shrink           Pavel Krivanek
slg              Steve Gilbert
sm               Simon Michael
sma              Stefan Matthias Aust
sn               Suslov Nikolay
spfa             Stephane Rollandin
sps              Steven Swerling
SqR              Andres Valloud
SqR!!!!            Andres Valloud
SqR!!!!!!!!          Andres Valloud
sr               Stephan Rudlof
ssa              Sam S. Adams
st               Samuel Tardieu
stephaneducasse  Stephane Ducasse
stp              Stephen Travis Pope
sumim            Masato Sumi
svp              Stephen Vincent Pair
sw               Scott Wallace
T2               Toshiyuki Takeda
tak              Takashi Yamamiya
tao              Tim Olson
tb               Todd Blanchard
TBn              Torsten Bergmann
tetha            Tetsuya Hayashi
tfei             The Fourth Estate, Inc.
th               Torge Husfeldt
ti               Tobias Isenberg
TJ               TJ Leone
tk               Ted Kaehler
tk               Thomas Kowark
tlk              Tom Koenig
TN               korakurider
TPR              Tim Rowledge
TRee             Trygve Reenskaug
Tsutomu          Tsutomu Hiroshima
tween            Andy Tween
vb               Vassili Bykov
vj               Vladimir Janousek
wiz              Jerome Peace
wod              Bill Dargel
ykoubo           Koji Yokokawa
yo               Yoshiki Ohshima
zz               Serge Stinckwich
jmv
jm
jmm
md 			Marcus Denker
sd  			Stephane Ducasse
tpr'! !

!SystemNavigation methodsFor: 'query' stamp: ''!
unimplemented
	"Answer an Array of each message that is sent by an expression in a method but is not implemented by any object in the system."
	
	| all unimplemented |
	all := IdentitySet new: Symbol instanceCount * 2.
	self allBehaviorsDo: [:cl | cl selectorsDo: [:aSelector | all add: aSelector]].

	unimplemented := IdentityDictionary new.
	self allBehaviorsDo: [:cl |
		 cl selectorsAndMethodsDo: [:sel :meth |
			meth messages do: [:m | | entry |
				(all includes: m) ifFalse: [
					entry := unimplemented at: m ifAbsent: [Array new].
					entry := entry copyWith: (cl name, '>', sel).
					unimplemented at: m put: entry]]]].
	^ unimplemented
! !

!SystemNavigation methodsFor: 'message sends' stamp: 'LucFabresse 10/12/2010 08:25'!
unsentMessagesInCategory: category
	^(self environment organization classesInCategory: category) inject: Set new into: [:unsentMessages :class|
		unsentMessages, (self unsentMessagesInClass: class)]
	! !

!SystemNavigation methodsFor: 'message sends' stamp: ''!
unsentMessagesInClass: aClass 
	|methReferences|
	methReferences := Set new.
	aClass selectorsDo: [:selector|
		(self isUnsentMessage: selector) ifTrue: [
			methReferences add: (SourcedMethodReference class: aClass selector: selector)]].
	^methReferences! !

!SystemNavigation methodsFor: 'message sends' stamp: 'LucFabresse 10/12/2010 08:28'!
unsentMessagesInPackageNamed: packageName
	| unsentMessages |
	unsentMessages := self unsentMessagesInCategory: packageName.	
	(self categoriesInPackageNamed: packageName) do: [:category|
		unsentMessages addAll: (self unsentMessagesInCategory: category)].
	^unsentMessages 
	! !

!SystemNavigation methodsFor: 'query' stamp: 'LucFabresse 11/2/2010 22:34'!
unusedBlocks
       "Answer all methods that contain a block that is not used (not
        sent a message, returned, passed as an argument, or assigned)."
       "SystemNavigation new unusedBlocks"
       "SystemNavigation new
               browseMessageList: SystemNavigation new unusedBlocks
               name: 'unused blocks'"
       ^self allSelect:
               [:m| | is |
               is := InstructionStream on: m.
               is scanFor: [:b| b = 143 and: [(m at: is thirdByte * 256 + is fourthByte + is pc + 4) = 135]]]! !

Object subclass: #SystemVersion
	instanceVariableNames: 'version date highestUpdate updates type major minor suffix'
	classVariableNames: 'Current'
	poolDictionaries: ''
	category: 'System-Support'!

!SystemVersion commentStamp: 'StephaneDucasse 12/8/2010 13:50' prior: 0!
I am responsible for maintaining what version of Squeak and the VM is running.  I also track all of the update items that have been included in the image.

I'm invoked at auto start to get the latest plugins, etc.

Some queries are
	SystemVersion current major
	SystemVersion current minor
	SystemVersion current suffix
	SystemVersion current	highestUpdate!

!SystemVersion class methodsFor: 'updating' stamp: ''!
check: pluginVersion andRequestPluginUpdate: updateURL
	"SystemVersion check: 'zzz' andRequestPluginUpdate: 'http://www.squeakland.org/installers/update.html' "

	"We don't have a decent versioning scheme yet, so we are basically checking for a nil VM version on the mac."
	(self pluginVersion: pluginVersion newerThan: self currentPluginVersion)
		ifFalse: [^true].
	(self confirm: 'There is a newer plugin version available. Do you want to install it now?')
		ifFalse: [^false].
	HTTPClient
		requestURL: updateURL , (OSPlatform platformName copyWithout: Character space) asLowercase , '.html'
		target: '_top'.
	^false! !

!SystemVersion class methodsFor: 'updating' stamp: 'Jb 10/23/2010 11:39'!
checkAndApplyUpdates: availableUpdate
	"SystemVersion checkAndApplyUpdates: nil"

	^(availableUpdate isNil
		or: [availableUpdate > SystemVersion current highestUpdate])
		ifTrue: [
			(self confirm: 'There are updates available. Do you want to install them now?')
				ifFalse: [^false].
			UpdateStreamer new saveLocally: false; upToNumber: availableUpdate.
			
			Smalltalk snapshot: true andQuit: false.
			true]
		ifFalse: [false]! !

!SystemVersion class methodsFor: 'accessing' stamp: ''!
current
	Current ifNil: [Current := SystemVersion new].
	^Current! !

!SystemVersion class methodsFor: 'updating' stamp: ''!
currentPluginVersion
	^Smalltalk vmVersion! !

!SystemVersion class methodsFor: 'instance creation' stamp: ''!
fromShortVersionString: aString
	"Create a new SystemVersion from the given short version string. Useful for testing."

	| parts |
	parts := self parseShortVersionString: aString.
	^self new
		type: parts first;
		major: parts second;
		minor: parts third;
		suffix: parts fourth;
		yourself! !

!SystemVersion class methodsFor: 'instance creation' stamp: ''!
newVersion: versionName
	"Install a new SystemVersion as the current version. Only preserve the higest update recorded.
	This method is a candidate for being removed, I am not sure it ever makes sense to do this."
	
	| newVersion |
	newVersion := self fromShortVersionString: versionName.
	newVersion
		highestUpdate: self current highestUpdate.
	Current := newVersion
! !

!SystemVersion class methodsFor: 'accessing' stamp: ''!
parseShortVersionString: aString 
	"Answer the parts of aString, expected in short format <type><major>.<minor><suffix>."

	| stream length type major minor suffix |

	stream := aString readStream.
	[stream peek isDigit] whileFalse: [stream next].
	length := stream position.
	stream reset.
	type := (stream next: length).
	major := (stream upTo: $.) asInteger.
	stream := stream upToEnd readStream.
	[stream peek notNil and: [stream peek isDigit]] whileTrue: [stream next].
	length := stream position.
	stream reset.
	minor := (stream next: length) asInteger.
	suffix := stream upToEnd.
	^{type. major. minor. suffix}! !

!SystemVersion class methodsFor: 'accessing' stamp: ''!
parseVersionString: versionString 
	"Answer the version of this release as version, date, update."
	"SystemVersion parseVersionString: 'Squeak3.1alpha of 28 February 2001 [latest update: #3966]' "
	| version date update |
	
	[ | stream | stream := versionString readStream.
	version := stream upToAll: ' of '.
	date := Date readFrom: stream.
	stream upToAll: ' #'.
	update := Number readFrom: stream ] 
		on: Error
		do: [ ^ nil ].
	^ {  version. date. update  }! !

!SystemVersion class methodsFor: 'accessing' stamp: ''!
pluginVersion: availableVersionString newerThan: currentVersionString
	| currentVersion availableVersion |
	(currentVersionString isEmptyOrNil
		or: [availableVersionString isEmptyOrNil])
		ifTrue: [^true].
	currentVersion := self parseVersionString: currentVersionString.
	availableVersion := self parseVersionString: availableVersionString.
	(currentVersion isNil
		or: [availableVersion isNil])
		ifTrue: [^false].
	^(currentVersion at: 2) < (availableVersion at: 2)! !

!SystemVersion methodsFor: 'accessing' stamp: ''!
date
	^date! !

!SystemVersion methodsFor: 'accessing' stamp: ''!
date: newDate
	date := newDate! !

!SystemVersion methodsFor: 'accessing' stamp: ''!
datedVersion
	"Answer the version of this release."

	^ self version asString , ' of ' , self date printString! !

!SystemVersion methodsFor: 'accessing' stamp: 'StephaneDucasse 12/8/2010 13:50'!
highestUpdate
	"Returns the highest number of available updates"

	| sortedUpdates |
	highestUpdate ifNil: [
		sortedUpdates := self updates asSortedCollection.
		highestUpdate := (sortedUpdates isEmpty
			ifTrue: [0]
			ifFalse: [sortedUpdates last])].
	^highestUpdate! !

!SystemVersion methodsFor: 'accessing' stamp: ''!
highestUpdate: anInteger
	highestUpdate := anInteger! !

!SystemVersion methodsFor: 'printing' stamp: ''!
imageVersionString
	"Print the version compatibile with naming the image."
	
	^String streamContents: [:str |
		str nextPutAll: self type;
			nextPut: $-;
			nextPutAll: self major printString;
			nextPut: $.;
			nextPutAll: self minor printString;
			nextPut: $-;
			nextPutAll: self highestUpdate printString;
			nextPut: $-;
			nextPutAll: self suffix]! !

!SystemVersion methodsFor: 'accessing' stamp: ''!
includesUpdate: anUpdate
	^self updates includes: anUpdate! !

!SystemVersion methodsFor: 'initialize' stamp: ''!
initialize
	super initialize.
	date := Date today.
	updates := Set new.
! !

!SystemVersion methodsFor: 'accessing' stamp: 'StephaneDucasse 12/8/2010 13:49'!
major
	"Answer the major number of a version. 1 in '1.2'"

	^ major! !

!SystemVersion methodsFor: 'accessing' stamp: ''!
major: anObject
	"Set the value of major"

	major := anObject! !

!SystemVersion methodsFor: 'accessing' stamp: ''!
majorMinorVersion
	"Return the major/minor version number of the form X.Y, without any 'alpha' or 'beta' or other suffix. 
	If the receiver needs upgrading, do it. Remove the guard in 1.3."
	"SystemVersion current majorMinorVersion"
	
	self upgrade.
	^String streamContents: [:str |
		str 
			nextPutAll: self type;
			nextPutAll: self major printString; nextPut: $.;
			nextPutAll: self minor printString]! !

!SystemVersion methodsFor: 'accessing' stamp: 'StephaneDucasse 12/8/2010 13:50'!
minor
	"Answer the minor number of a version. 2 in '1.2'"

	^ minor! !

!SystemVersion methodsFor: 'accessing' stamp: ''!
minor: anObject
	"Set the value of minor"

	minor := anObject! !

!SystemVersion methodsFor: 'printing' stamp: ''!
printOn: stream
	stream
		nextPutAll: self datedVersion;
		nextPutAll: ' update ' , self highestUpdate printString! !

!SystemVersion methodsFor: 'accessing' stamp: ''!
registerUpdate: update
	self updates add: update.
	self resetHighestUpdate! !

!SystemVersion methodsFor: 'accessing' stamp: ''!
resetHighestUpdate
	highestUpdate := nil! !

!SystemVersion methodsFor: 'printing' stamp: ''!
shortVersionString
	"Print a short version string.  Compatibile with senders that concatenate this with 
	#lastUpdateString or #highestUpdate. Essentially the old version format. Guard 
	against no state being set."
	
	self type ifNil: [^'No version set'].
	^String streamContents: [:str |
		str 
			nextPutAll: self type;
			nextPutAll: self major printString;
			nextPut: $.;
			nextPutAll: self minor printString;
			nextPutAll: self suffix]! !

!SystemVersion methodsFor: 'accessing' stamp: 'StephaneDucasse 12/8/2010 13:50'!
suffix
	"Answer the suffix (alpha, beta, rc  of a version. beta in '1.2beta'"

	^ suffix! !

!SystemVersion methodsFor: 'accessing' stamp: ''!
suffix: anObject
	"Set the value of suffix"

	suffix := anObject! !

!SystemVersion methodsFor: 'accessing' stamp: ''!
type
	"Answer the value of type"

	^ type! !

!SystemVersion methodsFor: 'accessing' stamp: ''!
type: anObject
	"Set the value of type"

	type := anObject! !

!SystemVersion methodsFor: 'accessing' stamp: ''!
unregisterUpdate: update
	self updates remove: update ifAbsent: []! !

!SystemVersion methodsFor: 'accessing' stamp: ''!
updates
	^updates! !

!SystemVersion methodsFor: 'upgrading' stamp: ''!
upgrade

	| parts |
	version ifNil: [^self].
	parts := self class parseShortVersionString: version.
	self type: parts first.
	self major: parts second.
	self minor: parts third.
	self suffix: parts fourth.
	version := nil.! !

!SystemVersion methodsFor: 'accessing' stamp: ''!
version
	"Answer the short version string. If the receiver needs upgrading, do it. Remove the guard in 1.3."
	
	self upgrade.
	^self shortVersionString! !

!SystemVersion methodsFor: 'accessing' stamp: 'FabrizioPerin 10/25/2010 11:19'!
version: newVersion

	"In the follow the implementation of testDowngrade as example of usage of this API
	
	| sv |
	sv := SystemVersion fromShortVersionString: 'Pharo1.2ALPHA'.
	self assert: sv shortVersionString = 'Pharo1.2ALPHA'.
	self assert: (sv instVarNamed: 'version') isNil.
	self assert: sv suffix = 'ALPHA'.
	sv version: sv shortVersionString.
	self assert: (sv instVarNamed: 'version') = 'Pharo1.2ALPHA'.
	self assert: sv type isNil.
	self assert: sv major isNil.
	self assert: sv minor isNil.
	self assert: sv suffix isNil"

	self deprecated: 'Use individual accessors type, major, minor, suffix. Version is not a string.' on: '28 Sep 2010' in: 'Pharo1.2'.
	
	version := newVersion.
	type := major := minor := suffix := nil! !

!ClassDescription methodsFor: '*System-Support' stamp: ''!
allUnreferencedClassVariables
	"Answer a list of the names of all the receiver's unreferenced class vars, including those defined in superclasses"

	^ self systemNavigation allUnreferencedClassVariablesOf: self! !
SmalltalkImage initialize!
ExternalSemaphoreTable initialize!
ExternalSettings initialize!
DefaultExternalDropHandler initialize!
SecurityManager initialize!
AutoStart initialize!
ExternalDropHandler initialize!
